#ifndef FRAME_CPP_INTERFACE__FRAME_CPP_HH
#define FRAME_CPP_INTERFACE__FRAME_CPP_HH

#include <string>

#include "general/types.hh"

#include "framecpp/config.h"

#define FRAME_SPEC_MIN		3
#define FRAME_SPEC_CURRENT	6
#define FRAME_SPEC_MAX		6

#define ENABLE_DEPRICATED	1

#if FRAME_SPEC_CURRENT > 5
#define	DAQ_SUPPORT 0
#else
#define	DAQ_SUPPORT 1
#endif /* FRAME_SPEC_CURRENT > 5 */

namespace FrameCPP
{
  //---------------------------------------------------------------------
  // enumerated type describing the byte ordering
  //---------------------------------------------------------------------
  enum byte_order_type {
    BYTE_ORDER_BIG_ENDIAN,
    BYTE_ORDER_LITTLE_ENDIAN,
#if WORDS_BIGENDIAN
    BYTE_ORDER_HOST = BYTE_ORDER_BIG_ENDIAN
#else	/* WORDS_BIGENDIAN */
    BYTE_ORDER_HOST = BYTE_ORDER_LITTLE_ENDIAN
#endif	/* WORDS_BIGENDIAN */

  };

  //---------------------------------------------------------------------
  // List of version namespaces
  //---------------------------------------------------------------------
  namespace Version_3_4_5{ }	// Supports version 3, 4, and 5 of spec
  namespace Version_6{ }	// Supports version 6 of spec

  //---------------------------------------------------------------------
  // Typedef for version information
  //---------------------------------------------------------------------

  typedef CHAR_U version_type;

  //---------------------------------------------------------------------
  // Typedef for STRING type in frames
  //---------------------------------------------------------------------

  typedef std::string STRING;

  //---------------------------------------------------------------------
  // Constants representing version information
  //---------------------------------------------------------------------
  //: Establish the default version namespace
#if FRAME_SPEC_CURRENT == 5
  namespace Version = Version_3_4_5;
#elif FRAME_SPEC_CURRENT == 6
  namespace Version = Version_6;
#else
#error Version namespace not defined for FRAME_SPEC_CURRENT
#endif /* PRAME_SPEC_CURRENT */

  //: The earliest supported frame specification
  const version_type FrameSpecMinimum( FRAME_SPEC_MIN );
  //: Currentently supported frame specification
  const version_type FrameSpecCurrent( FRAME_SPEC_CURRENT );
  //: Maximum supported frame spec
  // This may be one larger than the current frame specification
  const version_type FrameSpecMaximum( FRAME_SPEC_MAX );
  //: Byte ordering of host
  const int HostByteOrder( BYTE_ORDER_HOST );
} // namespace FrameCPP

#endif	/* FRAME_CPP_INTERFACE__FRAME_CPP_HH */
