#include "framecpp/config.h"

#include "framecpp/FrameCPP.hh"	// Need FRAME_SPEC_CURRENT

#if FRAME_SPEC_CURRENT <= 5

#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>

#if FRAME_SPEC_CURRENT == 5
#include "framecpp/generalreader.hh"
#include "framecpp/daqframe.hh"
#endif	/* FRAME_SPEC_CURRENT */

#include "framecpp/FrameH.hh"
#include "framecpp/FrameReader.hh"
#include "framecpp/VFrameReader.hh"

using namespace FrameCPP;

//-----------------------------------------------------------------------
// Structure to hold private info for FrameReader class
//-----------------------------------------------------------------------
struct FrameReader::private_type
{
public:
  enum reader_mode_type
  {
    DAQ,
    DAQ_UPDATE,
    SLOW,
    UNKNOWN
  };

  private_type( );

  version_type				m_version;
  reader_mode_type				m_reader_mode;
  std::auto_ptr<VFrameReader>		m_frame_reader;
  FrameH*				m_frame;
  //:TODO: Add members for daq reader and daq update
};

inline FrameReader::private_type::
private_type( )
  : m_reader_mode( UNKNOWN )
{
}

//-----------------------------------------------------------------------
// Public Methods for FrameReader
//-----------------------------------------------------------------------

FrameReader::
FrameReader( )
  : m_private( new private_type( ) )
{
}

FrameReader::
FrameReader( std::istream& InStream, bool AutoReset )
  : m_private( new private_type( ) )
{
#if FRAME_SPEC_CURRENT == 5
  //---------------------------------------------------------------------
  // Use the general reader which does both daq and standard frames.
  //---------------------------------------------------------------------
    p( ).m_frame_reader.reset( new Version::GeneralReader( InStream,
							   AutoReset ) );
#else
#error Unsupported frame spec: FRAME_SPEC_CURRENT
#endif
}

FrameReader::
~FrameReader( )
{
  delete m_private;
  m_private = 0;
}

void FrameReader::
SetStream( std::istream& InStream, bool AutoReset )
{
#if FRAME_SPEC_CURRENT == 5
  //---------------------------------------------------------------------
  // Use the general reader which does both daq and standard frames.
  //---------------------------------------------------------------------
    p( ).m_frame_reader.reset( new Version::GeneralReader( InStream,
							   AutoReset ) );
#else
#error Unsupported frame spec: FRAME_SPEC_CURRENT
#endif
}

bool FrameReader::
isEndOfFile( ) const
{
  return p( ).m_frame_reader->isEndOfFile( );
}

FrameH* FrameReader::
readFrame( )
{
  std::auto_ptr<FrameH>			frame;

  if ( ! p( ).m_frame_reader.get( ) )
  {
    throw std::logic_error( "FrameReader::readFrame: frame reader is null" );
  }
#if FRAME_SPEC_CURRENT == 5
  switch( p( ).m_reader_mode )
  {
  case private_type::DAQ_UPDATE:
    try
    {
      Version::GeneralReader* r( dynamic_cast< Version::GeneralReader* >
				 ( p( ).m_frame_reader.get( ) ) );
      Version::DaqFrame*
	daqframe( dynamic_cast< Version::DaqFrame* >( p( ).m_frame ) );
      if ( daqframe == (Version::DaqFrame*)NULL )
      {
	throw std::bad_cast();
      }
      daqframe->daqUpdate( r->GetStreamPtr( ) );
      return daqframe;
    }
    catch( ... )
    {
      p( ).m_reader_mode = private_type::UNKNOWN;
    }
    // :TRICKY: This fall through is intended
    // If there was a problem updating the frame, then reread the frame
  case private_type::UNKNOWN:
  case private_type::DAQ:
  case private_type::SLOW:
    frame.
      reset( dynamic_cast< Version::GeneralReader* >
	     ( p( ).m_frame_reader.get( ) )->readFrame( ) );
    if ( dynamic_cast< Version::DaqFrame* >( frame.get( ) ) )
    {
      p( ).m_reader_mode = private_type::DAQ_UPDATE;
    }
    else
    {
      p( ).m_reader_mode = private_type::SLOW;
    }
    break;
  default:
    std::ostringstream	oss;
    oss << "GeneralReader::readFrame: Unknown mode: " << p( ).m_reader_mode;
    throw std::range_error( oss.str( ).c_str( ) );
  }
  p( ).m_frame = frame.release( );
  return p( ).m_frame;
#else	/* FRAME_SPEC_CURRENT */
#error FrameReader::readFrame does not support FRAME_SPEC_CURRENT
#endif	/* FRAME_SPEC_CURRENT */
}

//-----------------------------------------------------------------------
// private - classes
//-----------------------------------------------------------------------

#endif /* FRAME_SPEC_CURRENT <= 5 */
