#ifndef FrameCppBaseHH
#define FrameCppBaseHH

#include <string>

#include "framecpp/config.h"
#include "types.hh"
#include "errors.hh"


namespace FrameCPP {

namespace Version_3_4_5 {

class Output;


//-----------------------------------------------------------------------------
//
//: The base class for all Frame Structure objects.
//
class Base
{
public:

    /* Constructors */
    Base();
    Base( const Base& base );

    /* Destructor */
    virtual ~Base();

    /* Required Methods */

    //: Get derived class ID.
    virtual INT_2U getClassId() const = 0;
    //: Give the name for the derived class
    std::string getClassName() const;

    
private: friend class Output;
    //: Write this object into the output data stream object.
    //!exc: write_failure - Write failure occured.
    virtual void write( Output& out ) const = 0;    
};



//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
//!exc: None.
//   
inline Base::Base()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const Base& base - Object to be copied.
//
//
//!exc: None.
//      
inline Base::Base( const Base& base )
{
}



 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#define DECLARE_DEFAULT_SIMPLE_CONTAINER( BIGNAME, LITTLENAME, TYPE )         \
    typedef FrameCPP::Version_3_4_5::Container< TYPE > BIGNAME##Container;    \
    typedef BIGNAME##Container::const_iterator const_iterator;                \
    typedef BIGNAME##Container::iterator iterator;                            \
    iterator begin()                                                          \
        { return m##BIGNAME.begin(); }                                        \
    const_iterator begin() const                                              \
        { return m##BIGNAME.begin(); }                                        \
    iterator end()                                                            \
        { return m##BIGNAME.end(); }                                          \
    const_iterator end() const                                                \
        { return m##BIGNAME.end(); }                                          \
    size_t size() const { return m##BIGNAME.size(); }                   \
    TYPE* get##BIGNAME( size_t index ) { return m##BIGNAME[ index ]; }        \
    const TYPE* get##BIGNAME( size_t index ) const                            \
        { return m##BIGNAME[ index ]; }                                       \
    TYPE* operator[]( size_t index ) { return m##BIGNAME[ index ]; }          \
    const TYPE* operator[]( size_t index ) const                              \
        { return m##BIGNAME[ index ]; }                                       \
    iterator append( const TYPE& LITTLENAME )                                 \
        { return m##BIGNAME.append( LITTLENAME ); }                           \
    iterator append( const TYPE* LITTLENAME )                                 \
        { return m##BIGNAME.append( LITTLENAME ); }                           \
    iterator append(                                                          \
        TYPE* LITTLENAME, bool allocate = true, bool owns = true )            \
        { return m##BIGNAME.append( LITTLENAME, allocate, owns ); }           \
    iterator insert(                                                          \
        iterator iter, const TYPE& LITTLENAME )                               \
        { return m##BIGNAME.insert( iter, LITTLENAME ); }                     \
    iterator insert(                                                          \
        iterator iter, const TYPE* LITTLENAME )                               \
        { return m##BIGNAME.insert( iter, LITTLENAME ); }                     \
    iterator insert(                                                          \
        iterator iter, TYPE* LITTLENAME, bool allocate = true,                \
        bool owns = true )                                                    \
        { return m##BIGNAME.insert( iter, LITTLENAME, allocate, owns ); }     \
    iterator insert(                                                          \
        size_t index, const TYPE& LITTLENAME )                                \
        { return m##BIGNAME.insert( index, LITTLENAME ); }                    \
    iterator insert(                                                          \
        size_t index, const TYPE* LITTLENAME )                                \
        { return m##BIGNAME.insert( index, LITTLENAME ); }                    \
    iterator insert(                                                          \
        size_t index, TYPE* LITTLENAME, bool allocate = true,                 \
        bool owns = true )                                                    \
        { return m##BIGNAME.insert( index, LITTLENAME, allocate, owns ); }    \
    void erase( unsigned int index )                                          \
        { return m##BIGNAME.erase( index ); }                                 \
    void erase(                                                               \
        iterator start, iterator finish )                                     \
        { return m##BIGNAME.erase( start, finish ); }                         


#define DECLARE_SIMPLE_CONTAINER( BIGNAME, LITTLENAME, TYPE)                  \
    typedef Container< TYPE > BIGNAME##Container;                             \
    typedef BIGNAME##Container::const_iterator const_##LITTLENAME##_iterator; \
    typedef BIGNAME##Container::iterator LITTLENAME##_iterator



#define DECLARE_DEFAULT_CONTAINER( BIGNAME, LITTLENAME, TYPE, KEY )           \
    typedef FrameCPP::SearchContainer< TYPE, &TYPE::KEY, FrameCPP::Version_3_4_5::Container< TYPE > > BIGNAME##Container;           \
    typedef BIGNAME##Container::const_iterator const_iterator;                \
    typedef BIGNAME##Container::iterator iterator;                            \
    typedef BIGNAME##Container::const_hash_iterator const_hash_iterator;      \
    typedef BIGNAME##Container::hash_iterator hash_iterator;                  \
    iterator begin()                                                          \
        { return m##BIGNAME.begin(); }                                        \
    const_iterator begin() const                                              \
        { return m##BIGNAME.begin(); }                                        \
    iterator end()                                                            \
        { return m##BIGNAME.end(); }                                          \
    const_iterator end() const                                                \
        { return m##BIGNAME.end(); }                                          \
    size_t size() const { return m##BIGNAME.size(); }                   \
    TYPE* get##BIGNAME( size_t index ) { return m##BIGNAME[ index ]; }        \
    const TYPE* get##BIGNAME( size_t index ) const                            \
        { return m##BIGNAME[ index ]; }                                       \
    TYPE* operator[]( size_t index ) { return m##BIGNAME[ index ]; }          \
    const TYPE* operator[]( size_t index ) const                              \
        { return m##BIGNAME[ index ]; }                                       \
    iterator append( const TYPE& LITTLENAME )                                 \
        { return m##BIGNAME.append( LITTLENAME ); }                           \
    iterator append( const TYPE* LITTLENAME )                                 \
        { return m##BIGNAME.append( LITTLENAME ); }                           \
    iterator append(                                                          \
        TYPE* LITTLENAME, bool allocate = true, bool owns = true )            \
        { return m##BIGNAME.append( LITTLENAME, allocate, owns ); }           \
    iterator insert(                                                          \
        iterator iter, const TYPE& LITTLENAME )                               \
        { return m##BIGNAME.insert( iter, LITTLENAME ); }                     \
    iterator insert(                                                          \
        iterator iter, const TYPE* LITTLENAME )                               \
        { return m##BIGNAME.insert( iter, LITTLENAME ); }                     \
    iterator insert(                                                          \
        iterator iter, TYPE* LITTLENAME, bool allocate = true,                \
        bool owns = true )                                                    \
        { return m##BIGNAME.insert( iter, LITTLENAME, allocate, owns ); }     \
    iterator insert(                                                          \
        size_t index, const TYPE& LITTLENAME )                                \
        { return m##BIGNAME.insert( index, LITTLENAME ); }                    \
    iterator insert(                                                          \
        size_t index, const TYPE* LITTLENAME )                                \
        { return m##BIGNAME.insert( index, LITTLENAME ); }                    \
    iterator insert(                                                          \
        size_t index, TYPE* LITTLENAME, bool allocate = true,                 \
        bool owns = true )                                                    \
        { return m##BIGNAME.insert( index, LITTLENAME, allocate, owns ); }    \
    void erase( unsigned int index )                                          \
        { return m##BIGNAME.erase( index ); }                                 \
    void erase(                                                               \
        iterator start, iterator finish )                                     \
        { return m##BIGNAME.erase( start, finish ); }                         \
    const_iterator find(                                                      \
        const std::string& name, const_iterator start ) const                 \
        { return m##BIGNAME.find( name, start ); }                            \
    iterator find(                                                            \
        const std::string& name, iterator start )                             \
        { return m##BIGNAME.find( name, start ); }                            \
    const_iterator regexFind(                                                 \
        const std::string& regex, const_iterator start ) const                \
        { return m##BIGNAME.regexFind( regex, start ); }                      \
    iterator regexFind(                                                       \
        const std::string& regex, iterator start )                            \
        { return m##BIGNAME.regexFind( regex, start ); }                      \
    std::pair< const_hash_iterator,                                           \
          const_hash_iterator >                                               \
        hashFind( const std::string& name ) const                             \
        { return m##BIGNAME.hashFind( name ); }                               \
    std::pair< hash_iterator, hash_iterator >                                 \
        hashFind( const std::string& name )                                   \
        { return m##BIGNAME.hashFind( name ); }        


#define DECLARE_CONTAINER( BIGNAME, LITTLENAME, TYPE, KEY )                   \
    typedef FrameCPP::SearchContainer< TYPE, &TYPE::KEY, FrameCPP::Version_3_4_5::Container< TYPE > > BIGNAME##Container;           \
    typedef BIGNAME##Container::const_iterator const_##LITTLENAME##_iterator; \
    typedef BIGNAME##Container::iterator LITTLENAME##_iterator;               \
    typedef BIGNAME##Container::const_hash_iterator                           \
        const_##LITTLENAME##_hash_iterator;                                   \
    typedef BIGNAME##Container::hash_iterator LITTLENAME##_hash_iterator

#endif // FrameCppBaseHH
