#include "framecpp/config.h"

#include "dimension.hh"


using FrameCPP::Version_3_4_5::Dimension;


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment operator.
//
//!param: const Dimension& dim - The object to assign from.
//
//!return: const Dimension& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Dimension& Dimension::operator=( const Dimension& dim )
{
    if ( this != &dim )
    {
        mNx = dim.mNx;
        mDx = dim.mDx;
        mStartX = dim.mStartX;
        mUnitX = dim.mUnitX;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
//!param: const Dimension& dim - The object to compare with.
//
//!return: bool - true if the objects are equal, false otherwise.
//
//!exc: None.   
//      
bool Dimension::operator==( const Dimension& dim ) const
{
    return ( ( mNx    == dim.mNx    ) &&
             ( mDx    == dim.mDx    ) &&
             ( mStartX== dim.mStartX) &&
             ( mUnitX == dim.mUnitX ) );
}


//-----------------------------------------------------------------------------
//
//: Not equal comparison.
//
//!param: const Dimension& dim - The object to compare with.
//
//!return: bool - false if the objects are equal, true otherwise.
//
//!exc: None.   
//      
bool Dimension::operator!=( const Dimension& dim ) const
{
    return !(dim == *this);
}
