#include "framecpp/config.h"

#include "dump.hh"
#include "frame.hh"
#include "adcdata.hh"
#include "msg.hh"
#include "history.hh"
#include "rawdata.hh"
#include "procdata.hh"
#include <iomanip>


const unsigned int INDENT_AMOUNT = 4;
const unsigned int DATA_WIDTH = 16;

std::ostream& operator<<( std::ostream& out, const Indent& indent )
{
    for ( unsigned int i = 0; i < indent.size; i++ )
    {
        out << " ";
    }

    return out;
}

//: Dump frame in ASCII.
void dump( std::ostream& out, FrameCPP::Version_3_4_5::Frame& fr, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Frame\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << fr.getName() << "\"\n"
        << Indent( d ) << "run: " << fr.getRun() << std::endl
        << Indent( d ) << "frame: " << fr.getFrame() << std::endl
        << Indent( d ) << "dataquality: " << fr.getDataQuality() << std::endl
        << Indent( d ) << "GTimeS: " << fr.getGTime().getSec() << std::endl
        << Indent( d ) << "GTimeN: " << fr.getGTime().getNSec() << std::endl
        << Indent( d ) << "ULeapS: " << fr.getULeapS() << std::endl
        << Indent( d ) << "dt: " << fr.getDt() << std::endl;

    if ( fr.refType().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refType().size(); ++i )
        {
            out << Indent( d ) << "type[ " << i << " ]: ";
            dump( out, *fr.refType()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "type: none" << std::endl;
    }

    if ( fr.refUser().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refUser().size(); ++i )
        {
            out << Indent( d ) << "user[ " << i << " ]: ";
            dump( out, *fr.refUser()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "user: none" << std::endl;
    }

    if ( fr.refDetectorSim().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refDetectorSim().size(); ++i )
        {
            out << Indent( d ) << "detectSim[ " << i << " ]: ";
            dump( out, *fr.refDetectorSim()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "detectorSim: none" << std::endl;
    }

    if ( fr.refDetectorProc().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refDetectorProc().size(); ++i )
        {
            out << Indent( d ) << "detectProc[ " << i << " ]: ";
            dump( out, *fr.refDetectorProc()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "detectorProc: none" << std::endl;
    }

    if ( fr.refHistory().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refHistory().size(); ++i )
        {
            out << Indent( d ) << "history[ " << i << " ]: ";
            dump( out, *fr.refHistory()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "history: none" << std::endl;
    }

    if ( fr.getRawData() != 0 )
    {
        out << std::endl << Indent( d ) << "rawData: ";
        dump( out, *fr.getRawData(), d );
    }
    else
    {
        out << Indent( d ) << "rawData: none" << std::endl;
    }
    
    if ( fr.refProcData().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refProcData().size(); ++i )
        {
            out << Indent( d ) << "procData[ " << i << " ]: ";
            dump( out, *fr.refProcData()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "procData: none" << std::endl;
    }

    if ( fr.refSimData().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refSimData().size(); ++i )
        {
            out << Indent( d ) << "simData[ " << i << " ]: ";
            dump( out, *fr.refSimData()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "simData: none" << std::endl;
    }

    if ( fr.refEvent().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refEvent().size(); ++i )
        {
            out << Indent( d ) << "Event[ " << i << " ]: ";
            dump( out, *fr.refEvent()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "Event: none" << std::endl;
    }

    if ( fr.refSimEvent().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refSimEvent().size(); ++i )
        {
            out << Indent( d ) << "SimEvent[ " << i << " ]: ";
            dump( out, *fr.refSimEvent()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "SimEvent: none" << std::endl;
    }

    if ( fr.refSummary().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refSummary().size(); ++i )
        {
            out << Indent( d ) << "summaryData[ " << i << " ]: ";
            dump( out, *fr.refSummary()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "summaryData: none" << std::endl;
    }

    if ( fr.refAux().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refAux().size(); ++i )
        {
            out << Indent( d ) << "auxData[ " << i << " ]: ";
            dump( out, *fr.refAux()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "auxData: none" << std::endl;
    }
        
    if ( fr.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < fr.refTable().size(); ++i )
        {
            out << Indent( d ) << "auxTable[ " << i << " ]: ";
            dump( out, *fr.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "auxTable: none" << std::endl;
    }
        
    out << Indent( i ) << "}" << std::endl;
}

void dump( std::ostream& out, FrameCPP::Version_3_4_5::AdcData& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "AdcData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"" << std::endl
        << Indent( d ) << "group: " << data.getChannelGroup() << std::endl
        << Indent( d ) << "channel number: " << data.getChannelNumber() << std::endl
        << Indent( d ) << "nBits: " << data.getNBits() << std::endl
        << Indent( d ) << "bias: " << data.getBias() << std::endl
        << Indent( d ) << "slope: " << data.getSlope() << std::endl
        << Indent( d ) << "units: \"" << data.getUnits() << "\"" << std::endl
        << Indent( d ) << "sampleRate: " << data.getSampleRate() << std::endl
        << Indent( d ) << "timeOffsetS: " << data.getTimeOffset().getSec()
        << std::endl
        << Indent( d ) << "timeOffsetN: " << data.getTimeOffset().getNSec()
        << std::endl
        << Indent( d ) << "fShift: " << data.getFShift() << std::endl
        << Indent( d ) << "phase: " << data.getPhase() << std::endl
        << Indent( d ) << "dataValid: " << data.getDataValid() << std::endl;
    
    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }

    if ( data.refAux().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refAux().size(); ++i )
        {
            out << Indent( d ) << "aux[ " << i << " ]: ";
            dump( out, *data.refAux()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "aux: none" << std::endl;
    }

    out << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::Detector& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Detector\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "longitudeD: " << data.getLongitude() << std::endl
        << Indent( d ) << "latitudeD: " << data.getLatitude() << std::endl
        << Indent( d ) << "elevation: " << data.getElevation() << std::endl
        << Indent( d ) << "armXAzimuth: " << data.getArmXAzimuth() << std::endl
        << Indent( d ) << "armYAzimuth: " << data.getArmYAzimuth() << std::endl
        << Indent( d ) << "armXAltitude: " << data.getArmXAltitude() << std::endl
        << Indent( d ) << "armYAltitude: " << data.getArmYAltitude() << std::endl
        << Indent( d ) << "armXMidpoint: " << data.getArmXMidpoint() << std::endl
        << Indent( d ) << "armYMidpoint: " << data.getArmYMidpoint() << std::endl
        << Indent( d ) << "localTime: " << data.getLocalTime() << std::endl
        << Indent( d ) << "dataQuality: " << data.getDataQuality() << std::endl
        << Indent( d ) << "qaBitList: " << data.getQaBitList() << std::endl;
    
    if ( data.refMore().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refMore().size(); ++i )
        {
            out << Indent( d ) << "more[ " << i << " ]: ";
            dump( out, *data.refMore()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "more: none" << std::endl;
    }

    if ( data.refMoreTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refMoreTable().size(); ++i )
        {
            out << Indent( d ) << "moreTable[ " << i << " ]: ";
            dump( out, *data.refMoreTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "moreTable: none" << std::endl;
    }

    if ( data.refStatData().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refStatData().size(); ++i )
        {
            out << Indent( d ) << "statData[ " << i << " ]: ";
            dump( out, *data.refStatData()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "statData: none" << std::endl;
    }
    
    out << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::Msg& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Msg\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "alarm: \"" << data.getAlarm() << "\"\n"
        << Indent( d ) << "message: \"" << data.getMessage() << "\"\n"
        << Indent( d ) << "severity: " << data.getSeverity() << std::endl
        << Indent( d ) << "GTimeS: " << data.getGTime().getSec() << std::endl
        << Indent( d ) << "GTimeN: " << data.getGTime().getNSec() << std::endl
        << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::History& h, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "History\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << h.getName() << "\"\n"
        << Indent( d ) << "time: " << h.getTime() << std::endl
        << Indent( d ) << "comment: \"" << h.getComment() << "\"\n"
        << Indent( i ) << "}" << std::endl;
}


void dump( std::ostream& out, FrameCPP::Version_3_4_5::RawData& r, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "RawData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << r.getName() << "\"\n";
    
    if ( r.refSerial().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < r.refSerial().size(); ++i )
        {
            out << Indent( d ) << "serData[ " << i << " ]: ";
            dump( out, *r.refSerial()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "serData: none" << std::endl;
    }
    
    if ( r.refAdc().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < r.refAdc().size(); ++i )
        {
            out << Indent( d ) << "adcData[ " << i << " ]: ";
            dump( out, *r.refAdc()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "adcData: none" << std::endl;
    }
    
    if ( r.refMsg().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < r.refMsg().size(); ++i )
        {
            out << Indent( d ) << "logMsg[ " << i << " ]: ";
            dump( out, *r.refMsg()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "logMsg: none" << std::endl;
    }

    if ( r.refMore().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < r.refMore().size(); ++i )
        {
            out << Indent( d ) << "more[ " << i << " ]: ";
            dump( out, *r.refMore()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "more: none" << std::endl;
    }
    
    out << Indent( i ) << "}" << std::endl;
}


void dump( std::ostream& out, FrameCPP::Version_3_4_5::ProcData& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "ProcData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"" << std::endl
        << Indent( d ) << "sampleRate: " << data.getSampleRate() << std::endl
        << Indent( d ) << "timeOffsetS: " << data.getTimeOffset().getSec() << std::endl
        << Indent( d ) << "timeOffsetN: " << data.getTimeOffset().getNSec() << std::endl
        << Indent( d ) << "fShift: " << data.getFShift() << std::endl
        << Indent( d ) << "phase: " << data.getPhase() << std::endl;

    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }

    if ( data.refAux().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refAux().size(); ++i )
        {
            out << Indent( d ) << "aux[ " << i << " ]: ";
            dump( out, *data.refAux()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "aux: none" << std::endl;
    }    

    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }

    out << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::SimData& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "SimData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"" << std::endl
        << Indent( d ) << "sampleRate: " << data.getSampleRate() << std::endl
        << Indent( d ) << "fShift: " << data.getFShift() << std::endl
        << Indent( d ) << "phase: " << data.getPhase() << std::endl;
    
    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }

    if ( data.refInput().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refInput().size(); ++i )
        {
            out << Indent( d ) << "input[ " << i << " ]: ";
            dump( out, *data.refInput()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "input: none" << std::endl;
    }

    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }

    out << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::SerData& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "SerData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "timeSec: " << data.getTime().getSec() << std::endl
        << Indent( d ) << "timeNsec: " << data.getTime().getNSec() << std::endl
        << Indent( d ) << "sampleRate: " << data.getSampleRate() << std::endl
        << Indent( d ) << "data: \"" << data.getData() << "\"\n";
    
    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "serial[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "serial: none" << std::endl;
    }

    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }

#ifdef not_def
    if ( data.refMore().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refMore().size(); ++i )
        {
            out << Indent( d ) << "more[ " << i << " ]: ";
            dump( out, *data.refMore()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "more: none" << std::endl;
    }
#endif

    out << Indent( i ) << "}" << std::endl;
}    


void dump( std::ostream& out, FrameCPP::Version_3_4_5::StatData& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "StatData\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"\n"
        << Indent( d ) << "representation: \"" << data.getRepresentation() << "\"\n"
        << Indent( d ) << "timeStart: " << data.getTimeStart() << std::endl
        << Indent( d ) << "timeEnd: " << data.getTimeEnd() << std::endl
        << Indent( d ) << "version: " << data.getVersion() << std::endl;

    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }
        
    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }

    out << Indent( i ) << "}" << std::endl;
}

void dump( std::ostream& out, FrameCPP::Version_3_4_5::SimEvent& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "SimEvent\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"\n"
        << Indent( d ) << "inputs: \"" << data.getInputs() << "\"\n"
        << Indent( d ) << "GTimeS: " << data.getGTime().getSec() << std::endl
        << Indent( d ) << "GTimeN: " << data.getGTime().getNSec() << std::endl
        << Indent( d ) << "timeBefore: " << data.getTimeBefore() << std::endl
        << Indent( d ) << "timeAfter: " << data.getTimeAfter() << std::endl
        << Indent( d ) << "amplitude: " << data.getAmplitude() << std::endl;

    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }

    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }
        
    out << Indent( i ) << "}" << std::endl;
}

void dump( std::ostream& out, FrameCPP::Version_3_4_5::Event& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Event\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"\n"
        << Indent( d ) << "inputs: \"" << data.getInputs() << "\"\n"
        << Indent( d ) << "GTimeS: " << data.getGTime().getSec() << std::endl
        << Indent( d ) << "GTimeN: " << data.getGTime().getNSec() << std::endl
        << Indent( d ) << "timeBefore: " << data.getTimeBefore() << std::endl
        << Indent( d ) << "timeAfter: " << data.getTimeAfter() << std::endl
        << Indent( d ) << "eventStatus: " << data.getEventStatus() << std::endl
        << Indent( d ) << "amplitude: " << data.getAmplitude() << std::endl

        << Indent( d ) << "probability: " << data.getProbability() << std::endl
        << Indent( d ) << "statistics: \"" << data.getStatistics() << "\"\n"
	<< Indent( d ) << "nParameters: " << data.getParameters().size() << std::endl;
    for (unsigned int j = 0; j < data.getParameters().size(); j++) {
      const std::pair<std::string, REAL_4>& p = data.getParameters()[j];
      out << Indent( d ) << "Parameter " << j << ": Name: \"" << p.first
	  << "\" Value: " << p.second <<std::endl;
    }

    if ( data.size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.size(); ++i )
        {
            out << Indent( d ) << "data[ " << i << " ]: ";
            dump( out, *data[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "data: none" << std::endl;
    }

    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }
        
    out << Indent( i ) << "}" << std::endl;
}


void dump( std::ostream& out, FrameCPP::Version_3_4_5::Summary& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Summary\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: \"" << data.getComment() << "\"\n"
        << Indent( d ) << "test: \"" << data.getTest() << "\"\n"
        << Indent( d ) << "GTimeS: " << data.getGTime().getSec() << std::endl
        << Indent( d ) << "GTimeN: " << data.getGTime().getNSec() << std::endl;
    
    if ( data.refMoments().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refMoments().size(); ++i )
        {
            out << Indent( d ) << "moments[ " << i << " ]: ";
            dump( out, *data.refMoments()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "moments: none" << std::endl;
    }
    if ( data.refTable().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refTable().size(); ++i )
        {
            out << Indent( d ) << "table[ " << i << " ]: ";
            dump( out, *data.refTable()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "table: none" << std::endl;
    }

        
    out << Indent( i ) << "}" << std::endl;
}


void dump( std::ostream& out, FrameCPP::Version_3_4_5::Vect& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Vect\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "compress: " << data.getCompress() << "\n"
        << Indent( d ) << "type: " << data.getType() << std::endl
        << Indent( d ) << "nData: " << data.getNData() << std::endl
        << Indent( d ) << "nBytes: " << data.getNBytes() << std::endl
        << Indent( d ) << "unitY: \"" << data.getUnitY() << "\"" << std::endl;

    out << Indent( d ) << "data: ";
#if __GNUC__ >= 3
    std::ios_base::fmtflags
#else
    long
#endif
 	flags = out.flags();
    for ( unsigned int n = 0; n < data.getNBytes(); n++ )
    {
        if ( n % DATA_WIDTH == 0 )
        {
            out << std::endl << Indent( d + INDENT_AMOUNT );
        }
        
        out << std::setfill( '0' ) << std::hex << std::setw( 2 )
            << int( ( (unsigned char*)(data.getData()) )[ n ] ) << std::flush;
    }
    out << std::endl;
    out.flags( flags );
    
    out << Indent( d ) << "nDim: " << data.getNDim() << std::endl;

    if ( data.getNDim() == 0 )
    {
        out << Indent( d ) << "nx: none" << std::endl;
        out << Indent( d ) << "dx: none" << std::endl;
        out << Indent( d ) << "startX: none" << std::endl;
        out << Indent( d ) << "unitX: none" << std::endl;
    }
    else
    {
        out << Indent( d ) << "nx: ";
        for ( unsigned int n = 0; n < data.getNDim(); n++ )
        {
            if ( n != 0 )
            {
                out << ", ";
            }
            out << data.getDimension( n ).getNx();
        }
        out << std::endl;
        
        out << Indent( d ) << "dx: ";
        for ( unsigned int n = 0; n < data.getNDim(); n++ )
        {
            if ( n != 0 )
            {
                out << ", ";
            }
            out << data.getDimension( n ).getDx();
        }
        out << std::endl;
        
        out << Indent( d ) << "startX: ";
        for ( unsigned int n = 0; n < data.getNDim(); n++ )
        {
            if ( n != 0 )
            {
                out << ", ";
            }
            out << data.getDimension( n ).getStartX();
        }
        out << std::endl;

        out << Indent( d ) << "unitX: ";
        for ( unsigned int n = 0; n < data.getNDim(); n++ )
        {
            if ( n != 0 )
            {
                out << ", ";
            }
            out << data.getDimension( n ).getUnitX();
        }
        out << std::endl;
    }
    
    out << Indent( i ) << "}" << std::endl;
}

void dump( std::ostream& out, FrameCPP::Version_3_4_5::Table& data, unsigned int i )
{
    int d = i + INDENT_AMOUNT;
    
    out << "Table\n"
        << Indent( i ) << "{\n"
        << Indent( d ) << "name: \"" << data.getName() << "\"\n"
        << Indent( d ) << "comment: " << data.getComment() << "\n"
        << Indent( d ) << "nColumn: " << data.getNColumns() << std::endl
        << Indent( d ) << "nRow: " << data.getNRows() << std::endl;

    if ( data.refColumns().size() != 0 )
    {
        out << std::endl;
        for ( unsigned int i = 0; i < data.refColumns().size(); ++i )
        {
            out << Indent( d ) << "aux[ " << i << " ]: ";
            dump( out, *data.refColumns()[ i ], d );
        }
    }
    else
    {
        out << Indent( d ) << "Columns: none" << std::endl;
    }

    out << Indent( i ) << "}" << std::endl;
}


