#ifndef ErrorsHH
#define ErrorsHH

#include <stdexcept>

//-----------------------------------------------------------------------------
//
//: Framecpp Error.
//
class framecpp_error : public std::runtime_error
{
public:

    framecpp_error( const std::string& msg, unsigned int code );
    unsigned int code() const throw();
    
private:

    unsigned int mCode;
};


inline framecpp_error::framecpp_error( const std::string& msg, unsigned int code )
        : runtime_error( msg ), mCode( code )
{
}

inline unsigned int framecpp_error::code() const throw()
{
    return mCode;
}

//-----------------------------------------------------------------------------
//
//: Write Failure.
//
class write_failure : public std::runtime_error
{
public:

    write_failure( const std::string& msg );
};


//-----------------------------------------------------------------------------
//
//: Read Failure.
//
class read_failure : public framecpp_error
{
public:

    read_failure( const std::string& msg, unsigned int code );
};


//-----------------------------------------------------------------------------
//
//: Not Found Error.
//
class not_found_error : public std::runtime_error
{
public:

    not_found_error( const std::string& msg );
};


//-----------------------------------------------------------------------------
//
//: Format Error.
//
class format_error : public std::runtime_error
{
public:

    format_error( const std::string& msg );
};


//-----------------------------------------------------------------------------
//
//: Frame Mismatch Error.
//
class frame_mismatch : public std::runtime_error
{
public:

    frame_mismatch( const std::string& msg );
};

//-----------------------------------------------------------------------------
//
//: DaqFrame Update Impossible Error.
//
class cannot_update : public std::runtime_error
{
public:

    cannot_update( const std::string& msg );
};


//-----------------------------------------------------------------------------
//
//: DaqFrame Concatenation Impossible Error.
//
class cannot_cat : public std::runtime_error
{
public:

    cannot_cat( const std::string& msg );
};

//-----------------------------------------------------------------------------
//
//: Unknown Error.
//
class unknown_error : public std::runtime_error
{
public:

    unknown_error( const std::string& msg );
};


#endif // ErrorsHH
