#include "framecpp/config.h"

#include "errors.hh"
#include "dictionary.hh"
#include "frameappendwritertoc.hh"

using namespace FrameCPP::Version_3_4_5;

/**
 * Read the TOC
 */   
//!exc: write_failure - Write failed.   
//!exc: read_failure - Read failed.
//            
FrameAppendWriterTOC::FrameAppendWriterTOC( std::iostream& out )
  : TOCReader( out ) // TOC is read
  , FrameWriterTOC( out ) // output initialized
{
  // impossible to append to older version files
  if (mFileHeader->getDictionary() != library.getCurrentVersionDictionary()) {
    throw read_failure("append doesn't support older data format versions", BADVERSION);
  }

  mClosed = false; // avoid file header writing by Output

  // note FrSHs that are present already to avoid duplicates
  for (TransMap::iterator iter = mTransMap.begin();
       iter != mTransMap.end(); iter++)
    shWrittenSet.insert(iter->second);
}


/**
 * Destructs the object
 */   
//   
//!exc: write_failure - Write failed.   
//               
FrameAppendWriterTOC::~FrameAppendWriterTOC()
{}

/**
 * Appends frame (overrired Output::writeFrame())
 */   
//   
//!exc: write_failure - Write failed.   
//               
void
FrameAppendWriterTOC::writeFrame( const Frame& Frame )
{
	try { 
		return appendFrame(Frame);
	} catch (read_failure e) {
		throw  write_failure(e.what());
	}
}

  // Get the TOC offset from the end of file structure
/**
 * Appends frame
 */
//!exc: write_failure - Write failed.   
//!exc: read_failure - Read failed.
//               
void
FrameAppendWriterTOC::appendFrame( const Frame& Frame )
{
  // Get the TOC offset from the end of file structure
  //
  mStream->seekg( -4, std::ios::end );
  INT_4U seekTOC = 0;
  *this >> seekTOC;

  // TOC is not present
  if ( !seekTOC ) 
    {
      throw read_failure( "TOC not present", TOCReader::NOTOC );
    }

  // Position output to the beginning of the TOC structure.
  // This leaves the FrSHs and FrSEs of the TOC between
  // the first and the second frames in the file.
  // This is OK according to the specification.

  mOut.seekp( - ((long) seekTOC), std::ios::end );

  this->FrameWriterTOC::writeFrame( Frame );
}

