#include "framecpp/config.h"

#include <memory>

#include "framecpp/Version3_4_5/daqframe.hh"
#include "framecpp/Version3_4_5/generalreader.hh"

using namespace FrameCPP::Version_3_4_5;

//-----------------------------------------------------------------------
//
//: Read a Frame
//
//!return: Frame* - Newly allocated.
//
//!exc: std::bad_alloc - Out of memory.
//!exc: read_failure - Read failed.
//
Frame* GeneralReader::
readFrame()
{
  std::auto_ptr<Frame> frame( FrameReaderBase::readFrame() );

  if ( frame.get( ) )
  {
    try
    {
      //-----------------------------------------------------------------
      // Try createing a daq frame
      //-----------------------------------------------------------------
      std::auto_ptr< DaqFrame >	daq_frame( new DaqFrame( *(frame.get( ) ) ) );

      // Move daq info to daq frame
      daq_frame->setDaqMaps( this->instanceOffsetMap,
			     this->classInstanceMultimap);
      daq_frame->setFileHeader(this->mFileHeader);
      return daq_frame.release( );
    }
    catch( ... )
    {
      // Not a daq frame. Continue with generic frame
    }

    // The following steps are here because they are not needed 
    // in the TOCReader. So they were transferred over here
    // from FrameReaderBase::readFrame()
    //

    // optional table of contents
    if ( readNextObjectHeader().getLocalClassId() == ID_TOC )
    {
      // Do TOC pseudo read (skip actually)
      dictionary->readElement( *this, ID_TOC );
    }
    
    ObjectHeader oh = readNextObjectHeader();
    if ( oh.getLocalClassId() == ID_ENDOFFILE )
    {
      readEndOfFile();
    }

  }
  return frame.release( );
}

ObjectHeader GeneralReader::
readNextObjectHeader()
{
  ObjectHeader hdr( FrameReaderBase::readNextObjectHeader() );
  INT_2U instance = hdr.getInstance();
  INT_2U classId =  hdr.getFileClassId();
  INT_2U localClassId =  hdr.getLocalClassId();

  if (classId != ID_SH && classId != ID_SE)
  {
    unsigned long cur_pos( mStream->tellg() );

    classInstanceMultimap.insert ( class_pair_type(localClassId, instance));
    adc_map_info_struct ami =
    { cur_pos,
      (localClassId == ID_ADCDATA ) ? adcDataVectorPtr(cur_pos) : 0,
      false
    };

    instanceOffsetMap.insert ( offset_pair_type (localClassId << 16 | instance,
					      ami ));
    mStream->seekg( cur_pos );
  }

  return hdr;
}

//!ignore_begin:
#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END

#define ADC_OFFSET_HEADER_END_DECL(offset) \
  INT_4U headerLength; \
  mStream->seekg (offset - 8); /* move to `header length' */ \
/*  mStream->read (&headerLength, 4); get the length */ \
  (*this) >> headerLength;
#define ADC_OFFSET_HEADER_END (offset - 8 + headerLength)
//!ignore_end:

//: Given the offset find the ADC data vector ptr.
INT_4U GeneralReader::
adcDataVectorPtr (int offset)
{
  ADC_OFFSET_HEADER_END_DECL(offset);
  mStream->seekg (ADC_OFFSET_HEADER_END - 12); // adjust to data ve
  INT_2U classId, instanceId;
  (*this) >> classId >> instanceId;
  return ID_VECT << 16 | instanceId;
}

#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END
