#ifndef ImageFrameWriterHH
#define ImageFrameWriterHH

// If it is a GNU compiler, support only 2.95.x and 3.2.x versions   
#if defined __GNUC__
   
#if 2 ==__GNUC__ && 95 ==__GNUC_MINOR__
   
#include <strstream>   
#define LDAS_USE_CGUC_2_95      
   
#else /* GCC 2.95 */

#include <sstream>
#define LDAS_USE_CGUC_3_2   
   
#endif /* GCC 2.95 */
   
#else    /* ! __GNUC__ */

// Use gcc-3.2.1 implementation
#include <sstream>
#define LDAS_USE_CGUC_3_2  
   
#endif  /* __GNUC__ */ 
   
   
#include "base.hh"
#include "framewritertoc.hh"

namespace FrameCPP {

  namespace Version_3_4_5 {
    class ImageFrameWriter;
  } // namespace - Version_3_4_5
} // namespace - FrameCPP

   
using namespace std;   

   
#ifdef LDAS_USE_CGUC_2_95
   
//: Creates frame image in user supplied output string stream `s'.
// Remembers offsets into `s.str()' for every structure instance
// written out by the means of inheriting TOC writer class and
// ultimately TOC class.
// Provides accessors and mutators for the frame file image.
//
class FrameCPP::Version_3_4_5::ImageFrameWriter : public FrameCPP::Version_3_4_5::FrameWriterTOC {
public:
  //!exc: write_failure - Write failed. 
  ImageFrameWriter (std::strstream &s)
    : FrameWriterTOC (s), ost(s), toc_offset(0)
    {}
  //!exc: write_failure - Write failed.    
  ~ImageFrameWriter() {}
private:
  //: frame file image created in this string stream.
  std::strstream &ost;
  //: offset to the beginnings of the End Of Frame structures.
  std::vector<INT_8U> EOFposition;
  //: offset to the beginning of the TOC structure.
  INT_8U toc_offset;

public:
  //: Vector instance offset map
  // Mapping works like this:
  // instanceId -> vector
  //
  // instanceId is ++ for every newly output Vect structure.
  // Vector has the element for each frame written.
  // Each vector element is an offset to the start of Vect structure
  // in the frame.
  // The offset is measured in `ost->str()'.
  
  std::map<INT_2U, std::vector<INT_4U> > vectInstanceOffsetMap;
  typedef std::map<INT_2U, std::vector<INT_4U> >::const_iterator OMI;

public:
  //: Get end of frame positions.
  inline const std::vector<INT_8U>& getPositionEOF() const { return EOFposition; }
  //: Get the table of contents offset.
  inline const INT_8U getTOCOffset() const { return toc_offset; }

  //: Overrides superclass. Called from the superclass.
  //!exc: write_failure - Write failed.    
  inline void startRecord (INT_2U classId, INT_4U length,
			   const FrameCPP::Version_3_4_5::Base* b)
  {
      INT_2U instance = getInstance(b);
      FrameWriterTOC::startRecord (classId, length, b); // call the superclass
      if (classId == ID_ENDOFFRAME)
        EOFposition.push_back(ost.pcount());
      else if (classId == ID_TOC)
        toc_offset = ost.pcount();
      else if (classId == ID_VECT) {
	vectInstanceOffsetMap[instance].push_back(ost.pcount());
      }
  }

  /* ADC structure access */
  //: Get ADC name string using the offset.
  inline char *adcName (int offset) {
    return name (offset);
  }
  //: Get ADC comment using offset.
  inline char *adcComment (int offset) {
    return name (offset + nameLength (offset) + 2);
  }

/* skip over the `name' and the `comment' strings */
#define ADC_NAME_COMMENT_OFFSET_DECL(offset) \
  int NL = offset + nameLength (offset) + 2
#define ADC_NAME_COMMENT_OFFSET (NL + nameLength (NL) + 2)

  //: Get ADC channel group number by offset.
  inline INT_4U adcChannelGroup (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET);
    INT_4U crate; ost.read ((char*)&crate, 4);
    return crate;
  }
  //: Get ADC channel number by offset.
  inline INT_4U adcChannelNumber (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 4);
    INT_4U cnum; ost.read ((char*)&cnum, 4);
    return cnum;
  }
  //: Get ADC NBits field by offset.
  inline INT_4U adcNBits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 8);
    INT_4U nbits; ost.read ((char*)&nbits, 4);
    return nbits;
  }
  //: Get ADC Bias by offset.
  inline REAL_4 adcBias (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 12);
    REAL_4 bias; ost.read ((char*)&bias, 4);
    return bias;
  }
  //: Get ADC Slope by offset.
  inline REAL_4 adcSlope (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 16);
    REAL_4 slope; ost.read ((char*)&slope, 4);
    return slope;
  }
  //: Get ADC Units string by offset.
  inline char *adcUnits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    return name (ADC_NAME_COMMENT_OFFSET + 20);
  }

#undef ADC_NAME_COMMENT_OFFSET_DECL
#undef ADC_NAME_COMMENT_OFFSET

/* These macros allow backwards movement from the ADC struct end */
#define ADC_OFFSET_HEADER_END_DECL(offset) \
  INT_4U headerLength; \
  ost.seekg (offset); /* move to `header length' */ \
  ost.read ((char*)&headerLength, 4); /* get the length */
#define ADC_OFFSET_HEADER_END (offset + headerLength)

  //: Get ADC sampling rate by offset.
  inline REAL_8 adcSampleRate (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 26);
    REAL_8 a; ost.read ((char*)&a, 8);
    return a;
  }
  //: Get ADC Time Offset seconds by offset.
  inline INT_4U adcTimeOffsetS (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 18);
    INT_4U a; ost.read ((char*)&a, 4);
    return a;
  }
  //: Get ADC Time offset nanoseconds by offset.
  inline INT_4U adcTimeOffsetN (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 14);
    INT_4U a; ost.read ((char*)&a, 4);
    return a;
  }
  //: Get ADC Frequency Shift by offset.
  inline REAL_8 adcFShift (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 10);
    REAL_8 fs; ost.read ((char*)&fs, 8);
    return fs;
  }
  //: Get ADC Data Valid flag by offset.
  inline INT_2U adcDataValid (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 2);
    INT_2U ovr; ost.read ((char*)&ovr, 2);
    return ovr;
  }
  //: Get pointer to the Data Valid flag.
  // Deferencing this pointer could produce SIGBUS.
  inline INT_2U* adcDataValidPtr (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    return (INT_2U*)(ost.str() + ADC_OFFSET_HEADER_END - 12 - 2);
  }

  //: Given the offset find the ADC data vector ptr.
  inline INT_4U adcDataVectorPtr (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    // std::cerr << "adc structure length is " << std::hex << headerLength << std::dec << std::endl;
    ost.seekg (ADC_OFFSET_HEADER_END - 12); // adjust to data ve
    INT_4U classInstance;
    ost.read ((char*)&classInstance, 4); // read the vec ptr in
    return classInstance;
  }

#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END

  /* Vector structure accessors */
  //: Given the offset to the start of the vector structure,
  //: move the input strstream position to the beginning of the `data'
  inline void seekVectorData (int offset) {
    ost.seekg (offset + nameLength (offset) + 2 + 12);
  }

  //: Get a pointer to the vector's data start by offset.
  inline char *vectorData (int offset) {
    return ost.str () + offset + nameLength (offset) + 2 + 12;
  }

  /* Frame header mutators */

  //: Set frame file attributes in all frames.
  inline void setFrameFileAttributes(INT_4S run, INT_4U frameNumber,
				     INT_4U dqual, INT_4U gps, INT_2U gpsInc,
				     INT_4U gpsn,
				     INT_2U leapS, INT_4S localTime)
  {
    INT_4U fn = frame.size();
    // This one loops over the frames in a file and updates them all.
    for (INT_4U i = 0; i < fn; i++) {
      ost.seekp (position[i] + 8 + nameLength (position[i] + 8) + 2);
      ost.write ((char*)&run, 4);
      INT_4U f = frameNumber + i;
      ost.write ((char*)&f, 4);
      ost.write ((char*)&dqual, 4);
      INT_4U gpsl = gpsInc*i + gps;
      ost.write ((char*)&gpsl, 4);
      ost.write ((char*)&gpsn, 4);
      ost.write ((char*)&leapS, 2);
      ost.write ((char*)&localTime, 4);

      ost.seekp (EOFposition[i]);
      ost.write ((char*)&run, 4);
      ost.write ((char*)&f, 4);
    }

    // Go to the TOC and update the values.
    ost.seekp (toc_offset);
    ost.write ((char*)&leapS, 2);
    ost.write ((char*)&localTime, 4);
    ost.seekp (toc_offset + 2 + 4 + 4);
    for (INT_4U i = 0; i < fn; i++) {
      INT_4U gpsl = gpsInc*i + gps;
      ost.write ((char*)&gpsl, 4);
    }
    for (INT_4U i = 0; i < fn; i++)
        ost.write ((char*)&gpsn, 4);
    ost.seekp (toc_offset + 2 + 4 + 4 + fn*(4 + 4 + 8));
    for (INT_4U i = 0; i < fn; i++)
    	ost.write ((char*)&run, 4);
    for (INT_4U i = 0; i < fn; i++) {
	INT_4U f = frameNumber + i;
    	ost.write ((char*)&f, 4);
    }
  }

  /*
   Both nameLength() and name() can be used in all
   structures that had `name' as the third element/
   The list of such strucutres include:
     FrSH, FrSE, Frame, FrADCData, FrDetector, FrHistory,
     FrRawData, FrProcData, FrSerData, FrStatData, FrVect,
     FrEvent, FrSummary.
  */

  inline int nameLength (int offset) {
    ost.seekg (offset);
    INT_2U nameLength;
    ost.read ((char*)&nameLength, 2);
    return nameLength;
  }

  inline char *name (int offset) {
    return ost.str () + offset + 2;
  }

};
   
   
#endif

   
   
#ifdef LDAS_USE_CGUC_3_2   
// NEW IMPLEMENTATION: using stringstream instead of strstream   
   
//: Creates frame image in user supplied output string stream `s'.
// Remembers offsets into `s.str()' for every structure instance
// written out by the means of inheriting TOC writer class and
// ultimately TOC class.
// Provides accessors and mutators for the frame file image.
//
class FrameCPP::Version_3_4_5::ImageFrameWriter : public FrameCPP::Version_3_4_5::FrameWriterTOC {
public:
  //!exc: write_failure - Write failed. 
  ImageFrameWriter (std::stringstream &s)
    : FrameWriterTOC (s), ost(s), toc_offset(0)
    {}
  //!exc: write_failure - Write failed.    
  ~ImageFrameWriter() {}
private:
  //: frame file image created in this string stream.
  std::stringstream &ost;
  //: offset to the beginnings of the End Of Frame structures.
  std::vector<INT_8U> EOFposition;
  //: offset to the beginning of the TOC structure.
  INT_8U toc_offset;

public:
  //: Vector instance offset map
  // Mapping works like this:
  // instanceId -> vector
  //
  // instanceId is ++ for every newly output Vect structure.
  // Vector has the element for each frame written.
  // Each vector element is an offset to the start of Vect structure
  // in the frame.
  // The offset is measured in `ost->str()'.
  
  std::map<INT_2U, std::vector<INT_4U> > vectInstanceOffsetMap;
  typedef std::map<INT_2U, std::vector<INT_4U> >::const_iterator OMI;

public:
  //: Get end of frame positions.
  inline const std::vector<INT_8U>& getPositionEOF() const { return EOFposition; }
  //: Get the table of contents offset.
  inline const INT_8U getTOCOffset() const { return toc_offset; }

  //: Overrides superclass. Called from the superclass.
  //!exc: write_failure - Write failed.    
  inline void startRecord (INT_2U classId, INT_4U length,
			   const FrameCPP::Version_3_4_5::Base* b)
  {
      INT_2U instance = getInstance(b);
      FrameWriterTOC::startRecord (classId, length, b); // call the superclass
      if (classId == ID_ENDOFFRAME)
        EOFposition.push_back(ost.str().size());
      else if (classId == ID_TOC)
        toc_offset = ost.str().size();
      else if (classId == ID_VECT) {
	vectInstanceOffsetMap[instance].push_back(ost.str().size());
      }
  }

  /* ADC structure access */
  //: Get ADC name string using the offset.
  inline const char *adcName (int offset) {
    return name (offset);
  }
  //: Get ADC comment using offset.
  inline const char *adcComment (int offset) {
    return name (offset + nameLength (offset) + 2);
  }

/* skip over the `name' and the `comment' strings */
#define ADC_NAME_COMMENT_OFFSET_DECL(offset) \
  int NL = offset + nameLength (offset) + 2
#define ADC_NAME_COMMENT_OFFSET (NL + nameLength (NL) + 2)

  //: Get ADC channel group number by offset.
  inline INT_4U adcChannelGroup (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET);
    INT_4U crate; ost.read ((char*)&crate, 4);
    return crate;
  }
  //: Get ADC channel number by offset.
  inline INT_4U adcChannelNumber (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 4);
    INT_4U cnum; ost.read ((char*)&cnum, 4);
    return cnum;
  }
  //: Get ADC NBits field by offset.
  inline INT_4U adcNBits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 8);
    INT_4U nbits; ost.read ((char*)&nbits, 4);
    return nbits;
  }
  //: Get ADC Bias by offset.
  inline REAL_4 adcBias (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 12);
    REAL_4 bias; ost.read ((char*)&bias, 4);
    return bias;
  }
  //: Get ADC Slope by offset.
  inline REAL_4 adcSlope (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    ost.seekg (ADC_NAME_COMMENT_OFFSET + 16);
    REAL_4 slope; ost.read ((char*)&slope, 4);
    return slope;
  }
  //: Get ADC Units string by offset.
  inline const char *adcUnits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    return name (ADC_NAME_COMMENT_OFFSET + 20);
  }

#undef ADC_NAME_COMMENT_OFFSET_DECL
#undef ADC_NAME_COMMENT_OFFSET

/* These macros allow backwards movement from the ADC struct end */
#define ADC_OFFSET_HEADER_END_DECL(offset) \
  INT_4U headerLength; \
  ost.seekg (offset); /* move to `header length' */ \
  ost.read ((char*)&headerLength, 4); /* get the length */
#define ADC_OFFSET_HEADER_END (offset + headerLength)

  //: Get ADC sampling rate by offset.
  inline REAL_8 adcSampleRate (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 26);
    REAL_8 a; ost.read ((char*)&a, 8);
    return a;
  }
  //: Get ADC Time Offset seconds by offset.
  inline INT_4U adcTimeOffsetS (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 18);
    INT_4U a; ost.read ((char*)&a, 4);
    return a;
  }
  //: Get ADC Time offset nanoseconds by offset.
  inline INT_4U adcTimeOffsetN (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 14);
    INT_4U a; ost.read ((char*)&a, 4);
    return a;
  }
  //: Get ADC Frequency Shift by offset.
  inline REAL_8 adcFShift (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 10);
    REAL_8 fs; ost.read ((char*)&fs, 8);
    return fs;
  }
  //: Get ADC Data Valid flag by offset.
  inline INT_2U adcDataValid (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    ost.seekg (ADC_OFFSET_HEADER_END - 12 - 2);
    INT_2U ovr; ost.read ((char*)&ovr, 2);
    return ovr;
  }
  //: Get pointer to the Data Valid flag.
  // Deferencing this pointer could produce SIGBUS.
  inline INT_2U* adcDataValidPtr (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    return (INT_2U*)(ost.str().c_str() + ADC_OFFSET_HEADER_END - 12 - 2);
  }

  //: Given the offset find the ADC data vector ptr.
  inline INT_4U adcDataVectorPtr (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    // std::cerr << "adc structure length is " << std::hex << headerLength << std::dec << std::endl;
    ost.seekg (ADC_OFFSET_HEADER_END - 12); // adjust to data ve
    INT_4U classInstance;
    ost.read ((char*)&classInstance, 4); // read the vec ptr in
    return classInstance;
  }

#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END

  /* Vector structure accessors */
  //: Given the offset to the start of the vector structure,
  //: move the input strstream position to the beginning of the `data'
  inline void seekVectorData (int offset) {
    ost.seekg (offset + nameLength (offset) + 2 + 12);
  }

  //: Get a pointer to the vector's data start by offset.
  inline const char *vectorData (int offset) {
    return ost.str().c_str() + offset + nameLength (offset) + 2 + 12;
  }

  /* Frame header mutators */

  //: Set frame file attributes in all frames.
  inline void setFrameFileAttributes(INT_4S run, INT_4U frameNumber,
				     INT_4U dqual, INT_4U gps, INT_2U gpsInc,
				     INT_4U gpsn,
				     INT_2U leapS, INT_4S localTime)
  {
    INT_4U fn = frame.size();
    // This one loops over the frames in a file and updates them all.
    for (INT_4U i = 0; i < fn; i++) {
      ost.seekp (position[i] + 8 + nameLength (position[i] + 8) + 2);
      ost.write ((char*)&run, 4);
      INT_4U f = frameNumber + i;
      ost.write ((char*)&f, 4);
      ost.write ((char*)&dqual, 4);
      INT_4U gpsl = gpsInc*i + gps;
      ost.write ((char*)&gpsl, 4);
      ost.write ((char*)&gpsn, 4);
      ost.write ((char*)&leapS, 2);
      ost.write ((char*)&localTime, 4);

      ost.seekp (EOFposition[i]);
      ost.write ((char*)&run, 4);
      ost.write ((char*)&f, 4);
    }

    // Go to the TOC and update the values.
    ost.seekp (toc_offset);
    ost.write ((char*)&leapS, 2);
    ost.write ((char*)&localTime, 4);
    ost.seekp (toc_offset + 2 + 4 + 4);
    for (INT_4U i = 0; i < fn; i++) {
      INT_4U gpsl = gpsInc*i + gps;
      ost.write ((char*)&gpsl, 4);
    }
    for (INT_4U i = 0; i < fn; i++)
        ost.write ((char*)&gpsn, 4);
    ost.seekp (toc_offset + 2 + 4 + 4 + fn*(4 + 4 + 8));
    for (INT_4U i = 0; i < fn; i++)
    	ost.write ((char*)&run, 4);
    for (INT_4U i = 0; i < fn; i++) {
	INT_4U f = frameNumber + i;
    	ost.write ((char*)&f, 4);
    }
  }

  /*
   Both nameLength() and name() can be used in all
   structures that had `name' as the third element/
   The list of such strucutres include:
     FrSH, FrSE, Frame, FrADCData, FrDetector, FrHistory,
     FrRawData, FrProcData, FrSerData, FrStatData, FrVect,
     FrEvent, FrSummary.
  */

  inline int nameLength (int offset) {
    ost.seekg (offset);
    INT_2U nameLength;
    ost.read ((char*)&nameLength, 2);
    return nameLength;
  }

  inline const char *name (int offset) {
    return ost.str().c_str() + offset + 2;
  }

};   

#endif   
   
#endif   
