#ifndef FrameCppRefHeadHH
#define FrameCppRefHeadHH


#include "refheadbase.hh"
#include <typeinfo>


namespace FrameCPP {

namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: A container reference.
//
// This class is used by the FrameReaderBase class to store a reference to a
// container to fill.
//
//!todo: Is this class really necessary?  It doesn't do much.
//
template< class T >
class RefHead : public RefHeadBase
{
public:
    
    RefHead( Container< T >& c );
    RefHead( const RefHead& ref );

    virtual ~RefHead();

    //!exc: std::bad_cast 
    virtual void addItem( Base* b );

    virtual const void* getContainer( ) const;
private:

    //: Referenced container.
    Container< T >& mContainer;
};


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: Container< T >& c - Referenced container.
//
template< class T >
inline RefHead< T >::RefHead( Container< T >& c )
        : mContainer( c )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const RefHead& ref - Object to be copied.
//
template< class T >
inline RefHead< T >::RefHead( const RefHead& ref )
        : mContainer( ref.mContainer )
{
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
template< class T >
RefHead< T >::~RefHead()
{
}


//-----------------------------------------------------------------------------
//
//: Adds an item to the referenced container.
//
//!param: Base* b - Item to be added.
//
//!exc: std::bad_cast    
//   
template< class T >
void RefHead< T >::addItem( Base* b )
{
    T* p = dynamic_cast< T* >( b );
    mContainer.append( p, false );
}

 
template< class T >
const void* RefHead< T >::getContainer( ) const
{
  return &mContainer;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppRefHeadHH
