#include "framecpp/config.h"

#include "sh.hh"
#include "input.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::SH;


//-----------------------------------------------------------------------------
//
//: Input constructor.
//
//!param: Input& in - Input data stream object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Data read failure.
//
SH::SH( Input& in )
        : mName(), mClass(), mComment(), mSE()
{
    in >> mName >> mClass >> mComment;
}

    
//-----------------------------------------------------------------------------
//
//: Destructor.
//
//!exc: None.   
//      
SH::~SH()
{}



//-----------------------------------------------------------------------------
//
//: Get the described structure name.
//
//!return: const std::string& -- Described structure name.
//
//!exc: None.   
//      
const std::string& SH::getName() const
{
    return mName;
}


   
//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const SH& sh - Object to be assigned.
//
//!return: const SH& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SH& SH::operator=( const SH& sh )
{
    if ( this != &sh )
    {
        mName = sh.mName;
        mClass = sh.mClass;
        mComment = sh.mComment;
        mSE = sh.mSE;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SH objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const SH& sh - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool SH::operator==( const SH& sh ) const
{
    return ( ( mName == sh.mName ) && ( mClass == sh.mClass )
             && ( mSE == sh.mSE ) );
}



//-----------------------------------------------------------------------------
//
//: Equal comparison (all data except class ID number).
//
// Determines whether two SH objects are equal. 
//
//!param: const SH& sh - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool SH::equal( const SH& sh ) const
{
    return ( ( mName == sh.mName ) && ( mSE == sh.mSE ) );
}


//-----------------------------------------------------------------------------
//
//: Read the SH from an Input object.
//
//!param: Input& in - Input data stream object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
SH* SH::read( Input& in )
{
    return new SH( in );
}


//-----------------------------------------------------------------------------
//
//: Writes the SH to an Output object.
//
//!param: Output& out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void SH::write( Output& out ) const
{
    out.startRecord( ID_SH, 16 + mName.size() + mComment.size(), this );
    out << mName << mClass << mComment;
    out.endRecord( this );
}

   

//-----------------------------------------------------------------------------
//
//: Get the class ID
//
//!return: INT_2U -- SH structure class ID.
//
//!exc: None.   
//      
INT_2U SH::getClassId() const
{
    return ID_SH;
}


   
