#include "framecpp/config.h"

#include "framecpp/Version3_4_5/adcdata.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrAdcData.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

using std::ceil;
using std::floor;

namespace {
  class MapAdcDataRef
    : public IFrameStream::MapRef< FrameCPP::Version_3_4_5::AdcData, FrAdcData >
  {
  public:
    MapAdcDataRef( IFrameStream& Stream,
		   const FrameCPP::Version_3_4_5::AdcData* AdcData5,
		   FrAdcData& AdcData6 );

  protected:
   
    // virtual
    void operator()( const void* Ref5Ptr,
                     const PTR_STRUCT& PtrStruct );
  }; // class - MapAdcDataRef
   
   
  MapAdcDataRef::
  MapAdcDataRef( IFrameStream& Stream,
		 const FrameCPP::Version_3_4_5::AdcData* AdcData5,
		 FrAdcData& AdcData6 )
    : IFrameStream::MapRef< FrameCPP::Version_3_4_5::AdcData,
			    FrAdcData >( Stream, AdcData5, AdcData6 )
  {
  }

  void MapAdcDataRef::operator()( const void* Ref5Ptr,
				  const PTR_STRUCT& PtrStruct )
  {
    if ( Ref5Ptr == &(m_struct_5->refData( ) ) )
    {
      m_stream.AddContainer( m_struct_6.RefData( ), PtrStruct );
    }
    else if ( Ref5Ptr == &(m_struct_5->refAux( ) ) )
    {
      m_stream.AddContainer( m_struct_6.RefAux( ), PtrStruct );
    }
    
    return;
  }   
} // namespace - anonymous


const CHAR* FrAdcData::getStaticName()
{
   static const CHAR* class_name( "FrAdcData" );
   return class_name;
}


//-----------------------------------------------------------------------
// FrAdcData::data_definition_type
//-----------------------------------------------------------------------
FrAdcData::data_definition_type::
data_definition_type( )
{
}

FrAdcData::data_definition_type::
data_definition_type( const data_definition_type& Source )
  : data( Source.data )
{
  copy_core( Source );
}

bool FrAdcData::data_definition_type::
operator==( const data_definition_type& RHS ) const
{
#if LM_INFO
#define LM_INFO_CHECK(a) \
  if ( ! ( a == RHS.a ) ) \
  { \
    std::cerr << "DEBUG: Comparison of " << #a << " failed" << std::endl; \
  }

  LM_INFO_CHECK( name );
  LM_INFO_CHECK( comment);
  LM_INFO_CHECK( channelGroup );
  LM_INFO_CHECK( channelNumber );
  LM_INFO_CHECK( nBits );
  LM_INFO_CHECK( bias );
  LM_INFO_CHECK( slope );
  LM_INFO_CHECK( units );
  LM_INFO_CHECK( fShift );
  LM_INFO_CHECK( timeOffset );
  LM_INFO_CHECK( dataValid );
  LM_INFO_CHECK( phase );
  LM_INFO_CHECK( data );
  LM_INFO_CHECK( aux );
#undef LM_INFO_CHECK
#endif /* LM_INFO */

  const bool	retval =
    ( ( this == &RHS ) ||	// Short circuit for comparison to self
      ( ( name == RHS.name ) &&
	( comment == RHS.comment ) &&
	( channelGroup == RHS.channelGroup ) &&
	( channelNumber == RHS.channelNumber ) &&
	( nBits == RHS.nBits ) &&
	( sampleRate == RHS.sampleRate ) &&
	( bias == RHS.bias ) &&
	( slope == RHS.slope ) &&
	( units == RHS.units ) &&
	( fShift == RHS.fShift ) &&
	( timeOffset == RHS.timeOffset ) &&
	( dataValid == RHS.dataValid ) &&
	( phase == RHS.phase ) &&
	( data == RHS.data ) &&
	( aux == RHS.aux ) ) );

  return retval;

}

void FrAdcData::data_definition_type::
copy_core( const data_definition_type& Source )
{
  name = Source.name;
  comment = Source.comment;
  channelGroup = Source.channelGroup;
  channelNumber = Source.channelNumber;
  nBits = Source.nBits;
  sampleRate = Source.sampleRate;
  bias = Source.bias;
  slope = Source.slope;
  units = Source.units;
  fShift = Source.fShift;
  timeOffset = Source.timeOffset;
  dataValid = Source.dataValid;
  phase = Source.phase;
  aux = Source.aux;
}

//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------

FrAdcData::
FrAdcData( )
  : FrBase( CLASS_FR_ADC_DATA )
{
}

FrAdcData::
FrAdcData( const FrAdcData& Source )
  : FrBase( CLASS_FR_ADC_DATA ),
    m_data( Source.m_data )
{
}

FrAdcData::
FrAdcData( const std::string& name, INT_4U group,
	   INT_4U channel, INT_4U nBits,
	   REAL_8 sampleRate,
	   REAL_4 bias,
	   REAL_4 slope,
	   const std::string& units,
	   REAL_8 fShift,
	   REAL_8 timeOffset,
	   INT_2U dataValid,
	   REAL_4 phase )
  : FrBase( CLASS_FR_ADC_DATA )
{
  m_data.name = name;
  m_data.channelGroup = group;
  m_data.channelNumber = channel;
  m_data.nBits = nBits;
  m_data.sampleRate = sampleRate;
  m_data.bias = bias;
  m_data.slope = slope;
  m_data.units = units;
  m_data.fShift = fShift;
  m_data.timeOffset = timeOffset;
  m_data.dataValid = dataValid;
  m_data.phase = phase;
}

#if ENABLE_DEPRICATED_FRAMECPP
FrAdcData::
FrAdcData( const std::string& name, INT_4U group,
	   INT_4U channel, INT_4U nBits,
	   const REAL_8& sampleRate,
	   REAL_4 bias,
	   REAL_4 slope,
	   const std::string& units,
	   const REAL_8& fShift,
	   const GPSTime& timeOffset,
	   INT_2U dataValid,
	   const REAL_4& phase )
  : FrBase( CLASS_FR_ADC_DATA )
{
  m_data.name = name;
  m_data.channelGroup = group;
  m_data.channelNumber = channel;
  m_data.nBits = nBits;
  m_data.sampleRate = sampleRate;
  m_data.bias = bias;
  m_data.slope = slope;
  m_data.units = units;
  m_data.fShift = fShift;
  m_data.timeOffset = timeOffset - GPSTime( 0, 0 );
  m_data.dataValid = dataValid;
  m_data.phase = phase;
}


const std::string& FrAdcData::getName( ) const
{
   return GetName( );
}

#endif // ENABLE_DEPRICATED_FRAMECPP

void FrAdcData::AppendComment( const std::string& comment )
{
   return FrameCPP::Version_6::AppendComment( m_data.comment, comment );
} 


FrAdcData::
~FrAdcData( )
{
}


const std::string& FrAdcData::GetName() const
{
   return m_data.name;
}


INT_8U FrAdcData::
Bytes( ) const
{
  return
    FrBase::Bytes( ) +
    m_data.name.Bytes( ) +
    m_data.comment.Bytes( ) +
    sizeof( m_data.channelGroup ) +
    sizeof( m_data.channelNumber ) +
    sizeof( m_data.nBits ) +
    sizeof( m_data.bias ) +
    sizeof( m_data.slope ) +
    m_data.units.Bytes( ) +
    sizeof( m_data.sampleRate ) +
    sizeof( m_data.timeOffset ) +
    sizeof( m_data.fShift ) +
    sizeof( m_data.phase ) +
    sizeof( m_data.dataValid ) +
    PTR_STRUCT::Bytes( ) +	// aux
    PTR_STRUCT::Bytes( ) +	// data
    PTR_STRUCT::Bytes( )	// next
    ;
}

FrAdcData& FrAdcData::
Merge( const FrAdcData& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrAdcData* FrAdcData::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrAdcData > ret( new FrAdcData );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

std::auto_ptr< FrAdcData > FrAdcData::
SubFrAdcData( REAL_8 Offset, REAL_8 Dt ) const
{
  //---------------------------------------------------------------------
  // Sanity checks
  //---------------------------------------------------------------------
  // Verify the existance of only a single FrVect component
  //---------------------------------------------------------------------
  if ( m_data.data.size( ) != 1 )
  {
    std::ostringstream	msg;

    msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	<< "data vectors of length " << m_data.data.size( )
	<< " currently are not supported";
    throw std::logic_error( msg.str( ) );
  }
  INT_4U	num_elements = m_data.data[0]->GetDim( 0 ).GetNx( );
  //---------------------------------------------------------------------
  //:TODO: Ensure the dimension of the FrVect is 1
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  //:TODO: Ensure the requested range is valid
  //---------------------------------------------------------------------
  if ( Offset > ( m_data.timeOffset +
		  ( num_elements / m_data.sampleRate ) ) )
  {
    std::ostringstream msg;
    msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	<< "The value for the parameter Offset (" << Offset
	<< ") is out of range ([0-"
	<< ( m_data.timeOffset + ( num_elements / m_data.sampleRate ) )
	<< "))";
    throw std::range_error( msg.str( ) );
  }
  //---------------------------------------------------------------------
  // Start moving of data
  //---------------------------------------------------------------------
  std::auto_ptr< FrAdcData >	retval( new FrAdcData );

  // Copy in the header information
  retval->m_data.copy_core( this->m_data );

  //:TODO: Get the slice of FrVect
  REAL_8	startSampleReal =
    ( Offset * retval->m_data.sampleRate ) -
    ( retval->m_data.timeOffset * retval->m_data.sampleRate );

  INT_4U	endSample =
    INT_4U( ceil ( startSampleReal +
		   ( retval->m_data.sampleRate * Dt ) ) );
  INT_4U	startSample = ( startSampleReal > REAL_8( 0 ) )
    ? INT_4U( floor( startSampleReal ) )
    : INT_4U( 0 );

  retval->m_data.data.append( this->m_data.data[ 0 ]->
			      SubFrVect( startSample, endSample ).release( ),
			      false, /* allocate */
			      true /* own */ );
  //---------------------------------------------------------------------
  // Shift the offset
  //---------------------------------------------------------------------
  if ( startSampleReal > 0 )
  {
    // There is no gap
    retval->m_data.timeOffset = REAL_8( 0 );
  }
  else
  {
    retval->m_data.timeOffset += Offset;
  }
  //---------------------------------------------------------------------
  // Return the new FrAdcData
  //---------------------------------------------------------------------
  return retval;
}
   
void FrAdcData::
Write( OFrameStream& Stream ) const
{
  try
  {
    Stream << m_data.name
	   << m_data.comment
	   << m_data.channelGroup
	   << m_data.channelNumber
	   << m_data.nBits
	   << m_data.bias
	   << m_data.slope
	   << m_data.units
	   << m_data.sampleRate
	   << m_data.timeOffset
	   << m_data.fShift
	   << m_data.phase
	   << m_data.dataValid
	   << m_data.data
	   << m_data.aux
	   << GetNext( )
      ;
  }
  catch ( const std::exception& Exception )
  {
    std::cerr << "DEBUG: Error Writing FrAdcData: " << m_data.name
	      << " exception: " << Exception.what( )
	      << std::endl;
    throw;
  }
}

void FrAdcData::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_ADC_DATA,
				"AdcData  Data Structure" ) );

  Stream.WriteDictionary( FrSE( "name", "STRING") );
  Stream.WriteDictionary( FrSE( "comment", "STRING") );
  Stream.WriteDictionary( FrSE( "channelGroup", "INT_4U") );
  Stream.WriteDictionary( FrSE( "channelNumber", "INT_4U") );
  Stream.WriteDictionary( FrSE( "nBits", "INT_4U") );
  Stream.WriteDictionary( FrSE( "bias", "REAL_4") );
  Stream.WriteDictionary( FrSE( "slope", "REAL_4") );
  Stream.WriteDictionary( FrSE( "units", "STRING") );
  Stream.WriteDictionary( FrSE( "sampleRate", "REAL_8") );
  Stream.WriteDictionary( FrSE( "timeOffset", "REAL_8") );
  Stream.WriteDictionary( FrSE( "fShift", "REAL_8") );
  Stream.WriteDictionary( FrSE( "phase", "REAL_4") );
  Stream.WriteDictionary( FrSE( "dataValid", "INT_2U") );

  Stream.WriteDictionary( FrSE( "data",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "aux",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );

  Stream.WriteDictionary( FrSE( "next",	PTR_STRUCT::Desc( getStaticName() ) ) );
}

void FrAdcData::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< AdcData >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      Stream >> m_data.name
	     >> m_data.comment
	     >> m_data.channelGroup
	     >> m_data.channelNumber
	     >> m_data.nBits
	     >> m_data.bias
	     >> m_data.slope
	     >> m_data.units
	     >> m_data.sampleRate
	     >> m_data.timeOffset
	     >> m_data.fShift
	     >> m_data.phase
	     >> m_data.dataValid
	     >> m_data.data
	     >> m_data.aux
	     >> next
	;
    }
    break;
  case 5:
    ver5.reset( AdcData::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( AdcData::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( AdcData::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->getName( );
    m_data.comment = ver5->getComment( );
    m_data.channelGroup = ver5->getChannelGroup( );
    m_data.channelNumber = ver5->getChannelNumber( );
    m_data.nBits = ver5->getNBits( );
    m_data.bias = ver5->getBias( );
    m_data.slope = ver5->getSlope( );
    m_data.units = ver5->getUnits( );
    m_data.sampleRate = ver5->getSampleRate( );
    m_data.timeOffset =
      GPSTime( ver5->getTimeOffset( ).getSec( ),
	       ver5->getTimeOffset( ).getNSec( ) ) -
      GPSTime( 0, 0 );
    m_data.fShift = ver5->getFShift( );
    m_data.phase = ver5->getPhase( );
    m_data.dataValid = ver5->getDataValid( );

    MapAdcDataRef	map_refs( Stream, ver5.get( ), *this );
    map_refs.Map();
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
