#ifndef FrameCPP_VERSION_6_FrDetector_HH
#define FrameCPP_VERSION_6_FrDetector_HH

#include <string>
#include <vector>

#include "general/types.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/Dimension.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"
#if ENABLE_DEPRICATED_FRAMECPP
#include "framecpp/Version6/FrStatData.hh"
#endif

namespace FrameCPP
{
  namespace Version_6
  {
    class FrDetector: public FrBase
    {
    public:
      typedef SearchContainer< FrVect, &FrVect::GetName >	aux_type;
      typedef SearchContainer< FrTable, &FrTable::GetName >	table_type;

      typedef aux_type::iterator aux_iterator;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef table_type::iterator table_iterator;
      typedef table_type::const_iterator const_table_iterator;

      static const CHAR* getStaticName();

      //: Default constructor
      FrDetector( );

      FrDetector( const std::string& Name,
		  const char* Prefix,
		  const REAL_8 Longitude,
		  const REAL_8 Latitude,
		  const REAL_4 Elevation,
		  const REAL_4 ArmXazimuth,
		  const REAL_4 ArmYazimuth,
		  const REAL_4 ArmXaltitude,
		  const REAL_4 ArmYaltitude,
		  const REAL_4 ArmXmidpoint,
		  const REAL_4 ArmYmidpoint,
		  const INT_4S LocalTime );

      FrDetector( const FrDetector& detector );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      const std::string& GetName( ) const;
      const CHAR* GetPrefix( ) const { return m_data.prefix; }
      REAL_8 GetLongitude( ) const { return m_data.longitude; }
      REAL_8 GetLatitude( ) const { return m_data.latitude; }
      REAL_4 GetElevation( ) const { return m_data.elevation; }
      REAL_4 GetArmXazimuth( ) const { return m_data.armXazimuth; }
      REAL_4 GetArmYazimuth( ) const { return m_data.armYazimuth; }
      REAL_4 GetArmXaltitude( ) const { return m_data.armXaltitude; }
      REAL_4 GetArmYaltitude( ) const { return m_data.armYaltitude; }
      REAL_4 GetArmXmidpoint( ) const { return m_data.armXmidpoint; }
      REAL_4 GetArmYmidpoint( ) const { return m_data.armYmidpoint; }
      INT_4S GetLocalTime( ) const { return m_data.localTime; }
      const aux_type& RefAux( ) const { return m_data.aux; }
      aux_type& RefAux( ) { return m_data.aux; }
      const table_type& RefTable( ) const { return m_data.table; }
      table_type& RefTable( ) { return m_data.table; }   
   
      //: Merge with another FrDetector
      FrDetector& Merge( const FrDetector& RHS );

      //: Read the structure from stream
      static FrDetector* ReadDynamic( IFrameStream& Stream,
				      const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct {
	STRING	name;
	CHAR	prefix[2];
	REAL_8	longitude;
	REAL_8	latitude;
	REAL_4	elevation;
	REAL_4	armXazimuth;
	REAL_4	armYazimuth;
	REAL_4	armXaltitude;
	REAL_4	armYaltitude;
	REAL_4	armXmidpoint;
	REAL_4	armYmidpoint;
	INT_4S	localTime;
	aux_type	aux;
	table_type	table;	
      } m_data;

      void read( IFrameStream& Stream );
   
#if ENABLE_DEPRICATED_FRAMECPP
    public:
      typedef SearchContainer< FrStatData, &FrStatData::GetName >	StatDataContainer;
      typedef aux_type	MoreContainer;
      typedef table_type	MoreTableContainer;

      typedef aux_iterator more_iterator;
      typedef const_aux_iterator const_more_iterator;
      typedef table_iterator moreTable_iterator;
      typedef const_table_iterator const_moreTable_iterator;
      typedef StatDataContainer::iterator statData_iterator;
      typedef StatDataContainer::const_iterator const_statData_iterator;

      FrDetector( const std::string& name,
		  REAL_8 longitude, REAL_8 latitude, REAL_4 elevation,
		  REAL_4 armXazimuth, REAL_4 armYazimuth,
		  REAL_4 armXaltitude, REAL_4 armYaltitude,
		  REAL_4 armXmidpoint, REAL_4 armYmidpoint,
		  INT_4S localTime, INT_4U dataQuality,
		  const std::string& qaBitList );

      const std::string& getName() const;

      REAL_8 getLongitude() const
      {
	return GetLongitude( );
      }
      REAL_8 getLatitude() const
      {
	return GetLatitude( );
      }
      REAL_4 getElevation() const
      {
	return GetElevation( );
      }
      REAL_4 getArmXAzimuth() const
      {
	return GetArmXazimuth( );
      }
      REAL_4 getArmYAzimuth() const
      {
	return GetArmYazimuth( );
      }
      REAL_4 getArmXAltitude() const
      {
	return GetArmXaltitude( );
      }
      REAL_4 getArmYAltitude() const
      {
	return GetArmYaltitude( );
      }
      REAL_4 getArmXMidpoint() const
      {
	return GetArmXmidpoint( );
      }
      REAL_4 getArmYMidpoint() const
      {
	return GetArmYmidpoint( );
      }
      INT_4S getLocalTime() const
      {
	return GetLocalTime( );
      }
      INT_4U getDataQuality() const
      {
	return 0;
      }
      const std::string& getQaBitList() const
      {
	return m_qa_bit_list;
      }
      MoreContainer& refMore()
      {
	return RefAux( );
      }
      const MoreContainer& refMore() const
      {
	return RefAux( );
      }
      MoreTableContainer& refMoreTable()
      {
	return RefTable( );
      }
      const MoreTableContainer& refMoreTable() const
      {
	return RefTable( );
      }
      StatDataContainer& refStatData()
      {
	return stat_data;
      }
      const StatDataContainer& refStatData() const
      {
	return stat_data;
      }

    private:
      StatDataContainer	stat_data;
      std::string m_qa_bit_list;
#endif
    };

  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrDetector_HH */
