#include "framecpp/config.h"

#include "framecpp/Version3_4_5/event.hh"

#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrEvent.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

#define LM_DEBUG 0

namespace {
  class MapEventRef
    : public IFrameStream::MapRef< FrameCPP::Version_3_4_5::Event, FrEvent >
  {
  public:
    MapEventRef( IFrameStream& Stream,
		 const FrameCPP::Version_3_4_5::Event* Event5,
		 FrEvent& Event6 )
      : IFrameStream::MapRef< FrameCPP::Version_3_4_5::Event, FrEvent >( Stream, Event5, Event6 )
    {
    }

  protected:
   
    // virtual
    void operator()( const void* Ref5Ptr, const PTR_STRUCT& PtrStruct );
   
  }; // class - MapEventRef
   
   
  void MapEventRef::operator()( const void* Ref5Ptr,
   		                const PTR_STRUCT& PtrStruct )
  {
      if ( Ref5Ptr == &(m_struct_5->refData( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefData( ), PtrStruct );
      }
      else if ( Ref5Ptr == &(m_struct_5->refTable( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefTable( ), PtrStruct );
      }
   
      return;
  }   
} // namepace - anonymous

const CHAR* FrEvent::getStaticName()
{
   static const CHAR* class_name( "FrEvent" );
   return class_name;
}


FrEvent::
FrEvent( )
  : FrBase( CLASS_FR_EVENT )
{
}

FrEvent::
FrEvent( const FrEvent& eventData )
  : FrBase( CLASS_FR_EVENT )
{
  m_data.name = eventData.m_data.name;
  m_data.comment = eventData.m_data.comment;
  m_data.inputs = eventData.m_data.inputs;
  m_data.GTime = eventData.m_data.GTime;
  m_data.timeBefore = eventData.m_data.timeBefore;
  m_data.timeAfter = eventData.m_data.timeAfter;
  m_data.eventStatus = eventData.m_data.eventStatus;
  m_data.amplitude = eventData.m_data.amplitude;
  m_data.probability = eventData.m_data.probability;
  m_data.statistics = eventData.m_data.statistics;
  m_data.Params = eventData.m_data.Params;
  m_data.data = eventData.m_data.data;
  m_data.table = eventData.m_data.table;
}

FrEvent::
FrEvent( const std::string& name, const std::string& comment,
	 const std::string& inputs,
	 const GPSTime& time,
	 REAL_4 timeBefore, REAL_4 timeAfter,
	 INT_4U eventStatus, REAL_4 amplitude,
	 REAL_4 prob,
	 const std::string& statistics,
	 const ParamList_type& parameters )
  : FrBase( CLASS_FR_EVENT )
{
  m_data.name = name;
  m_data.comment = comment;
  m_data.inputs = inputs;
  m_data.GTime = time;
  m_data.timeBefore = timeBefore;
  m_data.timeAfter = timeAfter;
  m_data.eventStatus = eventStatus;
  m_data.amplitude = amplitude;
  m_data.probability = prob;
  m_data.statistics = statistics;
  m_data.Params = parameters;
}


const std::string& FrEvent::GetName() const
{
   return m_data.name;
}


#if ENABLE_DEPRICATED_FRAMECPP
FrEvent::
FrEvent( const std::string& name, const std::string& comment,
	 const std::string& inputs,
	 const GPSTime& time,
	 REAL_4 timeBefore, REAL_4 timeAfter,
	 INT_4U eventStatus, REAL_4 amplitude,
	 REAL_4 prob,
	 const std::string& statistics,
	 const ParamList5_type& parameters )
  : FrBase( CLASS_FR_EVENT )
{
  m_data.name = name;
  m_data.comment = comment;
  m_data.inputs = inputs;
  m_data.GTime = time;
  m_data.timeBefore = timeBefore;
  m_data.timeAfter = timeAfter;
  m_data.eventStatus = eventStatus;
  m_data.amplitude = amplitude;
  m_data.probability = prob;
  m_data.statistics = statistics;
  for ( ParamList5_type::const_iterator i( parameters.begin( ) );
	i != parameters.end( );
	i++ )
  {
    m_data.Params.push_back( Param_type( (*i).first, (*i).second) );
  }
}


const std::string& FrEvent::getName() const
{
   return GetName();
}
#endif

INT_8U FrEvent::
Bytes( ) const
{
  INT_8U	ret( FrBase::Bytes( ) );

  ret +=
    m_data.name.Bytes( ) +
    m_data.comment.Bytes( ) +
    m_data.inputs.Bytes( ) +
    FrameCPP::Version_6::Bytes( m_data.GTime ) +
    sizeof( m_data.timeBefore ) +
    sizeof( m_data.timeAfter ) +
    sizeof( m_data.eventStatus ) +
    sizeof( m_data.amplitude ) +
    sizeof( m_data.probability ) +
    m_data.statistics.Bytes( ) +
    sizeof( INT_2U );	// nParam
  for ( ParamList_type::const_iterator i ( m_data.Params.begin( ) );
	i != ( m_data.Params.end( ) );
	i++ )
  {
    ret += sizeof( Param_type::second_type ) +
      (*i).first.Bytes( );
  }

  ret +=
    PTR_STRUCT::Bytes( ) +	// data
    PTR_STRUCT::Bytes( ) +	// table
    PTR_STRUCT::Bytes( )	// next
    ;
  return ret;
}

FrEvent& FrEvent::
Merge( const FrEvent& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrEvent* FrEvent::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrEvent > ret( new FrEvent );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrEvent::
Write( OFrameStream& Stream ) const
{
#if LM_DEBUG
  INT_8U head( Stream.tellp( ) );
  Stream << m_data.name;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.comment;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.inputs;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.GTime;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.timeBefore;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.timeAfter;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.eventStatus;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.amplitude;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.probability;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.statistics;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.Params;
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.data.containerCast( );
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << m_data.table.containerCast( );
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
  head = Stream.tellp( );
  Stream << GetNext( );
  std::cerr << "INFO: pos: " << ( Stream.tellp( ) - head ) << std::endl;
#else
  Stream << m_data.name
	 << m_data.comment
	 << m_data.inputs
	 << m_data.GTime
	 << m_data.timeBefore
	 << m_data.timeAfter
	 << m_data.eventStatus
	 << m_data.amplitude
	 << m_data.probability
	 << m_data.statistics
	 << m_data.Params
	 << m_data.data.containerCast( )
	 << m_data.table.containerCast( )
	 << GetNext( )
    ;
#endif
}

void FrEvent::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_EVENT,
				"Event Data Structure" ) );

  Stream.WriteDictionary( FrSE( "name", "STRING" ) );
  Stream.WriteDictionary( FrSE( "comment", "STRING" ) );
  Stream.WriteDictionary( FrSE( "inputs", "STRING" ) );
  Stream.WriteDictionary( FrSE( "GTimeS", "INT_4U" ) );
  Stream.WriteDictionary( FrSE( "GTimeN", "INT_4U" ) );
  Stream.WriteDictionary( FrSE( "timeBefore", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "timeAfter", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "eventStatus", "INT_4U" ) );
  Stream.WriteDictionary( FrSE( "amplitude", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "probability", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "statistics", "STRING" ) );
  Stream.WriteDictionary( FrSE( "nParam", "INT_2U" ) );
  Stream.WriteDictionary( FrSE( "parameters", "REAL_4[nParam]" ) );
  Stream.WriteDictionary( FrSE( "parameterNames", "STRING[nParam]" ) );

  Stream.WriteDictionary( FrSE( "data",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "table",
				PTR_STRUCT::Desc( FrTable::getStaticName() ) ) );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );
}

void FrEvent::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< Event >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      INT_2U	nParam;

      Stream >> m_data.name
	     >> m_data.comment
	     >> m_data.inputs
	     >> m_data.GTime
	     >> m_data.timeBefore
	     >> m_data.timeAfter
	     >> m_data.eventStatus
	     >> m_data.amplitude
	     >> m_data.probability
	     >> m_data.statistics
	     >> nParam
	;
      Stream.ReadVector( m_data.Params, nParam );

      Stream >> m_data.data.containerCast( )
	     >> m_data.table.containerCast( )
	     >> next
	;
    }
    break;
  case 5:
    ver5.reset( Event::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( Event::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( Event::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->getName( );
    m_data.comment = ver5->getComment( );
    m_data.comment = ver5->getInputs( );
    m_data.GTime =
      General::GPSTime( ver5->getGTime( ).getSec( ),
			ver5->getGTime( ).getNSec( ) );
    m_data.timeBefore = ver5->getTimeBefore( );
    m_data.timeBefore = ver5->getTimeAfter( );
    m_data.eventStatus = ver5->getEventStatus( );
    m_data.amplitude = ver5->getAmplitude( );
    m_data.probability = ver5->getProbability( );
    m_data.statistics = ver5->getStatistics( );

    m_data.Params.resize( 0 );
    for( std::vector< std::pair<std::string, REAL_4> >::const_iterator
	   p( ver5->getParameters( ).begin( ) );
	 p != ver5->getParameters( ).end( );
	 p++ )
    {
      m_data.Params.push_back( *p );
    }

    MapEventRef	map_refs( Stream, ver5.get( ), *this );
    map_refs.Map();
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
