#ifndef FrameCPP_VERSION_6_FrHeader_HH
#define	FrameCPP_VERSION_6_FrHeader_HH

#include <vector>

#include "general/types.hh"

#include "framecpp/Version6/FrBase.hh"

namespace FrameCPP
{
  namespace Version_6
  {

    class IFrameStream;
    class OFrameStream;

    class FrHeader
    {
    public:
      //: Retrieve the buffer holding the endian bytes for an 8 byte quantity
      const unsigned char* const Get8ByteEndianness( ) const;

      //: Return the byte order of the frame
      int GetByteOrder( ) const;

      //: Return the version number
      INT_2U GetLibraryRevision( ) const;

      //: Return the version number
      INT_2U GetVersion( ) const;

      //: Return the originator of the frame.
      const char* GetOriginator( ) const;
      
      //: Establish the originator of the frame.
      void SetOriginator( const std::string& Originator );
      
      //: Reorder the bytes
      template< class T >
      void Reorder( std::vector< T >& Data ) const;

      //: Reorder the bytes
      template< class T >
      void Reorder( T& Data ) const;

      //: Read header from stream
      void read( IFrameStream& fs );

      //: Write Header to stream
      void write( OFrameStream& fs );

    private:
      friend class IFrameStream;

      union FrHeader_type {
	struct
	{
	  char igwd[ 5 ];
	  unsigned char version;
	  unsigned char minorVersion;
	  unsigned char sizeInt2;
	  unsigned char sizeInt4;
	  unsigned char sizeInt8;
	  unsigned char sizeReal4;
	  unsigned char sizeReal8;
      
	  unsigned char x1234[ 2 ];
	  unsigned char x12345678[ 4 ];
	  unsigned char x123456789abcdef[ 8 ];
        
	  unsigned char smallPi[ 4 ];
	  unsigned char bigPi[ 8 ];
        
	  unsigned char AZ[ 2 ];
	} raw; // struct
      } m_data; // union - FrHeader_type
      bool m_reorder_bytes;
      int  m_byte_order;
    }; // class - FrHeader

    inline const unsigned char* const FrHeader::
    Get8ByteEndianness( ) const
    {
      return m_data.raw.x123456789abcdef;
    }

    inline int FrHeader::
    GetByteOrder( ) const
    {
      return m_byte_order;
    }

    inline INT_2U FrHeader::
    GetLibraryRevision( ) const
    {
      return m_data.raw.minorVersion;
    }

    inline INT_2U FrHeader::
    GetVersion( ) const
    {
      return m_data.raw.version;
    }
  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_FrHeader_HH */
