#include "framecpp/config.h"

#include "framecpp/Version3_4_5/history.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrHistory.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;


const CHAR* FrHistory::getStaticName()
{
   static const CHAR* class_name( "FrHistory" );
   return class_name;
}


FrHistory::
FrHistory( )
  : FrBase( CLASS_FR_HISTORY )
{
}

FrHistory::
FrHistory( const FrHistory& Source )
  : FrBase( CLASS_FR_HISTORY )
{
  m_data.name = Source.m_data.name;
  m_data.time = Source.m_data.time;
  m_data.comment = Source.m_data.comment;
}

FrHistory::
FrHistory( const std::string& Name,
	   INT_4U Time,
	   const std::string& Comment )
  : FrBase( CLASS_FR_HISTORY )
{
  m_data.name = Name;
  m_data.time = Time;
  m_data.comment = Comment;
}


std::string FrHistory::
ErrorInfo( ) const
{
  std::ostringstream	result;

  result << LibraryString << "::FrHistory: {Name: " << m_data.name
	 << " Time: " << m_data.time
	 << "} "
    ;
  return result.str( );
}

const std::string& FrHistory::GetName() const
{
   return m_data.name;   
}


#if ENABLE_DEPRICATED_FRAMECPP
const std::string& FrHistory::getName( ) const
{
   return GetName();
}
#endif

INT_8U FrHistory::
Bytes( ) const
{
  return FrBase::Bytes( ) +
    m_data.name.Bytes( ) +
    sizeof( INT_4U ) + //	time
    m_data.comment.Bytes( ) +
    PTR_STRUCT::Bytes( )
    ;
}

FrHistory* FrHistory::
Create( const FrameCPP::Version_3_4_5::History& Source )
{
  std::auto_ptr< FrHistory >	ret( new FrHistory );

  ret->m_data.name = Source.getName( );
  ret->m_data.time = Source.getTime( );
  ret->m_data.comment = Source.getComment( );
  return ret.release( );
}

FrHistory& FrHistory::
Merge( const FrHistory& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrHistory* FrHistory::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrHistory > ret( new FrHistory );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrHistory::
Write( OFrameStream& Stream ) const
{
  try {
    Stream << m_data.name
	   << m_data.time
	   << m_data.comment
	   << GetNext( )
      ;
  }
  catch( const std::exception& Exception )
  {
    std::cerr << "DEBUG: Error Writing FrHistory: " << m_data.name
	      << " exception: " << Exception.what( )
	      << std::endl;
    throw;
  }
}

void FrHistory::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_HISTORY,
				"History Structure" ) );
  Stream.WriteDictionary( FrSE( "name", "STRING", "") );
  Stream.WriteDictionary( FrSE( "time", "INT_4U", "") );
  Stream.WriteDictionary( FrSE( "comment", "STRING", "") );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );
 
  return;
}

void FrHistory::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< History >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {

      Stream >> m_data.name
	     >> m_data.time
	     >> m_data.comment
	     >> next
	;
    }
    break;
  case 5:
  case 4:
  case 3:
    ver5.reset( History::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->mName;
    m_data.time = ver5->mTime;
    m_data.comment = ver5->mComment;
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
