#ifndef FrameCPP_VERSION_6_FrHistory_HH
#define FrameCPP_VERSION_6_FrHistory_HH

#include "framecpp/FrameCPP.hh"

#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_3_4_5
  {
    class History;
  }

  namespace Version_6
  {
    class FrHistory: public FrBase
    {
    public:
      static const CHAR* getStaticName();

      //: Default constructor
      FrHistory( );

      FrHistory( const FrHistory& Source );

      FrHistory( const std::string& Name,
		 INT_4U Time,
		 const std::string& Comment );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      static FrHistory*
      Create( const ::FrameCPP::Version_3_4_5::History& Source );

      virtual std::string ErrorInfo( ) const;

      // Accessors
      const std::string& GetName() const;

      INT_4U GetTime( ) const
      {
         return m_data.time;
      }   
      const std::string& GetComment( ) const
      {
 	 return m_data.comment;
      }   

      //: Merge with another FrAdcData
      FrHistory& Merge( const FrHistory& RHS );

      //: Read the structure from stream
      static FrHistory* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct {
	STRING	name;
	INT_4U	time;
	STRING	comment;
      } m_data;

      void read( IFrameStream& Stream );

#if ENABLE_DEPRICATED_FRAMECPP
    public:
      const std::string& getName() const;

      INT_4U getTime( ) const
      {
	return GetTime( );
      }
      const std::string& getComment( ) const
      {
	return GetComment( );
      }
#endif	/* ENABLE_DEPRICATED_FRAMECPP */
    };

  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrHistory_HH */
