#include "framecpp/config.h"

#include <memory>

#include "framecpp/Version3_4_5/msg.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrMsg.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;


const CHAR* FrMsg::getStaticName()
{
   static const CHAR* class_name( "FrMsg" );
   return class_name;
}


FrMsg::
FrMsg( )
  : FrBase( CLASS_FR_MSG )
{
}

FrMsg::
FrMsg( const FrMsg& Source )
  : FrBase( CLASS_FR_MSG )
{
  m_data.alarm = Source.m_data.alarm;
  m_data.message = Source.m_data.message;
  m_data.severity = Source.m_data.severity;
  m_data.GTime = Source.m_data.GTime;
}

FrMsg::
FrMsg( const std::string& Alarm,
       const std::string& Message,
       INT_4U Severity,
       const GPSTime& GTime )
  : FrBase( CLASS_FR_MSG )
{
  m_data.alarm = Alarm;
  m_data.message = Message;
  m_data.severity = Severity;
  m_data.GTime = GTime;
}


const std::string& FrMsg::GetAlarm() const
{
   return m_data.alarm;
}


INT_8U FrMsg::
Bytes( ) const
{
  return FrBase::Bytes( ) +
    m_data.alarm.Bytes( ) +
    m_data.message.Bytes( ) +
    sizeof( m_data.severity ) +
    FrameCPP::Version_6::Bytes( m_data.GTime ) +
    PTR_STRUCT::Bytes( )	// next
    ;
}

FrMsg& FrMsg::
Merge( const FrMsg& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrMsg* FrMsg::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrMsg > ret( new FrMsg );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrMsg::
Write( OFrameStream& Stream ) const
{
  Stream << m_data.alarm
	 << m_data.message
	 << m_data.severity
	 << m_data.GTime
	 << GetNext( )
    ;
}

void FrMsg::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_MSG,
				"Msg Data Structure" ) );

  Stream.WriteDictionary( FrSE( "alarm", "STRING") );
  Stream.WriteDictionary( FrSE( "message", "STRING") );
  Stream.WriteDictionary( FrSE( "severity", "INT_4U") );
  Stream.WriteDictionary( FrSE( "GTimeS", "INT_4U") );
  Stream.WriteDictionary( FrSE( "GTimeN", "INT_4U") );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );

  return;
}

void FrMsg::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< Msg >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      Stream >> m_data.alarm
	     >> m_data.message
	     >> m_data.severity
	     >> m_data.GTime
	     >> next
	;
    }
    break;
  case 5:
  case 4:
    ver5.reset( Msg::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( Msg::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.alarm = ver5->getAlarm( );
    m_data.message = ver5->getMessage( );
    m_data.severity = ver5->getSeverity( );
    m_data.GTime = GPSTime( ver5->getGTime( ).getSec( ),
			    ver5->getGTime( ).getNSec( ) );
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
