#ifndef FrameCPP_VERSION_6_FrMsg_HH
#define FrameCPP_VERSION_6_FrMsg_HH

#include <vector>

#include "general/types.hh"

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrMsg: public FrBase
    {
    public:

      static const CHAR* getStaticName();

      //: Default constructor
      FrMsg( );

      FrMsg( const FrMsg& Source );

      FrMsg( const std::string& Alarm, const std::string& Message,
	     INT_4U Severity,
	     const GPSTime& GTime );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      //: Accessors
      const std::string& GetAlarm() const;
      const std::string& GetMessage( ) const { return m_data.message; }
      const INT_4U GetSeverity( ) const { return m_data.severity; }
      const GPSTime& GetGTime( ) const { return m_data.GTime; }   
   
      //: Merge with another FrAdcData
      FrMsg& Merge( const FrMsg& RHS );

      //: Read the structure from stream
      static FrMsg* ReadDynamic( IFrameStream& Stream,
				      const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct {
	STRING	alarm;
	STRING	message;
	INT_4U	severity;
	GPSTime	GTime;
      } m_data;

      void read( IFrameStream& Stream );
#if ENABLE_DEPRICATED_FRAMECPP
    public:
      inline const std::string& getAlarm( ) const
      {
	return GetAlarm( );
      }

      inline const std::string& getMessage( ) const
      {
	return GetMessage( );
      }
      
      inline const INT_4U getSeverity( ) const
      {
	return GetSeverity( );
      }

      const GPSTime& getGTime( ) const
      {
	return GetGTime( );
      }
#endif /* ENABLE_DEPRICATED_FRAMECPP */
    };


  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrMsg_HH */
