#include "framecpp/config.h"

#include "general/mutexlock.hh"
#include "general/Singleton.hh"
#include "general/gpstime.hh"

#include "framecpp/Version3_4_5/procdata.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrProcData.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

using std::ceil;
using std::floor;

namespace {
  class IDTypeToStringMapping
  {
    SINGLETON_TS_DECL( IDTypeToStringMapping );
  public:
    const std::string& GetString( INT_2U Id ) const;

  private:
    std::vector< std::string >	m_types;
  };

  class IDSubTypeToStringMappingFrequency
  {
    SINGLETON_TS_DECL( IDSubTypeToStringMappingFrequency );
  public:
    const std::string& GetString( INT_2U Id ) const;

  private:
    std::vector< std::string >	m_types;
  };

  class MapProcDataRef
    : public IFrameStream::MapRef< FrameCPP::Version_3_4_5::ProcData, FrProcData >
  {
  public:
    MapProcDataRef( IFrameStream& Stream,
		 const FrameCPP::Version_3_4_5::ProcData* ProcData5,
		 FrProcData& ProcData6 )
      : IFrameStream::MapRef< FrameCPP::Version_3_4_5::ProcData, FrProcData >( Stream, ProcData5, ProcData6 )
    {
    }

  protected:
   
    // virtual
    void operator()( const void* Ref5Ptr,
		     const PTR_STRUCT& PtrStruct );
   
  }; // class - MapProcDataRef
   
   
  void MapProcDataRef::operator()( const void* Ref5Ptr, 
                                   const PTR_STRUCT& PtrStruct )
  {
      if ( Ref5Ptr == &(m_struct_5->refData( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefData( ), PtrStruct );
      }
      else if ( Ref5Ptr == &(m_struct_5->refAux( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefAux( ), PtrStruct );
      }
      else if ( Ref5Ptr == &(m_struct_5->refTable( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefTable( ), PtrStruct );
      }
   
      return;
  }
   
} // namespace - anonymous


namespace FrameCPP
{
  namespace Version_6
  {
    template<  >
    IFrameStream& IFrameStream::
    ReadVector( FrProcData::AuxParamList_type& Data, INT_4U Size )
    {
      std::vector< REAL_8 >	auxParam;
      std::vector< STRING > auxParamNames;

      Data.resize( 0 );

      this->ReadVector( auxParam, Size );
      this->ReadVector( auxParamNames, Size );

      std::vector< FrProcData::AuxParam_type::auxParam_type >::const_iterator
	api( auxParam.begin( ) );
      std::vector< FrProcData::AuxParam_type::auxParamName_type >::const_iterator
	apni( auxParamNames.begin( ) );

      while ( api != auxParam.end( ) )
      {
	Data.push_back( FrProcData::AuxParam_type( *api, *apni ) );
#if old
	Data.back( ).auxParam = *api;
	Data.back( ).auxParamName = *apni;
#endif /* old */
	api++;
	apni++;
      }
      return *this;
    }

    template<  >
    OFrameStream& OFrameStream::
    operator<<( const FrProcData::AuxParamList_type& Data )
    {
      INT_2U	nAuxParam( Data.size( ) );

      std::vector< REAL_8 >	auxParam;
      std::vector< STRING > auxParamNames;

      for ( FrProcData::AuxParamList_type::const_iterator i ( Data.begin( ) );
	    i != Data.end( );
	    i++ )
      {
	auxParam.push_back( (*i).auxParam );
	auxParamNames.push_back( (*i).auxParamName );
      }

      *this << nAuxParam
	    << auxParam
	    << auxParamNames;

      return *this;
    }
  } // namespace - Version_6
} // namespace - FrameCPP


const CHAR* FrProcData::getStaticName()
{
   static const CHAR* class_name( "FrProcData" );
   return class_name;
}


FrProcData::data_definition_type::
data_definition_type( )
{
}

FrProcData::data_definition_type::
data_definition_type( const data_definition_type& Source )
{
  copy_core( Source );
  data = Source.data;
}

void FrProcData::data_definition_type::
copy_core( const data_definition_type& Source )
{
  name = Source.name;
  comment = Source.comment;
  type = Source.type;
  subType = Source.subType;
  timeOffset = Source.timeOffset;
  tRange = Source.tRange;
  fShift = Source.fShift;
  phase = Source.phase;
  fRange = Source.fRange;
  BW = Source.BW;
  AuxParam = Source.AuxParam;
  aux = Source.aux;
  table = Source.table;
  history = Source.history;
}

FrProcData::
FrProcData( )
  : FrBase( CLASS_FR_PROC_DATA ),
    m_synced_with_vector( false )
{
}

FrProcData::
FrProcData( const FrProcData& ProcData )
  : FrBase( CLASS_FR_PROC_DATA ),
    m_data( ProcData.m_data ),
    m_synced_with_vector( false )
{
}

FrProcData::
FrProcData( const std::string& Name,
	    const std::string& Comment,
	    INT_2U Type,
	    INT_2U SubType,
	    REAL_8 TimeOffset,
	    REAL_8 TRange,
	    REAL_8 FShift,
	    REAL_4 Phase,
	    REAL_8 FRange,
	    REAL_8 BW )
  : FrBase( CLASS_FR_PROC_DATA ),
    m_synced_with_vector( false )
{
  m_data.name = Name;
  m_data.comment = Comment;
  m_data.type = Type;
  m_data.subType = SubType;
  m_data.timeOffset = TimeOffset;
  m_data.tRange = TRange;
  m_data.fShift = FShift;
  m_data.phase = Phase;
  m_data.fRange = FRange;
  m_data.BW = BW;
}

FrProcData::
FrProcData( const std::string& name,
	    REAL_8 sampleRate, const GPSTime& timeOffset,
	    REAL_8 fShift, REAL_4 phase )
  : FrBase( CLASS_FR_PROC_DATA ),
    m_synced_with_vector( false )

{
  m_data.name = name;
  m_data.comment = std::string( "" );
  m_data.type = 0;
  m_data.subType = 0;
  m_data.timeOffset = timeOffset - GPSTime( 0, 0 );
  m_data.tRange = 0;
  m_data.fShift = fShift;
  m_data.phase = phase;
  m_data.fRange = 0;
  m_data.BW = 0;
}


const std::string& FrProcData::GetName() const
{
   return m_data.name;
}


void FrProcData::AppendComment( const std::string& comment )
{
   return FrameCPP::Version_6::AppendComment( m_data.comment, comment );
} 


#if ENABLE_DEPRICATED_FRAMECPP
const std::string& FrProcData::getName() const
{
   return GetName( );
}
#endif


INT_8U FrProcData::
Bytes( ) const
{
  INT_8U	ret( FrBase::Bytes( ) );

  ret +=
    m_data.name.Bytes( ) +
    m_data.comment.Bytes( ) +
    sizeof( m_data.type ) +
    sizeof( m_data.subType ) +
    sizeof( m_data.timeOffset ) +
    sizeof( m_data.tRange ) +
    sizeof( m_data.fShift ) +
    sizeof( m_data.phase ) +
    sizeof( m_data.fRange ) +
    sizeof( m_data.BW ) +
    sizeof( INT_2U );	// nAuxParam
  for ( AuxParamList_type::const_iterator i ( m_data.AuxParam.begin( ) );
	i != ( m_data.AuxParam.end( ) );
	i++ )
  {
    ret += sizeof( AuxParam_type::auxParam_type ) +
      (*i).auxParamName.Bytes( );
  }

  ret +=
    PTR_STRUCT::Bytes( ) +	// aux
    PTR_STRUCT::Bytes( ) +	// data
    PTR_STRUCT::Bytes( ) +	// table
    PTR_STRUCT::Bytes( ) +	// history
    PTR_STRUCT::Bytes( )	// next
    ;
  return ret;
}

const std::string& FrProcData::
IDTypeToString( INT_2U Type )
{
  return IDTypeToStringMapping::Instance( ).GetString( Type );
}

const std::string& FrProcData::
IDSubTypeToString( INT_2U Type, INT_2U SubType )
{
  static const std::string	empty( "" );
  switch ( Type )
  {
  case FREQUENCY_SERIES:
  return IDSubTypeToStringMappingFrequency::Instance( ).GetString( Type );
    break;
  default:
    return empty;
    break;
  }
}

FrProcData& FrProcData::
Merge( const FrProcData& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrProcData* FrProcData::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrProcData > ret( new FrProcData );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

std::auto_ptr< FrProcData > FrProcData::
SubFrProcData( REAL_8 Offset, REAL_8 Dt ) const
{
  std::auto_ptr< FrProcData >	retval;

  //---------------------------------------------------------------------
  // Process the types that are understood
  //---------------------------------------------------------------------
  switch( GetType( ) )
  {
  case TIME_SERIES:
    retval = sub_time_series( Offset, Dt );
    break;
  default:
    {
      //-----------------------------------------------------------------
      // This type is not supported so throw an exception
      //-----------------------------------------------------------------
      std::ostringstream	msg;
      msg << "Currently cannot take a subsection of type: "
	  << IDTypeToString( GetType( ) );
      const std::string& subtype( IDSubTypeToString( GetType( ),
						     GetSubType( ) ) );
      if ( subtype.length( ) > 0 )
      {
	msg << " subType: " << subtype;
      }
      throw std::runtime_error( msg.str( ) );
    }
    break;
  }

  return retval;
}

void FrProcData::
Write( OFrameStream& Stream ) const
{
  try
  {
    Stream << m_data.name
	   << m_data.comment
	   << m_data.type
	   << m_data.subType
	   << m_data.timeOffset
	   << m_data.tRange
	   << m_data.fShift
	   << m_data.phase
	   << m_data.fRange
	   << m_data.BW
	   << m_data.AuxParam
	   << m_data.data.containerCast( )
	   << m_data.aux.containerCast( )
	   << m_data.table.containerCast( )
	   << m_data.history.containerCast( )
	   << GetNext( )
      ;
  }
  catch( const std::exception& Exception )
  {
    std::cerr << "DEBUG: Error Writing FrProcData: " << m_data.name
	      << " exception: " << Exception.what( )
	      << std::endl;
    throw;
  }
}

void FrProcData::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_PROC_DATA,
				"Post-processed Data Structure" ) );

  Stream.WriteDictionary( FrSE( "name", "STRING" ) );
  Stream.WriteDictionary( FrSE( "comment", "STRING" ) );
  Stream.WriteDictionary( FrSE( "type", "INT_2U" ) );
  Stream.WriteDictionary( FrSE( "subType", "INT_2U" ) );
  Stream.WriteDictionary( FrSE( "timeOffset", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "tRange", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "fShift", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "phase", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "fRange", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "BW", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "nAuxParam", "INT_2U" ) );
  Stream.WriteDictionary( FrSE( "auxParam", "REAL_8[nAuxParam]" ) );
  Stream.WriteDictionary( FrSE( "auxParamNames", "STRING[nAuxParam]" ) );

  Stream.WriteDictionary( FrSE( "data",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "aux",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "table",
				PTR_STRUCT::Desc( FrTable::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "history",
				PTR_STRUCT::Desc( FrHistory::getStaticName() ) ) );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );

  return;
}

void FrProcData::
sync_with_vector( ) const
{
  if ( ( m_synced_with_vector == false ) &&
       ( RefData( ).size( ) > 0 ) )
  {
    switch( GetType( ) )
    {
    case TIME_SERIES:
      m_data.tRange = ( RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) *
			RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
      break;
    case FREQUENCY_SERIES:
      m_data.fRange = ( RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) *
			RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
      break;
    }
    m_synced_with_vector = true;
  }
}

void FrProcData::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< ProcData >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      INT_2U	nAuxParam;

      Stream >> m_data.name
	     >> m_data.comment
	     >> m_data.type
	     >> m_data.subType
	     >> m_data.timeOffset
	     >> m_data.tRange
	     >> m_data.fShift
	     >> m_data.phase
	     >> m_data.fRange
	     >> m_data.BW
	     >> nAuxParam
	;
      Stream.ReadVector( m_data.AuxParam, nAuxParam );

      Stream >> m_data.data.containerCast( )
	     >> m_data.aux.containerCast( )
	     >> m_data.table.containerCast( )
	     >> m_data.history.containerCast( )
	     >> next
	;
    }
    break;
  case 5:
    ver5.reset( ProcData::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( ProcData::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( ProcData::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->getName( );
    m_data.comment = ver5->getComment( );
    if ( ver5->getSampleRate( ) == -1 )
    {
      m_data.type = FREQUENCY_SERIES;
    }
    else
    {
      m_data.type = TIME_SERIES;
    }
    m_data.subType = 0;
    m_data.timeOffset =
      General::GPSTime( ver5->getTimeOffset( ).getSec( ),
			ver5->getTimeOffset( ).getNSec( ) ) -
			General::GPSTime( 0, 0 );
    m_data.tRange = 0; //:TODO: Calculate after linking
    m_data.fShift = ver5->getFShift( );
    m_data.phase = ver5->getPhase( );
    m_data.fRange = 0; //:TODO: Calculate after linking
    m_data.BW = 0;

    MapProcDataRef	map_refs( Stream, ver5.get( ), *this );
    map_refs.Map();
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}

// Routine to take a subsection when the data type is TimeSeries data
std::auto_ptr< FrProcData > FrProcData::
sub_time_series( REAL_8 Offset, REAL_8 Dt ) const
{
  //---------------------------------------------------------------------
  // Sanity checks
  //---------------------------------------------------------------------
  // Verify the existance of only a single FrVect component
  //---------------------------------------------------------------------
  if ( m_data.data.size( ) != 1 )
  {
    std::ostringstream	msg;

    msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	<< "data vectors of length " << m_data.data.size( )
	<< " currently are not supported";
    throw std::logic_error( msg.str( ) );
  }
  INT_4U	num_elements = m_data.data[0]->GetDim( 0 ).GetNx( );
  //---------------------------------------------------------------------
  //:TODO: Ensure the dimension of the FrVect is 1
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  //:TODO: Ensure the requested range is valid
  //---------------------------------------------------------------------
  REAL_8	sample_rate = 1.0 / m_data.data[ 0 ]->GetDim( 0 ).GetDx( );
  if ( Offset > ( m_data.timeOffset +
		  ( num_elements / sample_rate ) ) )
  {
    std::ostringstream msg;
    msg << "METHOD: frameAPI::FrAdcData::SubFrAdcData: "
	<< "The value for the parameter Offset (" << Offset
	<< ") is out of range ([0-"
	<< ( m_data.timeOffset + ( num_elements / sample_rate ) )
	<< "))";
    throw std::range_error( msg.str( ) );
  }
  //---------------------------------------------------------------------
  // Start moving of data
  //---------------------------------------------------------------------
  std::auto_ptr< FrProcData >	retval( new FrProcData );
  // Copy in the header information
  retval->m_data.copy_core( this->m_data );

  //:TODO: Get the slice of FrVect
  REAL_8	startSampleReal =
    ( Offset * sample_rate ) -
    ( retval->m_data.timeOffset * sample_rate );

  INT_4U	endSample =
    INT_4U( ceil ( startSampleReal +
		   ( Dt * sample_rate ) ) );
  INT_4U	startSample = ( startSampleReal > REAL_8( 0 ) )
    ? INT_4U( floor( startSampleReal ) )
    : INT_4U( 0 );

  retval->m_data.data.append( this->m_data.data[ 0 ]->SubFrVect( startSample, endSample ).release( ),
			      false, /* allocate */
			      true /* own */ );
  //---------------------------------------------------------------------
  // Shift the offset
  //---------------------------------------------------------------------
  if ( startSampleReal > 0 )
  {
    // There is no gap
    retval->m_data.timeOffset = REAL_8( 0 );
  }
  else
  {
    retval->m_data.timeOffset += Offset;
  }
  //---------------------------------------------------------------------
  // :TODO: Recalculate tRange
  //---------------------------------------------------------------------
  retval->SetTRange( retval->RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) * 
		     retval->RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
  //---------------------------------------------------------------------
  // Return the new FrAdcData
  //---------------------------------------------------------------------
  return retval;
}

namespace
{
  //---------------------------------------------------------------------
  // Initialization of singleton
  //---------------------------------------------------------------------
  SINGLETON_TS_INST( IDTypeToStringMapping );
  //---------------------------------------------------------------------
  // Default constructor
  //---------------------------------------------------------------------
  IDTypeToStringMapping::
  IDTypeToStringMapping( )
  {
    m_types.push_back( "Unknown/user defined" );
    m_types.push_back( "Time Series" );
    m_types.push_back( "Frequency Series" );
    m_types.push_back( "Other 1D Series data" );
    m_types.push_back( "Time-Frequency" );
    m_types.push_back( "Wavelets" );
    m_types.push_back( "Multi-dimensional" );
  }

  const std::string& IDTypeToStringMapping::
  GetString( INT_2U Type ) const
  {
    if ( Type >= m_types.size( ) )
    {
      std::ostringstream	msg;
      msg << "Cannot map FrProcData type value of " << Type
	  << " to string value because it is out of range";
      throw std::range_error( msg.str( ) );
    }
    return m_types[ Type ];
  }

  //---------------------------------------------------------------------
  // Initialization of singleton
  //---------------------------------------------------------------------
  SINGLETON_TS_INST( IDSubTypeToStringMappingFrequency );
  //---------------------------------------------------------------------
  // Default constructor
  //---------------------------------------------------------------------
  IDSubTypeToStringMappingFrequency::
  IDSubTypeToStringMappingFrequency( )
  {
    m_types.push_back( "Unknown/user defined" );
    m_types.push_back( "Amplitude Spectral Density" );
    m_types.push_back( "Power Spectral Density" );
    m_types.push_back( "Cross Spectral Density" );
    m_types.push_back( "Coherence" );
    m_types.push_back( "Transfer Function" );
  }

  const std::string& IDSubTypeToStringMappingFrequency::
  GetString( INT_2U Type ) const
  {
    if ( Type >= m_types.size( ) )
    {
      std::ostringstream	msg;
      msg << "Cannot map FrProcData type value of " << Type
	  << " to string value because it is out of range";
      throw std::range_error( msg.str( ) );
    }
    return m_types[ Type ];
  }
} // namespace - anonymous
