#ifndef FrameCPP_VERSION_6_FrRawData_HH
#define FrameCPP_VERSION_6_FrRawData_HH

#include <string>
#include <vector>

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/FrAdcData.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrMsg.hh"
#include "framecpp/Version6/FrSerData.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrRawData: public FrBase
    {
    public:
      typedef SearchContainer< FrSerData, &FrSerData::GetName >	firstSer_type;
      typedef SearchContainer< FrAdcData, &FrAdcData::GetName >	firstAdc_type;
      typedef SearchContainer< FrTable, &FrTable::GetName >	firstTable_type;
      typedef SearchContainer< FrMsg, &FrMsg::GetAlarm >	logMsg_type;
      typedef SearchContainer< FrVect, &FrVect::GetName >	more_type;

      typedef firstSer_type::iterator firstSer_iterator;
      typedef firstSer_type::const_iterator const_firstSer_iterator;
      typedef firstAdc_type::iterator firstAdc_iterator;
      typedef firstAdc_type::const_iterator const_firstAdc_iterator;
      typedef firstTable_type::iterator firstTable_iterator;
      typedef firstTable_type::const_iterator const_firstTable_iterator;
      typedef logMsg_type::iterator logMsg_iterator;
      typedef logMsg_type::const_iterator const_logMsg_iterator;
      typedef more_type::iterator more_iterator;
      typedef more_type::const_iterator const_more_iterator;

      typedef firstSer_type::hash_iterator firstSer_hash_iterator;
      typedef firstSer_type::const_hash_iterator const_firstSer_hash_iterator;
      typedef firstAdc_type::hash_iterator firstAdc_hash_iterator;
      typedef firstAdc_type::const_hash_iterator const_firstAdc_hash_iterator;
      typedef firstTable_type::hash_iterator firstTable_hash_iterator;
      typedef firstTable_type::const_hash_iterator const_firstTable_hash_iterator;
      typedef logMsg_type::hash_iterator logMsg_hash_iterator;
      typedef logMsg_type::const_hash_iterator const_logMsg_hash_iterator;
      typedef more_type::hash_iterator more_hash_iterator;
      typedef more_type::const_hash_iterator const_more_hash_iterator;

      static const CHAR* getStaticName();

      //: Default constructor
      FrRawData( );
      FrRawData( const FrRawData& rawData );
      FrRawData( const std::string& name );
      ~FrRawData( );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      const std::string& GetName() const;
      const firstSer_type& RefFirstSer( ) const { return m_data.firstSer; }
      firstSer_type& RefFirstSer( ) { return m_data.firstSer; }
      const firstAdc_type& RefFirstAdc( ) const { return m_data.firstAdc; }
      firstAdc_type&  RefFirstAdc( ) { return m_data.firstAdc; }
      const firstTable_type& RefFirstTable( ) const { return m_data.firstTable; }
      firstTable_type& RefFirstTable( ) { return m_data.firstTable; }
      const logMsg_type& RefLogMsg( ) const { return m_data.logMsg; }
      logMsg_type& RefLogMsg( ) { return m_data.logMsg; }
      const more_type& RefMore( ) const { return m_data.more; }
      more_type& RefMore( ) { return m_data.more; }
   
      //: Read the structure from stream
      static FrRawData* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct m_data_type {
	STRING		name;
	firstSer_type	firstSer;
	firstAdc_type	firstAdc;
	firstTable_type	firstTable;
	logMsg_type	logMsg;
	more_type		more;

	m_data_type( );
      } m_data;

      void read( IFrameStream& Stream );
#if ENABLE_DEPRICATED_FRAMECPP
    public:
      typedef firstSer_type SerDataContainer;
      typedef firstAdc_type AdcDataContainer;
      typedef firstTable_type TableContainer;
      typedef logMsg_type MsgContainer;
      typedef more_type MoreContainer;

      typedef firstSer_iterator serData_iterator;
      typedef const_firstSer_iterator const_serData_iterator;
      typedef firstAdc_iterator adcData_iterator;
      typedef const_firstAdc_iterator const_adcData_iterator;
      typedef firstTable_iterator table_iterator;
      typedef const_firstTable_iterator const_table_iterator;
      typedef logMsg_iterator msg_iterator;
      typedef const_logMsg_iterator const_msg_iterator;

      typedef firstSer_hash_iterator serData_hash_iterator;
      typedef const_firstSer_hash_iterator const_serData_hash_iterator;
      typedef firstAdc_hash_iterator adcData_hash_iterator;
      typedef const_firstAdc_hash_iterator const_adcData_hash_iterator;
      typedef firstTable_hash_iterator table_hash_iterator;
      typedef const_firstTable_hash_iterator const_table_hash_iterator;
      typedef logMsg_hash_iterator msg_hash_iterator;
      typedef const_logMsg_hash_iterator const_msg_hash_iterator;

      const std::string& getName() const;

      SerDataContainer& refSerial() { return RefFirstSer( ); }
      const SerDataContainer& refSerial() const { return RefFirstSer( ); }
      AdcDataContainer& refAdc() { return RefFirstAdc( ); }
      const AdcDataContainer& refAdc() const { return RefFirstAdc( ); }
      //const AdcData* findAdc( INT_4U channel ) const;
      //AdcData* findAdc( INT_4U channel );
      TableContainer& refTable() { return RefFirstTable( ); }
      const TableContainer& refTable() const { return RefFirstTable( ); }
      MsgContainer& refMsg() { return RefLogMsg( ); }
      const MsgContainer& refMsg() const { return RefLogMsg( ); }
      MoreContainer& refMore() { return RefMore( ); }
      const MoreContainer& refMore() const { return RefMore( ); }

#endif
    };
  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrRawData_HH */
