#ifndef FrameCPP_VERSION_6_FrSerData_HH
#define FrameCPP_VERSION_6_FrSerData_HH

#include <vector>

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrHistory.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrSerData: public FrBase
    {
    public:
      typedef SearchContainer< FrVect, &FrVect::GetName >	serial_type;
      typedef SearchContainer< FrTable, &FrTable::GetName >	table_type;

      typedef serial_type::iterator iterator;
      typedef serial_type::const_iterator const_iterator;
      typedef table_type::iterator table_iterator;
      typedef table_type::const_iterator const_table_iterator;

      static const CHAR* getStaticName();

      //: Default constructor
      FrSerData( );
      FrSerData( const FrSerData& serData );
      FrSerData( const std::string& name,
		 const GPSTime& time,
		 REAL_4 sampleRate );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      const std::string& GetName() const;
      const GPSTime& GetTime( ) const { return m_data.time; }
      REAL_8 GetSampleRate( ) const { return m_data.sampleRate; }
      const std::string& GetData( ) const { return m_data.data; }
      void SetData( const std::string& Data )
      {
         m_data.data = Data;
         return;
      }

      const serial_type& RefSerial( ) const { return m_data.serial; }
      serial_type& RefSerial( ) { return m_data.serial; }
      const table_type& RefTable( ) const { return m_data.table; }
      table_type& RefTable( ) { return m_data.table; }   
   
      //: Merge with another FrAdcData
      FrSerData& Merge( const FrSerData& RHS );

      //: Read the structure from stream
      static FrSerData* ReadDynamic( IFrameStream& Stream,
				     const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct {
	STRING		name;
	GPSTime		time;
	REAL_4		sampleRate;
	STRING		data;
	serial_type	serial;
	table_type		table;
      } m_data;

      void read( IFrameStream& Stream );
#if ENABLE_DEPRICATED_FRAMECPP
    public:
      typedef serial_type SerialContainer;
      typedef table_type TableContainer;

      const std::string& getName( ) const;

      const GPSTime& getTime( ) const
      {
	return GetTime( );
      }
       REAL_8 getSampleRate( ) const
      {
	return GetSampleRate( );
      }
      const std::string& getData( ) const
      {
	return GetData( );
      }

      void setData( const std::string& data )
      {
	SetData( data );
      }

      const SerialContainer&	refData( ) const
      {
	return RefSerial( );
      }

      SerialContainer&		refData( )
      {
	return RefSerial( );
      }
      const TableContainer&	refTable( ) const
      {
	return RefTable( );
      }
      TableContainer&		refTable( )
      {
	return RefTable( );
      }
#endif	/* ENABLE_DEPRICATED_FRAMECPP */
    };
  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrSerData_HH */
