#include "framecpp/config.h"

#include "framecpp/Version3_4_5/simdata.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrSimData.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

namespace {
  class MapSimDataRef
    : public IFrameStream::MapRef< FrameCPP::Version_3_4_5::SimData, FrSimData >
  {
  public:
    MapSimDataRef( IFrameStream& Stream,
		 const FrameCPP::Version_3_4_5::SimData* SimData5,
		 FrSimData& SimData6 )
      : IFrameStream::MapRef< FrameCPP::Version_3_4_5::SimData, FrSimData >( Stream, SimData5, SimData6 )
    {
    }

  protected:
   
    // virtual
    void operator()( const void* Ref5Ptr,
                     const PTR_STRUCT& PtrStruct );

  }; // class - MapSimDataRef
   
   
  void MapSimDataRef::operator()( const void* Ref5Ptr,
          		          const PTR_STRUCT& PtrStruct )
  {
      if ( Ref5Ptr == &(m_struct_5->refData( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefData( ), PtrStruct );
      }
      else if ( Ref5Ptr == &(m_struct_5->refInput( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefInput( ), PtrStruct );
      }
      else if ( Ref5Ptr == &(m_struct_5->refTable( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefTable( ), PtrStruct );
      }
   
      return;
  }
   
} // namespace - anonymous


const CHAR* FrSimData::getStaticName()
{
   static const CHAR* class_name( "FrSimData" );
   return class_name;
}


FrSimData::
FrSimData( )
  : FrBase( CLASS_FR_SIM_DATA )
{
}

FrSimData::
FrSimData( const FrSimData& SimData )
  : FrBase( CLASS_FR_SIM_DATA )
{
  m_data.name = SimData.m_data.name;
  m_data.comment = SimData.m_data.comment;
  m_data.sampleRate = SimData.m_data.sampleRate;
  m_data.timeOffset = SimData.m_data.timeOffset;
  m_data.fShift = SimData.m_data.fShift;
  m_data.phase = SimData.m_data.phase;
  m_data.data = SimData.m_data.data;
  m_data.input = SimData.m_data.input;
  m_data.table = SimData.m_data.table;
}

FrSimData::
FrSimData( const std::string& name,
	   const std::string& comment,
	   REAL_4 sampleRate, 
	   REAL_8 fshift,
	   REAL_4 phase )
  : FrBase( CLASS_FR_SIM_DATA )
{
  m_data.name = name;
  m_data.comment = comment;
  m_data.sampleRate = sampleRate;
  m_data.timeOffset = 0;
  m_data.fShift = fshift;
  m_data.phase = phase;
}


const std::string& FrSimData::GetName() const
{ 
   return m_data.name;
}


#if ENABLE_DEPRICATED_FRAMECPP
const std::string& FrSimData::getName() const
{
   return GetName( );
}
#endif


INT_8U FrSimData::
Bytes( ) const
{
  return
    FrBase::Bytes( ) +
    m_data.name.Bytes( ) +
    m_data.comment.Bytes( ) +
    sizeof( m_data.sampleRate ) +
    sizeof( m_data.timeOffset ) +
    sizeof( m_data.fShift ) +
    sizeof( m_data.phase ) +
    PTR_STRUCT::Bytes( ) +	// data
    PTR_STRUCT::Bytes( ) +	// input
    PTR_STRUCT::Bytes( ) +	// table
    PTR_STRUCT::Bytes( )	// next
    ;
}

FrSimData& FrSimData::
Merge( const FrSimData& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrSimData* FrSimData::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrSimData > ret( new FrSimData );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrSimData::
Write( OFrameStream& Stream ) const
{
  Stream << m_data.name
	 << m_data.comment
	 << m_data.sampleRate
	 << m_data.timeOffset
	 << m_data.fShift
	 << m_data.phase
	 << m_data.data.containerCast( )
	 << m_data.input.containerCast( )
	 << m_data.table.containerCast( )
	 << GetNext( )
    ;
}

void FrSimData::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_SIM_DATA,
				"Simulated Data Structure" ) );

  Stream.WriteDictionary( FrSE( "name", "STRING" ) );
  Stream.WriteDictionary( FrSE( "comment", "STRING" ) );
  Stream.WriteDictionary( FrSE( "sampleRate", "REAL_4" ) );
  Stream.WriteDictionary( FrSE( "timeOffset", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "fShift", "REAL_8" ) );
  Stream.WriteDictionary( FrSE( "phase", "REAL_4" ) );

  Stream.WriteDictionary( FrSE( "data",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "input",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );
  Stream.WriteDictionary( FrSE( "table",
				PTR_STRUCT::Desc( FrTable::getStaticName() ) ) );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );


  return;
}

void FrSimData::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< SimData >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      Stream >> m_data.name
	     >> m_data.comment
	     >> m_data.sampleRate
	     >> m_data.timeOffset
	     >> m_data.fShift
	     >> m_data.phase
	     >> m_data.data.containerCast( )
	     >> m_data.input.containerCast( )
	     >> m_data.table.containerCast( )
	     >> next
	;
    }
    break;
  case 5:
    ver5.reset( SimData::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( SimData::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( SimData::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->getName( );
    m_data.comment = ver5->getComment( );
    m_data.sampleRate = ver5->getSampleRate( );
    m_data.timeOffset = 0;
    m_data.fShift = ver5->getFShift( );
    m_data.phase = ver5->getPhase( );

    MapSimDataRef	map_refs( Stream, ver5.get( ), *this );
    map_refs.Map();
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
