#include "framecpp/config.h"

#include <unistd.h>

#include <cassert>

#include <algorithm>
#include <iostream>
#include <list>
#include <memory>
#include <sstream>
#include <string>

#include "general/gpstime.hh"
#include "general/mutexlock.hh"

#include "framecpp/Version3_4_5/framereaderbase.hh"
#include "framecpp/Version3_4_5/input.hh"
#include "framecpp/Version3_4_5/time.hh"
#include "framecpp/Version3_4_5/toc.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrameH.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"

using namespace std;
using namespace FrameCPP::Version_6;

#define SINGLE_FRTOC_READ 0

#define LM_DEBUG 0

#if LM_DEBUG
#define LM_INFO( a ) std::cerr << a << "( Line: " << __LINE__ << " File: " << __FILE__ << ")" << std::endl;
#else
#define	LM_INFO( a )
#endif


const CHAR* FrTOC::getStaticName()
{
   static const CHAR* class_name( "FrTOC" );
   return class_name;
}


#define write_map_element( LM_Stream, LM_Map, LM_Element ) \
  for ( Map##LM_Map##_type::const_iterator i( m_##LM_Map.begin( ) ); \
	i != m_##LM_Map.end( ); \
	i++ ) \
  { \
    LM_Stream << (*i).LM_Element; \
  }

namespace
{
  typedef std::list< STRING >	alphabetical_type;
  
  template< class Input, class Output >
  void copy_data( const Input& In, Output& Out )
  {
    Out.resize( In.size( ) );
    std::copy( In.begin( ), In.end( ), Out.begin( ) );
  }

  INT_8U Bytes( const std::vector< STRING >& Base )
  {
    INT_8U	ret( 0 );
    for( std::vector< STRING >::const_iterator i( Base.begin( ) );
	 i != Base.end( );
	 i++ )
    {
      ret += (*i).Bytes( );
    }
    return ret;
  }
  
  INT_8U Bytes( const General::hash_map< std::string, FrTOC::position_list_type >& Map )
  {
    INT_8U	ret( sizeof( INT_4U ) );
    INT_4U	count( 0 );

    for( General::hash_map< std::string, FrTOC::position_list_type >::const_iterator
	   i( Map.begin( ) );
	 i != Map.end( );
	 i++, count++ )
    {
      assert( count < Map.size( ) );
      ret += STRING::Bytes( (*i).first );
      ret += ( sizeof( FrTOC::position_type ) * (*i).second.size( ) );
    }
    return ret;
  }

  INT_8U Bytes( const General::hash_map< std::string, FrTOC::position_type >& Map )
  {
    INT_8U	ret( sizeof( INT_4U ) );
    INT_4U	count( 0 );

    for( General::hash_map< std::string, FrTOC::position_type >::const_iterator
	   i( Map.begin( ) );
	 i != Map.end( );
	 i++, count++ )
    {
      assert( count < Map.size( ) );
      ret += STRING::Bytes( (*i).first ) +
	sizeof( FrTOC::position_type );
    }
    return ret;
  }

  INT_8U Bytes( const General::hash_map< std::string, FrTOC::EventList_type >& Map )
  {
    INT_8U	ret( sizeof( INT_4U ) );
    INT_4U	l( 0 );
    INT_4U	count( 0 );

    for( General::hash_map< std::string, FrTOC::EventList_type>::const_iterator
	   i( Map.begin( ) );
	 i != Map.end( );
	 i++, count++ )
    {
      assert( count < Map.size( ) );
      ret += STRING::Bytes( (*i).first );
      l += (*i).second.size( );
    }
    ret +=
      ( sizeof( INT_4U ) * Map.size( ) ) + // nEvent
      ( sizeof( INT_4U ) * l ) + // GTimeS
      ( sizeof( INT_4U ) * l ) + // GTimeN
      ( sizeof( REAL_4 ) * l ) + // amplitude
      ( sizeof( INT_8U ) * l );  // position

    return ret;
  }

  template< class LM_Map >
  IFrameStream&
  read_event_type_map( IFrameStream& Stream,
		       General::hash_map< std::string, LM_Map >& Map )
  {
    INT_4U size;
    Stream >> size;
    if ( ( size ) && ( size != FrTOC::NO_DATA_AVAILABLE ) )
    {
      INT_4U	l( 0 );

      std::vector< STRING > name;
      std::vector< INT_4U > n;

      Stream.ReadVector( name, size );
      Stream.ReadVector( n, size );

      for( std::vector< INT_4U >::const_iterator i( n.begin( ) );
	   i != n.end( );
	   i++ )
      {
	if ( *i != FrTOC::NO_DATA_AVAILABLE )
	{
	  l += *i;
	}
      }

      std::vector< INT_4U >	GTimeS;
      std::vector< INT_4U >	GTimeN;
      std::vector< REAL_4 >	amplitude;
      std::vector< INT_8U >	position;

      Stream.ReadVector( GTimeS, l );
      Stream.ReadVector( GTimeN, l );
      Stream.ReadVector( amplitude, l );
      Stream.ReadVector( position, l );

      INT_4U	start( 0 );
      INT_4U	end( 0 );

      std::vector< INT_4U >::const_iterator	iGTimeS( GTimeS.begin( ) );
      std::vector< INT_4U >::const_iterator	iGTimeN( GTimeN.begin( ) );
      std::vector< REAL_4 >::const_iterator	iamplitude( amplitude.begin( ) );
      std::vector< INT_8U >::const_iterator	iposition( position.begin( ) );
      for ( INT_4U x = 0; x < n.size( ); x++ )
      {
	FrTOC::EventList_type&	e( Map[ name[ x ] ] );
	start = end;
	end += ( n[ x ] == FrTOC::NO_DATA_AVAILABLE ) ? 0 : n[ x ];
	e.resize( end - start );
	for ( FrTOC::EventList_type::iterator y( e.begin( ) );
	      y != e.end( );
	      y++, iGTimeS++, iGTimeN++, iamplitude++, iposition++ )
	{
	  (*y).GTimeS = *iGTimeS;
	  (*y).GTimeN = *iGTimeN;
	  (*y).amplitude = *iamplitude;
	  (*y).position = *iposition;
	}
      }
    }
    return Stream;
  }

  IFrameStream&
  read_position_map( IFrameStream& Stream,
		     General::hash_map< std::string, FrTOC::position_list_type >& Map,
		     std::vector< STRING >& IndexRef,
		     INT_4U NFrame )
  {
    INT_4U size;
    Stream >> size;
    if ( ( size ) && ( size != FrTOC::NO_DATA_AVAILABLE ) )
    {
      FrTOC::position_list_type	position;

      Stream.ReadVector( IndexRef, size );
      Stream.ReadVector( position, size * NFrame );

      for ( INT_4U x = 0; x < size; x++ )
      {
        const STRING& index( IndexRef[ x ] );
	//FrTOC::position_list_type&	data( Map[ IndexRef[ x ] ] );
	Map[ index ].resize( NFrame );
	std::copy( position.begin() + x * NFrame,
		   position.begin() + ( x + 1 ) * NFrame,
		   Map[ index ].begin() );
      }

    }
    return Stream;
  }

  IFrameStream&
  read_position_map( IFrameStream& Stream,
		     General::hash_map< std::string, FrTOC::position_list_type >& Map,
		     INT_4U NFrame )
  {
    std::vector< STRING > index_ref;

    read_position_map( Stream, Map, index_ref, NFrame );
    index_ref.erase( index_ref.begin( ), index_ref.end( ) );
    return Stream;
  }

  IFrameStream&
  read_position_map( IFrameStream& Stream,
		     General::hash_map< std::string, FrTOC::position_type >& Map,
		     INT_4U /* NFrame */ )
  {
    INT_4U size;
    Stream >> size;
    if ( ( size ) && ( size != FrTOC::NO_DATA_AVAILABLE ) )
    {
      std::vector< STRING >	name;
      FrTOC::position_list_type	position;

      Stream.ReadVector( name, size );
      Stream.ReadVector( position, size );

      for ( INT_4U x = 0; x < size; x++ )
      {
	FrTOC::position_type&	data( Map[ name[ x ] ] );
	data = position[ x ];
      }

    }
    return Stream;
  }
#if 0
  IFrameStream&
  read_position_map( IFrameStream& Stream,
		     General::hash_map< std::string, FrTOC::position_type >& Map,
		     INT_4U NFrame )
  {
    std::vector< STRING > index_ref;

    read_position_map( Stream, Map, index_ref, NFrame );
    index_ref.erase( index_ref.begin( ), index_ref.end( ) );
    return Stream;
  }
#endif

  template< class LM_Map >
  OFrameStream&
  write_event_type_map( OFrameStream& Stream, const LM_Map& Map )
  {
    INT_4U size( Map.size( ) );

    alphabetical_type	alphabetical;
    INT_4U	l( 0 );

    for ( typename LM_Map::const_iterator i( Map.begin( ) );
	  i != Map.end( );
	  i++ )
    {
      alphabetical.push_back( (*i).first );
      l += (*i).second.size( );
    }
    alphabetical.sort( );

    std::vector< INT_4U >	nEvent( Map.size( ) );
    std::vector< INT_4U >	GTimeS( l );
    std::vector< INT_4U >	GTimeN( l );
    std::vector< REAL_4 >	amplitude( l );
    std::vector< INT_8U >	position( l );

    std::vector< INT_4U >::iterator	inEvent( nEvent.begin( ) );
    std::vector< INT_4U >::iterator	iGTimeS( GTimeS.begin( ) );
    std::vector< INT_4U >::iterator	iGTimeN( GTimeN.begin( ) );
    std::vector< REAL_4 >::iterator	iamplitude( amplitude.begin( ) );
    std::vector< INT_8U >::iterator	iposition( position.begin( ) );

    for ( alphabetical_type::const_iterator i( alphabetical.begin( ) );
	  i != alphabetical.end( );
	  i++, inEvent++ )
    {
      INT_4U			end( Map.find( *i )->second.size( ) );
      *inEvent = end;
      const FrTOC::EventList_type&	events( Map.find( *i )->second );
      for( INT_4U x = 0;
	   x < end;
	   x++, iGTimeS++, iGTimeN++, iamplitude++, iposition++ )
      {
	*iGTimeS = events[ x ].GTimeS;
	*iGTimeN = events[ x ].GTimeN;
	*iamplitude = events[ x ].amplitude;
	*iposition = events[ x ].position;
      }
    }

    Stream << size
	   << alphabetical
	   << nEvent
	   << GTimeS
	   << GTimeN
	   << amplitude
	   << position;

    return Stream;
  }

  template< class LM_Map >
  OFrameStream&
  write_position_map( OFrameStream& Stream, const LM_Map& Map )
  {
    INT_4U size( Map.size( ) );
    Stream << size;
    for ( typename LM_Map::const_iterator i( Map.begin( ) );
	  i != Map.end( );
	  i++ )
    {
      Stream << (*i).first;
    }
    for ( typename LM_Map::const_iterator i( Map.begin( ) );
	  i != Map.end( );
	  i++ )
    {
      Stream << (*i).second;
    }
    return Stream;
  }

  template< >
  OFrameStream&
  write_position_map( OFrameStream& Stream, const FrTOC::MapDetector_type& Map )
  {
    INT_4U size( Map.size( ) );

    alphabetical_type	alphabetical;

    for ( FrTOC::MapDetector_type::const_iterator i ( Map.begin( ) );
	  i != Map.end( );
	  i++ )
    {
      alphabetical.push_back( (*i).first );
    }
    alphabetical.sort( );

    std::vector< FrTOC::position_type > p;
    for ( alphabetical_type::const_iterator i( alphabetical.begin( ) );
	  i != alphabetical.end( );
	  i++ )
    {
      p.push_back( Map.find( *i )->second );
    }

    Stream << size
	   << alphabetical
	   << p;
    return Stream;
  }

}

class FrTOC::TOCCallback: public Dictionary::Callback
{
public:
  TOCCallback( FrTOC& TOC )
    : m_toc( TOC )
  {
  }

private:
  INT_8U tell( );

public:
   void operator()( FrBase& Base );
   
private:
  FrTOC&	m_toc;
};
   
   
void FrTOC::TOCCallback::operator()( FrBase& Base )
{
    switch( Base.GetClassId( ) )
    {
    case FrBase::CLASS_FR_SH:
      {
	const FrSH&	sh( dynamic_cast< const FrSH& >( Base ) );
	m_toc.m_SHid.push_back( sh.GetClass( ) );
	m_toc.m_SHname.push_back( sh.GetName( ) );
      }
      break;
    case CLASS_FRAME_H:
      {
	const FrameH&	frame( dynamic_cast< const FrameH& >( Base ) );
	if ( m_toc.GetNFrame( ) == 0 )
	{
	  m_toc.m_ULeapS = frame.GetULeapS( );
	}
	m_toc.m_dataQuality.push_back( frame.GetDataQuality( ) );
	m_toc.m_GTimeS.push_back( frame.GetGTime( ).GetSeconds( ) );
	m_toc.m_GTimeN.push_back( frame.GetGTime( ).GetNanoseconds( ) );
	m_toc.m_dt.push_back( frame.GetDt( ) );
	m_toc.m_runs.push_back( frame.GetRun( ) );
	m_toc.m_frame.push_back( frame.GetFrame( ) );
	// Record the position of this header
	m_toc.m_positionH.push_back( tell( ) );
	// Assume the next structure will not be supplied.
	// If they are, then over ride the value with the actual position
	m_toc.m_nFirstADC.push_back( (INT_8U)0 );
	m_toc.m_nFirstSer.push_back( (INT_8U)0 );
	m_toc.m_nFirstTable.push_back( (INT_8U)0 );
	m_toc.m_nFirstMsg.push_back( (INT_8U)0 );
      }
      break;
    case CLASS_FR_DETECTOR:
      {
	const FrDetector&
	  data( static_cast< const FrDetector& >( Base ) );

	if ( m_toc.m_Detector.find( data.GetName( ) ) ==
	     m_toc.m_Detector.end( ) )
	{
	  m_toc.m_Detector[ data.GetName( ) ] = tell( );
	}
      }
      break;
    case CLASS_FR_STAT_DATA:
      {
	const FrStatData&
	  data( static_cast< const FrStatData& >( Base ) );
	std::string	detector_name =
	  ( data.GetDetector( ) )
	  ? data.GetDetector( )->GetName( )
	  : "";

	FrTOC::StatType_type&
	  s = m_toc.m_StatType[ StatTypeKey( data.GetName( ),
					     detector_name ) ];

	s.tStart.push_back( data.GetTimeStart( ) );
	s.tEnd.push_back( data.GetTimeEnd( ) );
	s.version.push_back( data.GetVersion( ) );
	s.positionStat.push_back( tell( ) );
      }
      break;
    case CLASS_FR_ADC_DATA:
      {
	if ( m_toc.m_nFirstADC.back( ) == 0 )
	{
	  m_toc.m_nFirstADC.back( ) = tell( );
	}
	const FrAdcData&
	  data( static_cast< const FrAdcData& >( Base ) );

	INT_4U size( m_toc.m_ADC.size( ) );

	FrTOC::ADC_type&	adc( m_toc.m_ADC[ data.GetName( ) ] );
	if ( size == m_toc.m_ADC.size( ) )
	{
	  //:TODO: Sanity checks on data
	}
	else
	{
	  m_toc.m_ADCIndex.push_back( data.GetName( ) );
	  adc.channelID = data.GetChannelNumber( );
	  adc.groupID = data.GetChannelGroup( );
	}
	adc.positionADC.resize( m_toc.GetNFrame( ), 0 );
	adc.positionADC.back( ) = tell( );
      }
      break;
    case CLASS_FR_PROC_DATA:
      {
	const FrProcData&
	  data( static_cast< const FrProcData& >( Base ) );

	INT_4U	old_size = m_toc.m_Proc.size( );
	m_toc.m_Proc[ data.GetName( ) ].push_back( tell( ) );
	if ( m_toc.m_Proc.size( ) != old_size )
	{
	  m_toc.m_ProcIndex.push_back( data.GetName( ) );
	}
      }
      break;
    case CLASS_FR_SIM_DATA:
      {
	const FrSimData&
	  data( static_cast< const FrSimData& >( Base ) );

	m_toc.m_Sim[ data.GetName( ) ].push_back( tell( ) );
      }
      break;
    case CLASS_FR_SER_DATA:
      {
	if ( m_toc.m_nFirstSer.back( ) == 0 )
	{
	  m_toc.m_nFirstSer.back( ) = tell( );
	}
	const FrSerData&
	  data( static_cast< const FrSerData& >( Base ) );

	m_toc.m_Ser[ data.GetName( ) ].push_back( tell( ) );
      }
      break;
    case CLASS_FR_SUMMARY:
      {
	const FrSummary&
	  data( static_cast< const FrSummary& >( Base ) );

	m_toc.m_Summary[ data.GetName( ) ].push_back( tell( ) );
      }
      break;
    case CLASS_FR_EVENT:
      {
	const FrEvent&
	  data( static_cast< const FrEvent& >( Base ) );

	FrTOC::EventList_type&
	  event( m_toc.m_EventType[ data.GetName( ) ] );
	event.push_back( Event_type( data.GetGTime( ),
				  data.GetAmplitude( ),
				  tell( ) ) );
      }
      break;
    case CLASS_FR_SIM_EVENT:
      {
	const FrSimEvent&
	  data( static_cast< const FrSimEvent& >( Base ) );

	FrTOC::EventList_type&
	  event( m_toc.m_SimEventType[ data.GetName( ) ] );
	event.push_back( Event_type( data.GetGTime( ),
				  data.GetAmplitude( ),
				  tell( ) ) );
      }
      break;
    case CLASS_FR_TABLE:
      if ( m_toc.m_nFirstTable.back( ) == 0 )
      {
	m_toc.m_nFirstTable.back( ) = tell( );
      }
      break;
    case CLASS_FR_MSG:
      if ( m_toc.m_nFirstMsg.back( ) == 0 )
      {
	m_toc.m_nFirstMsg.back( ) = tell( );
      }
      break;
    default:
      break;
    }
}


const INT_4U	FrTOC::NO_DATA_AVAILABLE = 0xFFFFFFFF;

   
INT_8U FrTOC::TOCCallback::tell( )
{
   if ( m_toc.m_input_stream )
   {
     std::streampos pos( m_toc.m_input_stream->tellg( ) );

     if ( pos == (std::streampos)(-1) )
     {
        throw std::runtime_error( "Invalid stream position" );
     }
     return (INT_8U)( pos ) - m_toc.m_input_stream->GetLengthOfLastObject( );
   }
   
   std::streampos pos( m_toc.m_output_stream->tellp( ) );

   if ( pos == (std::streampos)(-1) )
   {
     throw std::runtime_error( "Invalid stream position" );
   }
   
   return pos;
}   
   
//=======================================================================
//=======================================================================

FrTOC::StatTypeKey::
StatTypeKey( const std::string& Name, const std::string& Detector )
  : s_name( Name ),
    s_detector( Detector )
{
}

bool FrTOC::StatTypeKey::
operator<( const StatTypeKey& RHS ) const
{
  int retval = s_name.compare( RHS.s_name );
  switch( retval )
  {
  case -1: return true;
  case 1: return false;
  default:
    return s_detector < RHS.s_detector;
    break;
  }
}
   
bool FrTOC::StatTypeKey::
operator==( const StatTypeKey& RHS ) const
{
  return ( ( s_name == RHS.s_name ) &&
	   ( s_detector == RHS.s_detector ) );
}

//=======================================================================
//=======================================================================
FrTOC::
FrTOC( )
  : FrBase( CLASS_FR_TOC ),
    m_dictionary( (Dictionary*)NULL ),
    m_input_stream( (IFrameStream*)NULL ),
    m_output_stream( (OFrameStream*)NULL ),
    m_callback( (Dictionary::Callback*)NULL )
{
}

FrTOC::
~FrTOC( )
{
  if ( m_callback != (Dictionary::Callback*)NULL )
  {
    delete m_callback;
    m_callback = (Dictionary::Callback*)NULL;
  }
}

void FrTOC::
BuildDynamically( IFrameStream& Stream )
{
  //---------------------------------------------------------------------
  // Choose to ignore most structures to improve reading speed
  //---------------------------------------------------------------------

  Stream.m_dictionary.DisableLibraryReaderAll( );
  Stream.m_dictionary.TrackObjects( false );
  Stream.m_dictionary.EnableLibraryReader( FrBase::CLASS_FR_SH );
  Stream.m_dictionary.EnableLibraryReader( FrBase::CLASS_FR_SE );
  RegisterCallbacks( Stream.m_dictionary, Stream );

  //---------------------------------------------------------------------
  // Read through all the objects building the toc along the way
  //---------------------------------------------------------------------

  std::auto_ptr< FrBase >	object;
  while( Stream.good( ) && ( Stream.eof( ) == false ) )
  {
    object.reset( Stream.ReadObject( ) );
  }

  //---------------------------------------------------------------------
  // Resume mode of reading everything
  //---------------------------------------------------------------------

  UnregisterCallbacks( );
  Stream.m_dictionary.TrackObjects( true );
  Stream.m_dictionary.EnableLibraryReaderAll( );
  Stream.clear( );
}

INT_8U FrTOC::
Bytes( ) const
{
  INT_8U	ret( FrBase::Bytes( ) );
  ret +=
    sizeof( m_ULeapS ) +
    sizeof( INT_4U ) + // nFrame
    ( GetNFrame( ) * ( sizeof( INT_4U ) + /* dataQuality */
		       sizeof( INT_4U ) + /* GTimeS */
		       sizeof( INT_4U ) + /* GTimeN */
		       sizeof( REAL_8 ) + /* dt */
		       sizeof( INT_4S ) + /* runs */
		       sizeof( INT_4U ) + /* frame */
		       sizeof( position_type ) + /* positionH */
		       sizeof( position_type ) + /* nFirstADC */
		       sizeof( position_type ) + /* nFirstSer */
		       sizeof( position_type ) + /* nFirstTable */
		       sizeof( position_type )   /* nFirstMsg */
		       ) )
    ;
  LM_INFO( ret );
  ret +=
    sizeof( INT_4U ) + // nSH
    GetSHid( ).size( ) * sizeof( INT_2U ) +
    ::Bytes( GetSHname( ) );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += 
    ::Bytes( GetDetector( ) );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret +=
    sizeof( INT_4U ); // nStatType
  for ( FrTOC::MapStatType_type::const_iterator i( m_StatType.begin( ) );
	i != m_StatType.end( );
	i++ )
  {
    ret +=
      STRING::Bytes( (*i).first.s_name ) + // nameStat
      STRING::Bytes( (*i).first.s_detector ) + // detector
      sizeof( INT_4U ) + // nStatInstance
      ( (*i).second.tStart.size( ) * ( sizeof( INT_4U ) + // tStart
				       sizeof( INT_4U ) + // tEnd
				       sizeof( INT_4U ) + // version
				       sizeof( FrTOC::position_type ) // positionStat
				       ) )
      ;
  }
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret +=
    sizeof( INT_4U ); //nADC
  for( MapADC_type::const_iterator i( m_ADC.begin( ) );
       i != m_ADC.end( );
       i++ )
  {
    ret +=
      STRING::Bytes( (*i).first ) +
      sizeof( INT_4U ) + // channelID
      sizeof( INT_4U ) + // groupID
      ( GetNFrame( ) * sizeof( position_type ) ); // positionADC
  }
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_Proc );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_Sim );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_Ser );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_Summary );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_EventType );
  LM_INFO( ret );
  //---------------------------------------------------------------------
  ret += ::Bytes( m_SimEventType );
  LM_INFO( ret );
  return ret;
}

std::string FrTOC::
GetClassName( INT_2U ClassId ) const
{
  std::ostringstream	result;

  typedef std::vector< INT_2U > shid_v_type;

  const shid_v_type&	shid( GetSHid( ) );
  INT_4U	offset = shid.size( );
  
  for( shid_v_type::const_iterator
	 begin = shid.begin( ),
	 cur = shid.begin( ),
	 end = shid.end( );
       cur != end;
       ++cur )
  {
    if ( (*cur) == ClassId )
    {
      offset = (cur - begin);
      break;
    }
  }
  if ( offset == shid.size( ) )
  {
    result << ClassId;
  }
  else
  {
    result << GetSHname( )[ offset ];
  }
  return result.str( );
}

FrTOC::MapADC_type::const_iterator FrTOC::
GetADC( const std::string& Channel ) const
{
  return m_ADC.find( Channel );
}

FrTOC::MapADC_type::const_iterator FrTOC::
GetADC( INT_4U Channel ) const
{
  if ( Channel >= GetADC( ).size( ) )
  {
    return GetADC( ).end( );
  }
  return GetADC( ).find( m_ADCIndex[ Channel ] );
}

FrTOC::MapProc_type::const_iterator FrTOC::
GetProc( const std::string& Channel ) const
{
  return GetProc( ).find( Channel );
}

FrTOC::MapProc_type::const_iterator FrTOC::
GetProc( INT_4U Channel ) const
{
  if ( Channel >= GetProc( ).size( ) )
  {
    return GetProc( ).end( );
  }
  return GetProc( ).find( m_ProcIndex[ Channel ] );
}

FrTOC::MapSer_type::const_iterator FrTOC::
GetSer( const std::string& Channel ) const
{
  return GetSer( ).find( Channel );
}

FrTOC::MapSer_type::const_iterator FrTOC::
GetSer( INT_4U Channel ) const
{
  if ( Channel >= GetSer( ).size( ) )
  {
    return GetSer( ).end( );
  }
  return GetSer( ).find( m_SerIndex[ Channel ] );
}

FrTOC::MapSim_type::const_iterator FrTOC::
GetSim( const std::string& Channel ) const
{
  return GetSim( ).find( Channel );
}

FrTOC::MapSim_type::const_iterator FrTOC::
GetSim( INT_4U Channel ) const
{
  if ( Channel >= GetSim( ).size( ) )
  {
    return GetSim( ).end( );
  }
  return GetSim( ).find( m_SimIndex[ Channel ] );
}

FrTOC::MapEventType_type::const_iterator FrTOC::
GetEventType( const std::string& Channel ) const
{
  return GetEventType( ).find( Channel );
}

FrTOC::MapSimEventType_type::const_iterator FrTOC::
GetSimEventType( const std::string& Channel ) const
{
  return GetSimEventType( ).find( Channel );
}

FrTOC* FrTOC::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrTOC > ret( new FrTOC );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrTOC::
RegisterCallbacks( Dictionary& Dictionary, IFrameStream& Stream )
{
  if ( &Dictionary != m_dictionary )
  {
    UnregisterCallbacks( );
    if ( m_callback == (Dictionary::Callback*)NULL )
    {
      m_callback = new TOCCallback( *this );
    }
    m_dictionary = &Dictionary;
    m_input_stream = &Stream;
    m_output_stream = (OFrameStream*)NULL;
    register_callbacks( );
  }
}

void FrTOC::
RegisterCallbacks( Dictionary& Dictionary, OFrameStream& Stream )
{
  if ( &Dictionary != m_dictionary )
  {
    UnregisterCallbacks( );
    if ( m_callback == (Dictionary::Callback*)NULL )
    {
      m_callback = new TOCCallback( *this );
    }
    m_dictionary = &Dictionary;
    m_input_stream = (IFrameStream*)NULL;
    m_output_stream = &Stream;
    register_callbacks( );
  }
}

void FrTOC::
UnregisterCallbacks( )
{
  if ( m_dictionary )
  {
    for ( std::vector< Dictionary::callback_key_type >::iterator
	    i ( m_callback_keys.begin( ) );
	  i != m_callback_keys.end( );
	  i++ )
    {
      m_dictionary->RemoveCallback( *i );
    }
    m_dictionary = (Dictionary*)NULL;
    m_input_stream = (IFrameStream*)NULL;
    m_output_stream = (OFrameStream*)NULL;
    m_callback_keys.erase( m_callback_keys.begin( ),
			   m_callback_keys.end( ) );
  }
}

void FrTOC::
Write( OFrameStream& Stream ) const
{
#if LM_DEBUG
  INT_8U head( Stream.tellp( ) - FrBase::Bytes( ) );
#endif
  LM_INFO( ( Stream.tellp( ) - head ) );
  Stream
    //--------------------------------
    << GetULeapS( )
    << GetNFrame( )
    << GetDataQuality( )
    << GetGTimeS( )
    << GetGTimeN( )
    << GetDt( )
    << GetRuns( )
    << GetFrame( )
    << GetPositionH( )
    << GetNFirstADC( )
    << GetNFirstSer( )
    << GetNFirstTable( )
    << GetNFirstMsg( )
    ;
  LM_INFO( ( Stream.tellp( ) - head ) );
  //----------------------------------
  Stream
    << GetNSH( )
    << GetSHid( )
    << GetSHname( )
    //--------------------------------
    ;
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_position_map( Stream, m_Detector );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  Stream << GetNStatType( )
    ;
  for ( FrTOC::MapStatType_type::const_iterator i( GetStatType( ).begin( ) );
	i != GetStatType( ).end( );
	i++ )
  {
    Stream << (*i).first.s_name // nameStat
	   << (*i).first.s_detector
	   << (INT_4U)( (*i).second.tStart.size( ) )
	   << (*i).second.tStart
	   << (*i).second.tEnd
	   << (*i).second.version
	   << (*i).second.positionStat
      ;
  }
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  Stream << GetNADC( );
  write_map_element( Stream, ADC, first );
  write_map_element( Stream, ADC, second.channelID );
  write_map_element( Stream, ADC, second.groupID );
  write_map_element( Stream, ADC, second.positionADC );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_position_map( Stream, m_Proc );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_position_map( Stream, m_Sim );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_position_map( Stream, m_Ser );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_position_map( Stream, m_Summary );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_event_type_map( Stream, m_EventType );
  LM_INFO( ( Stream.tellp( ) - head ) );
  //--------------------------------
  write_event_type_map( Stream, m_SimEventType );
  LM_INFO( ( Stream.tellp( ) - head ) );
}

void FrTOC::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary
    ( FrSH( getStaticName(), CLASS_FR_TOC, "Table of Contents Structure" ) );
  Stream.WriteDictionary
    ( FrSE( "ULeapS", "INT_2S", "From the first FrameH in the file" ) );
  Stream.WriteDictionary
    ( FrSE( "nFrame", "INT_4U", "Number of frames in the file" ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "dataQuality", "*INT_4U", "Array of integer QA words from each FrameH (size of nFrames)" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeS", "*INT_4U", "Array of integer GPS frae times (size of nFrames" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeN", "*INT_4U", "Array of integer GPS residual nanoseconds for the frame (size of nFrame" ) );
  Stream.WriteDictionary
    ( FrSE( "dt", "*REAL_8", "Array of frame durations in seconds( size of nFrames)" ) );
  Stream.WriteDictionary
    ( FrSE( "runs", "*INT_4S", "Array of run numbers (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "frame", "*INT_4U", "Array of frame numbers (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "positionH", "*INT_8U", "Array of FrameH positions, in bytes, from the bginning of file (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "nFirstADC", "*INT_8U", "Array of first FrADCData positions, in bytes, from beginning of file (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "nFirstSer", "*INT_8U", "Array of first FrSerData positions in bytes, from beginning of file (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "nFirstTable", "*INT_8U", "Array of first FrTable positions, in bytes, from beginning of file (size of nFrame)" ) );
  Stream.WriteDictionary
    ( FrSE( "nFirstMsg", "*INT_8U", "Array of first FrMsg positions, in bytes, from beginning of file (size of nFrame)" ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nSH", "INT_4U", "Number of FrSH structures in the file" ) );
  Stream.WriteDictionary
    ( FrSE( "SHid", "*INT_2U", "Array of FrSH IDs (size of nSH)" ) );
  Stream.WriteDictionary
    ( FrSE( "SHname", "*STRING", "Array of FrSH names (size of nSH)" ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nDetector", "INT_4U",
	    "Number of distinct types of FrDetector in the file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameDetector", "*STRING", "Array of FrDetector names (size of nDetector). They appear alphabetically" ) );
  Stream.WriteDictionary
    ( FrSE( "positionDetector", "*INT_8U", "Array of FrDetector positions from the beginning of file (size of nDetector)." ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nStatType", "INT_4U",
	    "Number of static data block types in the file." ) );
  Stream.WriteDictionary( FrSE( "nameStat", "STRING", "FrStatData name" ) );
  Stream.WriteDictionary( FrSE( "detector", "STRING", "Detector name" ) );
  Stream.WriteDictionary( FrSE( "nStatInstance", "INT_4U",
				"Number of instance for this FrStatData" ) );
  Stream.WriteDictionary
    ( FrSE( "tStart", "*INT_4U",
	    "Array of GPS integer start times, in seconds" ) );
  Stream.WriteDictionary
    ( FrSE( "tEnd", "*INT_4U",
	    "Array of GPS integer end times, in seconds" ) );
  Stream.WriteDictionary( FrSE( "version", "*INT_4U",
				"Array of version time" ) );
  Stream.WriteDictionary
    ( FrSE( "positionStat", "*INT_8U",
	    "Array of FrStatData positions from beginning of file" ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nADC", "INT_4U",
	    "Number of unique FrAdcData names in file." ) );
  Stream.WriteDictionary
    ( FrSE( "name", "*STRING", "Array of FrAdcData names" ) );
  Stream.WriteDictionary
    ( FrSE( "channelID", "*INT_4U",
	    "Array of ADC channel IDs" ) );
  Stream.WriteDictionary
    ( FrSE( "groupID", "*INT_4U",
	    "Array of ADC group IDs" ) );
  Stream.WriteDictionary
    ( FrSE( "positionADC", "*INT_8U",
	    "Array of lists of FrAdcData offset positions, in bytes, from"
	    " beginning of file (size of nFrame*nADC)"
	    ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nProc", "INT_4U", "Number of unique FrProcData names in file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameProc", "*STRING", "Array of FrProcData names" ) );
  Stream.WriteDictionary
    ( FrSE( "positionProc", "*INT_8U",
	    "Array of lists of FrProcData offset positions, in bytes, from"
	    " beginning of file (size of nFrame*nProc)"
	    ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nSim", "INT_4U", "Number of unique FrSimData names in file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameSim", "*STRING", "Array of FrSimData names" ) );
  Stream.WriteDictionary
    ( FrSE( "positionSim", "*INT_8U",
	    "Array of lists of FrSimData offset positions, in bytes, from"
	    " beginning of file (size of nFrame*nSim)"
	    ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nSer", "INT_4U", "Number of unique FrSerData names in file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameSer", "*STRING", "Array of FrSerData names" ) );
  Stream.WriteDictionary
    ( FrSE( "positionSer", "*INT_8U",
	    "Array of lists of FrSerData offset positions, in bytes, from"
	    " beginning of file (size of nFrame*nSer)"
	    ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nSummary", "INT_4U", "Number of unique FrSummary names in file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameSum", "*STRING", "Array of FrSummary names" ) );
  Stream.WriteDictionary
    ( FrSE( "positionSum", "*INT_8U",
	    "Array of lists of FrSummary offset positions, in bytes, from"
	    " beginning of file (size of nFrame*nSummary)"
	    ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nEventType", "INT_4U", "Number of FrEvent in the file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameEvent", "*STRING", "Array of FrEvent names" ) );
  Stream.WriteDictionary
    ( FrSE( "nEvent", "*INT_4U", "Number of FrEvent for each type of FrEvent (size of nEventType)" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeSEvent", "*INT_4U", "GPS time in integer seconds" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeNEvent", "*INT_4U",
	    "Residual GPS time in integer nanoseconds" ) );
  Stream.WriteDictionary
    ( FrSE( "amplitudeEvent", "*REAL_4", "Event amplitude" ) );
  Stream.WriteDictionary
    ( FrSE( "positionEvent", "*INT_8U",
	    "Array of FrEvent positions, in bytes, from beginning of file" ) );
  //---------------------------------------------------------------------
  Stream.WriteDictionary
    ( FrSE( "nSimEventType", "INT_4U", "Number of FrSimEvent in the file" ) );
  Stream.WriteDictionary
    ( FrSE( "nameSimEvent", "*STRING", "Array of FrSimEvent names" ) );
  Stream.WriteDictionary
    ( FrSE( "nSimEvent", "*INT_4U", "Number of FrEvent for each type of FrEvent (size of nEventType)" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeSSim", "*INT_4U", "GPS time in integer seconds" ) );
  Stream.WriteDictionary
    ( FrSE( "GTimeNSim", "*INT_4U",
	    "Residual GPS time in integer nanoseconds" ) );
  Stream.WriteDictionary
    ( FrSE( "amplitudeSimEvent", "*REAL_4", "SimEvent amplitude" ) );
  Stream.WriteDictionary
    ( FrSE( "positionSimEvent", "*INT_8U",
	    "Array of FrSimEvent positions, in bytes, from beginning of file" ) );
  
}

void FrTOC::
read( IFrameStream& Stream )
{
#if SINGLE_FRTOC_READ
  //---------------------------------------------------------------------
  // This method is very memory intensive. Only allow one at a time
  //   to be created to reduce memory contention
  //---------------------------------------------------------------------

  static pthread_mutex_t	lock_key = PTHREAD_MUTEX_INITIALIZER;
  MutexLock	lock( lock_key );
#endif /* SINGLE_FRTOC_READ */

  //---------------------------------------------------------------------
  // This reader will read 4, 5, and 6 versions of the TOC.
  //---------------------------------------------------------------------
  using namespace FrameCPP::Version_3_4_5;
  std::auto_ptr< TOC > toc5;

  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      INT_4U	size;

      //-----------------------------------------------------------------
      Stream >> m_ULeapS
	     >> size;	// nFrame
      Stream.ReadVector( m_dataQuality, size );
      Stream.ReadVector( m_GTimeS, size );
      Stream.ReadVector( m_GTimeN, size );
      Stream.ReadVector( m_dt, size );
      Stream.ReadVector( m_runs, size );
      Stream.ReadVector( m_frame, size );
      Stream.ReadVector( m_positionH, size );
      Stream.ReadVector( m_nFirstADC, size );
      Stream.ReadVector( m_nFirstSer, size );
      Stream.ReadVector( m_nFirstTable, size );
      Stream.ReadVector( m_nFirstMsg, size );
      //-----------------------------------------------------------------
      Stream >> size;	// nSH
      Stream.ReadVector( m_SHid, size );
      Stream.ReadVector( m_SHname, size );
      //-----------------------------------------------------------------
      read_position_map( Stream, m_Detector, 1 );
      //-----------------------------------------------------------------
      Stream >> size; // nStatType
      if ( size == NO_DATA_AVAILABLE )
      {
	size = 0;
      }
      for ( INT_4U x = 0; x < size; x++ )
      {
	INT_4U	nStatInstance;
	STRING	name;
	STRING	detector;

	Stream >> name
	       >> detector
	       >> nStatInstance;

	StatType_type& s( m_StatType[ StatTypeKey( name, detector ) ] );

	Stream.ReadVector( s.tStart, nStatInstance );
	Stream.ReadVector( s.tEnd, nStatInstance );
	Stream.ReadVector( s.version, nStatInstance );
	Stream.ReadVector( s.positionStat, nStatInstance );
      }
      //-----------------------------------------------------------------
      Stream >> size; // nADC
      if ( size == NO_DATA_AVAILABLE )
      {
	size = 0;
      }
      if ( size )
      {
	INT_4U	nFrame( GetNFrame( ) );

	std::vector< INT_4U > channelID;
	std::vector< INT_4U > groupID;
	position_list_type positionADC;

	Stream.ReadVector( m_ADCIndex, size );
	Stream.ReadVector( channelID, size );
	Stream.ReadVector( groupID, size );
	Stream.ReadVector( positionADC, size * nFrame );

	ADC_type*	adc;
	for( INT_4U x = 0; x < size; x++ )
	{
	  adc = &( m_ADC[ m_ADCIndex[ x ] ]);
	  adc->channelID = channelID[ x ];
	  adc->groupID = groupID[ x ];

	  adc->positionADC.resize( nFrame );
	  std::copy( positionADC.begin( ) + ( x * nFrame ),
		     positionADC.begin( ) + ( ( x + 1 ) *  nFrame ),
		     adc->positionADC.begin( ) );
	}
      }
      //-----------------------------------------------------------------
      read_position_map( Stream, m_Proc, m_ProcIndex, GetNFrame( ) );
      //-----------------------------------------------------------------
      read_position_map( Stream, m_Sim, m_SimIndex, GetNFrame( ) );
      //-----------------------------------------------------------------
      read_position_map( Stream, m_Ser, m_SerIndex, GetNFrame( ) );
      //-----------------------------------------------------------------
      read_position_map( Stream, m_Summary, GetNFrame( ) );
      //-----------------------------------------------------------------
      read_event_type_map( Stream, m_EventType );
      //-----------------------------------------------------------------
      read_event_type_map( Stream, m_SimEventType );
    }
    break;
  case 5:
  case 4:
    try {
      toc5.reset( new TOC );

      // for debugging only!!!
      //cout << "Going to sleep..." << endl;
      //sleep( 120 );
      
      toc5->reallyRead( Stream.GetExposedFrame5Reader( ) );
    }
    catch( ... )
    {
      toc5.reset( (TOC*)NULL );
    }
    break;
  }
  if ( toc5.get( ) )
  {
    //-------------------------------------------------------------------
    // Move in what was learned
    //-------------------------------------------------------------------
    m_ULeapS = toc5->getULeapS( );
    
    copy_data( toc5->dataQuality, m_dataQuality );
    m_GTimeS.resize( toc5->getGTime( ).size( ) );
    m_GTimeN.resize( toc5->getGTime( ).size( ) );
    std::vector< INT_4U >::iterator	gtsi( m_GTimeS.begin( ) );
    std::vector< INT_4U >::iterator	gtni( m_GTimeN.begin( ) );
    for( std::vector< Time >::const_iterator
	   gti(  toc5->getGTime( ).begin( ) );
	 gti != toc5->getGTime( ).end( );
	 gti++, gtsi++, gtni++ )
    {
      *gtsi = (*gti).getSec( );
      *gtni = (*gti).getNSec( );
    }
    copy_data( toc5->getDt( ), m_dt );
    copy_data( toc5->runs, m_runs );
    copy_data( toc5->getFrame( ), m_frame );
    copy_data( toc5->getPositionH( ), m_positionH );
    copy_data( toc5->getFirstADCPosition( ), m_nFirstADC );
    copy_data( toc5->getFirstSerPosition( ), m_nFirstSer );
    copy_data( toc5->getFirstTablePosition( ), m_nFirstTable );
    copy_data( toc5->getFirstMsgPosition( ), m_nFirstMsg );
    //-------------------------------------------------------------------
    copy_data( toc5->SHid, m_SHid );
    copy_data( toc5->SHname, m_SHname );
    //-------------------------------------------------------------------
    for ( TOC::DNPI i( toc5->detectorNamePositionMap.begin( ) );
	  i != toc5->detectorNamePositionMap.end( );
	  i++ )
    {
      position_type&	pos( m_Detector[ (*i).first ] );
      pos = (*i).second;
    }
    //-------------------------------------------------------------------
    struct StatType_type stat_struct;
    INT_4U nStatInstance;
    for ( TOC::SNPI i( toc5->statNamePositionMap.begin( ) );
	  i != toc5->statNamePositionMap.end( );
	  i++ )
    {
      nStatInstance = (*i).second.second.size( );
      stat_struct.tStart.resize( nStatInstance );
      stat_struct.tEnd.resize( nStatInstance );
      stat_struct.version.resize( nStatInstance );
      stat_struct.positionStat.resize( nStatInstance );
      for ( INT_4U x = 0; x < nStatInstance; x++ )
      {
	stat_struct.tStart[ x ] = (*i).second.second[ x ].first[ 0 ];
	stat_struct.tEnd[ x ] = (*i).second.second[ x ].first[ 1 ];
	stat_struct.version[ x ] = (*i).second.second[ x ].first[ 2 ];
	stat_struct.positionStat[ x ] = (*i).second.second[ x ].second;
      }
      m_StatType[ StatTypeKey( STRING( (*i).first ),
			       STRING( (*i).second.first ) ) ] = stat_struct;
    }
    //-------------------------------------------------------------------
    m_ADC.resize( toc5->adcNamePositionMap.size( ) );
    m_ADCIndex.erase( m_ADCIndex.begin( ), m_ADCIndex.end( ) );

    for ( TOC::ANPI i( toc5->adcNamePositionMap.begin( ) );
	  i != toc5->adcNamePositionMap.end( );
	  i++ )
    {
      m_ADCIndex.push_back( (*i).first );

      ADC_type&	adc( m_ADC[ (*i).first ] );

      adc.channelID = (*i).second.second[ 0 ];
      adc.groupID = (*i).second.second[ 1 ];
      adc.positionADC = (*i).second.first;
    }
    //-------------------------------------------------------------------
    m_ProcIndex.erase( m_ProcIndex.begin( ), m_ProcIndex.end( ) );
    for ( TOC::PNPI i( toc5->procNamePositionMap.begin( ) );
	  i != toc5->procNamePositionMap.end( );
	  i++ )
    {
      m_ProcIndex.push_back( (*i).first );

      position_list_type&	pos( m_Proc[ (*i).first ] );

      pos = (*i).second;
    }
    //-------------------------------------------------------------------
    m_SimIndex.erase( m_SimIndex.begin( ), m_SimIndex.end( ) );
    for ( TOC::SIMNPI i( toc5->simNamePositionMap.begin( ) );
	  i != toc5->simNamePositionMap.end( );
	  i++ )
    {
      m_SimIndex.push_back( (*i).first );

      position_list_type&	pos( m_Sim[ (*i).first ] );
      pos = (*i).second;
    }
    //-------------------------------------------------------------------
    m_SerIndex.erase( m_SerIndex.begin( ), m_SerIndex.end( ) );
    for ( TOC::SERNPI i( toc5->serNamePositionMap.begin( ) );
	  i != toc5->serNamePositionMap.end( );
	  i++ )
    {
      m_SerIndex.push_back( (*i).first );

      position_list_type&	pos( m_Ser[ (*i).first ] );
      pos = (*i).second;
    }
    //-------------------------------------------------------------------
    for ( TOC::SUMNPI i( toc5->sumNamePositionMap.begin( ) );
	  i != toc5->sumNamePositionMap.end( );
	  i++ )
    {
      position_list_type&	pos( m_Summary[ (*i).first ] );
      pos = (*i).second;
    }
    //-------------------------------------------------------------------
    for ( TOC::EVENTNPI i( toc5->eventNamePositionMap.begin( ) );
	  i != toc5->eventNamePositionMap.end( );
	  i++ )
    {
      FrTOC::EventList_type& event( m_EventType[ (*i).first ] );
      event.resize( (*i).second.size( ) );
      FrTOC::EventList_type::iterator pos( event.begin( ) );
      std::vector< TOC::eventAttr >::const_iterator b( (*i).second.begin( ) );
      std::vector< TOC::eventAttr >::const_iterator e( (*i).second.end( ) );
      while ( b != e )
      {
	(*pos).GTimeS = (*b).gtime.getSec( );
	(*pos).GTimeN = (*b).gtime.getNSec( );
	(*pos).amplitude = (*b).amplitude;
	(*pos).position = (*b).position;
	b++;
	pos++;
      }
    }
    //-------------------------------------------------------------------
    for ( TOC::SIMEVENTNPI i( toc5->simEventNamePositionMap.begin( ) );
	  i != toc5->simEventNamePositionMap.end( );
	  i++ )
    {
      FrTOC::EventList_type& event( m_SimEventType[ (*i).first ] );
      event.resize( (*i).second.size( ) );
      FrTOC::EventList_type::iterator pos( event.begin( ) );
      std::vector< TOC::eventAttr >::const_iterator b( (*i).second.begin( ) );
      std::vector< TOC::eventAttr >::const_iterator e( (*i).second.end( ) );
      while ( b != e )
      {
	(*pos).GTimeS = (*b).gtime.getSec( );
	(*pos).GTimeN = (*b).gtime.getNSec( );
	(*pos).amplitude = (*b).amplitude;
	(*pos).position = (*b).position;
	b++;
	pos++;
      }
    }

    //-------------------------------------------------------------------
    // Ensure the dataQuality vector is correct
    //-------------------------------------------------------------------
    if ( ( m_dataQuality.size( ) == 0 ) && ( m_dt.size( ) != 0 ) )
    {
      m_dataQuality.resize( m_dt.size( ) );
      std::fill( m_dataQuality.begin( ), m_dataQuality.end( ), 0 );
    }
  }
}

void FrTOC::
register_callbacks( )
{
  static INT_2U classes[] = {
    CLASS_FRAME_H,
    FrBase::CLASS_FR_SH,
    CLASS_FR_DETECTOR,
    CLASS_FR_STAT_DATA,
    CLASS_FR_ADC_DATA,
    CLASS_FR_PROC_DATA,
    CLASS_FR_SIM_DATA,
    CLASS_FR_SER_DATA,
    CLASS_FR_SUMMARY,
    CLASS_FR_EVENT,
    CLASS_FR_SIM_EVENT,
    CLASS_FR_TABLE,
    CLASS_FR_MSG,
  };
  if ( m_dictionary )
  {
    for ( INT_4U x = 0; x < sizeof( classes ) / sizeof( *classes ); x++ )
    {
      m_callback_keys.push_back( m_dictionary->AddCallback( classes[ x ],
							    *m_callback ) );
      m_dictionary->EnableLibraryReader( classes[ x ] );
    }
  }
}
