#ifndef FrameCPP_VERSION_6_FrTOC_HH
#define FrameCPP_VERSION_6_FrTOC_HH

#include <map>
#include <vector>

#include "general/hash_map.hh"
#include "general/types.hh"
#include "general/gpstime.hh"

#include "framecpp/Version6/Dictionary.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class IFrameStream;
    class OFrameStream;

    class FrTOC: public FrBase
    {
    public:
      class TOCCallback;

      static const INT_4U	NO_DATA_AVAILABLE;

      typedef INT_8U			position_type;
      typedef std::vector< position_type >	position_list_type;
      // order by data then nFrame
      typedef std::vector< position_list_type > position_dbl_list_type;

      class StatTypeKey {
      public:
	std::string	s_name;
	std::string	s_detector;

	StatTypeKey( const std::string& Name,
		     const std::string& Detector );
	bool operator<( const StatTypeKey& RHS ) const;
	bool operator==( const StatTypeKey& RHS ) const;
	bool operator!=( const StatTypeKey& RHS ) const;
      };

      struct StatType_type
      {
	std::vector< INT_4U >	tStart;
	std::vector< INT_4U >	tEnd;
	std::vector< INT_4U >	version;
	position_list_type	positionStat;
      };

      class StatTypeKeyHash
      {
      public:
	size_t operator()( const StatTypeKey& Key ) const
	{
	  General::hash< const char* > h;

	  std::string key( Key.s_name );
	  key += Key.s_detector;
	  return h( key.c_str( ) );
	}
      };

      typedef General::hash_map< StatTypeKey, StatType_type, StatTypeKeyHash >
      MapStatType_type;

      struct ADC_type
      {
	INT_4U	channelID;
	INT_4U	groupID;
	position_list_type	positionADC;
      };

      typedef General::hash_map< std::string, ADC_type >
      MapADC_type;
      typedef std::vector< STRING >				IndexADC_type;

      static const CHAR* getStaticName();


      typedef General::hash_map< std::string, position_type >
      MapDetector_type;
      typedef General::hash_map< std::string, position_list_type >
      MapProc_type;
      typedef std::vector< STRING >				IndexProc_type;
      typedef General::hash_map< std::string, position_list_type >
      MapSim_type;
      typedef std::vector< STRING >				IndexSim_type;
      typedef General::hash_map< std::string, position_list_type >
      MapSer_type;
      typedef std::vector< STRING >				IndexSer_type;
      typedef General::hash_map< std::string, position_list_type >
      MapSummary_type;

      struct Event_type
      {
	INT_4U		GTimeS;
	INT_4U		GTimeN;
	REAL_4		amplitude;
	position_type	position;

	inline Event_type( )
	{
	};

	inline Event_type( const General::GPSTime& GTime,
			REAL_4 	       Amplitude,
			position_type     Position )
	  : GTimeS( GTime.GetSeconds( ) ),
	    GTimeN( GTime.GetNanoseconds( ) ),
	    amplitude( Amplitude ),
	    position( Position )
	{
	}
      };

      typedef std::vector< Event_type >	EventList_type;

      typedef General::hash_map< std::string, EventList_type >
      MapEventType_type;
      typedef General::hash_map< std::string, EventList_type >
      MapSimEventType_type;

      FrTOC( );

      ~FrTOC( );

      void BuildDynamically( IFrameStream& Stream );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      std::string GetClassName( INT_2U ClassId ) const;

      const std::vector< INT_4U >&	GetChannelID( ) const;
      const std::vector< INT_4U >&	GetDataQuality( ) const;
      const std::vector< REAL_8 >&	GetDt( ) const;
      const std::vector< INT_4U >&	GetFrame( ) const;
      const std::vector< INT_4U >&	GetGroupID( ) const;
      const std::vector< INT_4U >&	GetGTimeN( ) const;
      const std::vector< INT_4U >&	GetGTimeS( ) const;
      const std::vector< STRING >&	GetName( ) const;
      INT_4U				GetNFrame( ) const;
      const position_list_type&		GetNFirstADC( ) const;
      const position_list_type&		GetNFirstMsg( ) const;
      const position_list_type&		GetNFirstSer( ) const;
      const position_list_type&		GetNFirstTable( ) const;
      INT_4U				GetNSH( ) const;
      INT_4U				GetNStatType( ) const;
      const position_dbl_list_type&	GetPositionADC( ) const;
      const position_list_type&		GetPositionH( ) const;
      const std::vector< INT_4S >&	GetRuns( ) const;
      const std::vector< INT_2U >&	GetSHid( ) const;
      const std::vector< STRING >&	GetSHname( ) const;
      const MapStatType_type&		GetStatType( ) const;
      INT_2S				GetULeapS( ) const;

      //:TODO: Put these into the correct place
      INT_4U				GetNADC( ) const;
      const MapADC_type&			GetADC( ) const;
      MapADC_type::const_iterator 		GetADC( const std::string& Name ) const;
      MapADC_type::const_iterator 		GetADC( INT_4U index ) const;
      INT_4U				GetNDetector( ) const;
      const MapDetector_type& 		GetDetector( ) const;
      INT_4U				GetNProc( ) const;
      const MapProc_type& 			GetProc( ) const;
      MapProc_type::const_iterator 	GetProc( const std::string& Name ) const;
      MapProc_type::const_iterator 	GetProc( INT_4U index ) const;

      INT_4U				GetNSim( ) const;
      const MapSim_type& 		GetSim( ) const;
      MapSim_type::const_iterator 	GetSim( const std::string& Name ) const;
      MapSim_type::const_iterator 	GetSim( INT_4U index ) const;

      INT_4U				GetNSer( ) const;
      const MapSer_type& 		GetSer( ) const;
      MapSer_type::const_iterator 	GetSer( const std::string& Name ) const;
      MapSer_type::const_iterator 	GetSer( INT_4U index ) const;
      INT_4U				GetNSummary( ) const;
      const MapSummary_type& 		GetSummary( ) const;
      
      INT_4U				GetNEventType( ) const;
      const MapEventType_type& 		GetEventType( ) const;
      MapEventType_type::const_iterator 	GetEventType( const std::string& Name ) const;

      INT_4U				GetNSimEventType( ) const;
      const MapSimEventType_type& 	GetSimEventType( ) const;
      MapSimEventType_type::const_iterator 	GetSimEventType( const std::string& Name ) const;
      //:TODO: End - Put these into the correct place

      void RegisterCallbacks( Dictionary& Dictionary, OFrameStream& Stream );

      void RegisterCallbacks( Dictionary& Dictionary, IFrameStream& Stream );

      void UnregisterCallbacks( );

      //: Read the structure from stream
      static FrTOC* ReadDynamic( IFrameStream& Stream,
				 const FrBase& Base );

      virtual void Write( OFrameStream& Stream ) const;

      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      friend class IFrameStream;
      friend class TOCCallback;

      INT_2S			m_ULeapS;
      std::vector< INT_4U >	m_dataQuality;
      std::vector< INT_4U >	m_GTimeS;
      std::vector< INT_4U >	m_GTimeN;
      std::vector< REAL_8 >	m_dt;
      std::vector< INT_4S >	m_runs;
      std::vector< INT_4U >	m_frame;
      position_list_type		m_positionH;
      position_list_type		m_nFirstADC;
      position_list_type		m_nFirstSer;
      position_list_type		m_nFirstTable;
      position_list_type		m_nFirstMsg;

      std::vector< INT_2U >	m_SHid;
      std::vector< STRING >	m_SHname;

      MapDetector_type		m_Detector;

      MapStatType_type		m_StatType;
      MapADC_type			m_ADC;
      IndexADC_type		m_ADCIndex;

      //:TODO: Put these into the correct place
      MapProc_type		m_Proc;
      IndexProc_type		m_ProcIndex;
      MapSim_type		m_Sim;
      IndexSim_type		m_SimIndex;
      MapSer_type		m_Ser;
      IndexSer_type		m_SerIndex;
      MapSummary_type		m_Summary;
      MapEventType_type		m_EventType;
      MapSimEventType_type	m_SimEventType;
      //:TODO: End - Put these into the correct place

      Dictionary*				m_dictionary;
      IFrameStream*				m_input_stream;
      OFrameStream*				m_output_stream;
      std::vector< Dictionary::callback_key_type >	m_callback_keys;
      Dictionary::Callback*			m_callback;

      void read( IFrameStream& Stream );

      void register_callbacks( );
    };

    inline bool FrTOC::StatTypeKey::
    operator!=( const StatTypeKey& RHS ) const
    {
      return ! ( *this == RHS );
    }

    inline const std::vector< INT_4U >& FrTOC::
    GetDataQuality( ) const
    {
      return m_dataQuality;
    }

    inline const std::vector< REAL_8 >& FrTOC::
    GetDt( ) const
    {
      return m_dt;
    }

    inline const std::vector< INT_4U >& FrTOC::
    GetFrame( ) const
    {
      return m_frame;
    }

    inline const std::vector< INT_4U >& FrTOC::
    GetGTimeN( ) const
    {
      return m_GTimeN;
    }

    inline const std::vector< INT_4U >& FrTOC::
    GetGTimeS( ) const
    {
      return m_GTimeS;
    }

    inline INT_4U FrTOC::
    GetNADC( ) const
    {
      return m_ADC.size( );
    }

    inline const FrTOC::MapADC_type& FrTOC::
    GetADC( ) const
    {
      return m_ADC;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstADC( ) const
    {
      return m_nFirstADC;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstMsg( ) const
    {
      return m_nFirstMsg;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstSer( ) const
    {
      return m_nFirstSer;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstTable( ) const
    {
      return m_nFirstTable;
    }

    inline INT_4U FrTOC::
    GetNFrame( ) const
    {
      return m_dataQuality.size( );
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetPositionH( ) const
    {
      return m_positionH;
    }

    inline const std::vector< INT_4S >& FrTOC::
    GetRuns( ) const
    {
      return m_runs;
    }

    inline INT_4U FrTOC::
    GetNSH( ) const
    {
      return m_SHid.size( );
    }

    inline INT_4U FrTOC::
    GetNStatType( ) const
    {
      return m_StatType.size( );
    }

    inline const std::vector< INT_2U >& FrTOC::
    GetSHid( ) const
    {
      return m_SHid;
    }

    inline const std::vector< STRING >& FrTOC::
    GetSHname( ) const
    {
      return m_SHname;
    }

    inline const FrTOC::MapStatType_type& FrTOC::
    GetStatType( ) const
    {
      return m_StatType;
    }

    inline INT_2S FrTOC::
    GetULeapS( ) const
    {
      return m_ULeapS;
    }

    //:TODO: Put these into there proper location
    inline INT_4U FrTOC::
    GetNDetector( ) const
    {
      return m_Detector.size( );
    }

    inline const FrTOC::MapDetector_type& FrTOC::
    GetDetector( ) const
    {
      return m_Detector;
    }

    inline INT_4U FrTOC::
    GetNProc( ) const
    {
      return m_Proc.size( );
    }

    inline const FrTOC::MapProc_type& FrTOC::
    GetProc( ) const
    {
      return m_Proc;
    }

    inline INT_4U FrTOC::
    GetNSim( ) const
    {
      return m_Sim.size( );
    }
    inline const FrTOC::MapSim_type& FrTOC::
    GetSim( ) const
    {
      return m_Sim;
    }

    inline INT_4U FrTOC::
    GetNSer( ) const
    {
      return m_Ser.size( );
    }
    inline const FrTOC::MapSer_type& FrTOC::
    GetSer( ) const
    {
      return m_Ser;
    }

    inline INT_4U FrTOC::
    GetNSummary( ) const
    {
      return m_Summary.size( );
    }
    inline const FrTOC::MapSummary_type& FrTOC::
    GetSummary( ) const
    {
      return m_Summary;
    }

    inline INT_4U FrTOC::
    GetNEventType( ) const
    {
      return m_EventType.size( );
    }
    inline const FrTOC::MapEventType_type& FrTOC::
    GetEventType( ) const
    {
      return m_EventType;
    }

    inline INT_4U FrTOC::
    GetNSimEventType( ) const
    {
      return m_SimEventType.size( );
    }
    inline const FrTOC::MapSimEventType_type& FrTOC::
    GetSimEventType( ) const
    {
      return m_SimEventType;
    }

    //:TODO: End - Put these into there proper location

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_FrTOC_HH */
