#include "framecpp/config.h"

#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

namespace {
  class MapTableRef
    : public IFrameStream::MapRef< FrameCPP::Version_3_4_5::Table, FrTable >
  {
  public:
    MapTableRef( IFrameStream& Stream,
		 const FrameCPP::Version_3_4_5::Table* Table5,
		 FrTable& Table6 )
      : IFrameStream::MapRef< FrameCPP::Version_3_4_5::Table, FrTable >( Stream, Table5, Table6 )
    {
    }

  protected:
   
    // virtual
    void operator()( const void* Ref5Ptr,
                     const PTR_STRUCT& PtrStruct );

  }; // class - MapTableRef
   
   
  void MapTableRef::operator()( const void* Ref5Ptr,
			        const PTR_STRUCT& PtrStruct )
  {
      if ( Ref5Ptr == &(m_struct_5->refColumns( ) ) )
      {
	m_stream.AddContainer( m_struct_6.RefColumn( ), PtrStruct );
      }
   
      return;
  }   
   
} // namespace - anonymous


const CHAR* FrTable::getStaticName()
{
   static const CHAR* class_name( "FrTable" );
   return class_name;
}


FrTable::
FrTable( )
  : FrBase( CLASS_FR_TABLE )
{
}
FrTable::
FrTable( const FrTable& table )
  : FrBase( CLASS_FR_TABLE )
{
  m_data.name = table.m_data.name;
  m_data.comment = table.m_data.comment;
  m_data.nRow = table.m_data.nRow;
  m_data.column = table.m_data.column;
}

FrTable::
FrTable( const std::string& name, INT_4U nrows )
  : FrBase( CLASS_FR_TABLE )
{
  m_data.name = name;
  m_data.nRow = nrows;
}


const std::string& FrTable::GetName( ) const
{
   return m_data.name;
}


void FrTable::AppendComment( const std::string& comment )
{
   return FrameCPP::Version_6::AppendComment( m_data.comment, comment );
}


INT_8U FrTable::
Bytes( ) const
{
  INT_8U ret( FrBase::Bytes( ) );

  ret +=
    m_data.name.Bytes( ) +
    m_data.comment.Bytes( ) +
    sizeof( INT_2U ) + // nColumn
    sizeof( m_data.nRow ) +
    PTR_STRUCT::Bytes( ) + 	// column
    PTR_STRUCT::Bytes( )	// next
    ;
  for ( column_type::const_iterator c( m_data.column.begin( ) );
	c != m_data.column.end( );
	c++ )
  {
    ret += STRING::Bytes( (*c)->GetName( ) );
  }
  return ret;
}

FrTable& FrTable::
Merge( const FrTable& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrTable* FrTable::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrTable > ret( new FrTable );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrTable::
Write( OFrameStream& Stream ) const
{
  INT_2U nColumn( m_data.columnName.size( ) );

  Stream << m_data.name
	 << m_data.comment
	 << nColumn
	 << m_data.nRow
	 << m_data.columnName
	 << m_data.column.containerCast( )
	 << GetNext( )
    ;
}

void FrTable::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_TABLE,
				"Table  Data Structure" ) );

  Stream.WriteDictionary( FrSE( "name", "STRING", "") );
  Stream.WriteDictionary( FrSE( "comment", "STRING", "") );
  Stream.WriteDictionary( FrSE( "nColumn", "INT_2U", "") );
  Stream.WriteDictionary( FrSE( "nRow", "INT_4U", "") );
  Stream.WriteDictionary( FrSE( "columnName", "STRING[nColumn]", "") );
  Stream.WriteDictionary( FrSE( "column",
				PTR_STRUCT::Desc( FrVect::getStaticName() ) ) );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );

  return;
}

FrTable& FrTable::
operator+=( const FrTable& RHS )
{
  //---------------------------------------------------------------------
  //:TODO: Sanity checks
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  // Return the concatinated table
  //---------------------------------------------------------------------
  return *this;
}

void FrTable::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< Table >	ver5;
  PTR_STRUCT			next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      INT_2U	nColumn;

      Stream >> m_data.name
	     >> m_data.comment
	     >> nColumn
	     >> m_data.nRow;
      Stream.ReadVector( m_data.columnName, nColumn );
      Stream >> m_data.column.containerCast( )
	     >> next;
    }
    break;
  case 5:
    ver5.reset( Table::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( Table::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->getName( );
    m_data.comment = ver5->getComment( );
    m_data.nRow = ver5->getNRows( );

    MapTableRef	map_table( Stream, ver5.get( ), *this );
    map_table.Map();
  }
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}
