#ifndef FrameCPP_VERSION_6_FrTable_HH
#define FrameCPP_VERSION_6_FrTable_HH

#include <vector>

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/Dimension.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrTable: public FrBase
    {
    public:
      typedef SearchContainer< FrVect, &FrVect::GetName >	column_type;

      typedef column_type::iterator iterator;
      typedef column_type::const_iterator const_iterator;

      static const CHAR* getStaticName();

      //: Default constructor
      FrTable( );
      FrTable( const FrTable& table );
      FrTable( const std::string& name, INT_4U nrows );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      const std::string& GetName( ) const;
   
      const std::string& GetComment() const
      {
         return m_data.comment;
      }   
   
      INT_2U GetNColumn() const
      {
         return RefColumn( ).size( );
      }
   

      INT_4U GetNRow() const
      {
         return m_data.nRow;
      }      

      /* Mutators */
      void AppendComment( const std::string& comment );

      const column_type&	RefColumn( ) const
      {
         return m_data.column;
      }
   
      column_type&		RefColumn( )
      {
         return m_data.column;
      }

      //: Merge with another FrAdcData
      FrTable& Merge( const FrTable& RHS );

      //: Read the structure from stream
      static FrTable* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

      //: Concatination operator
      FrTable& operator+=( const FrTable& RHS );

    private:
      struct {
	STRING			name;
	STRING			comment;
	INT_4U			nRow;
	std::vector<STRING>	columnName;
	column_type		column;
      } m_data;

      void read( IFrameStream& Stream );

#if ENABLE_DEPRICATED_FRAMECPP
    public:
      typedef column_type ColumnContainer;

      const std::string& getName( ) const
      {
	return GetName( );
      }
      const std::string& getComment() const
      {
	return GetComment( );
      }
      INT_2U getNColumns() const
      {
	return GetNColumn( );
      }
      INT_4U getNRows() const
      {
	return GetNRow( );
      }

      /* Mutators */
      void appendComment( const std::string& comment )
      {
	return AppendComment( comment );
      }

      const ColumnContainer&	refColumns( ) const
      {
	return RefColumn( );
      }
      ColumnContainer&		refColumns( )
      {
	return RefColumn( );
      }
#endif /* ENABLE_DEPRICATED_FRAMECPP */
    };

  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrTable_HH */

