#ifndef FrameCPP_VERSION_6_FrVect_HH
#define FrameCPP_VERSION_6_FrVect_HH

#include <memory>
#include <vector>
#include <string>

#include "general/types.hh"
#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/Compression.hh"

#include "framecpp/Version6/Dimension.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/GPSTime.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrVect
      : public FrBase
    {
    public:
      // using FrameCPP::Compression;

      enum compression_scheme_type {
	BIGENDIAN_RAW
	  = Compression::MODE_RAW,
	BIGENDIAN_GZIP
	  = Compression::MODE_GZIP,
	BIGENDIAN_DIFF
	  = Compression::MODE_DIFF,
        BIGENDIAN_DIFF_GZIP
	  = Compression::MODE_DIFF_GZIP,
	BIGENDIAN_ZERO_SUPPRESS_SHORT
	  = Compression::MODE_ZERO_SUPPRESS_SHORT,
	BIGENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER
	  = Compression::MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER,
	BIGENDIAN_ZERO_SUPPRESS_INT_FLOAT
	  = Compression::MODE_ZERO_SUPPRESS_INT_FLOAT,
	BIGENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP
	  = Compression::MODE_ZERO_SUPPRESS_OTHERWISE_GZIP,
	//---------------------------------------------------------
	LITTLEENDIAN_RAW
	  = 256 + BIGENDIAN_RAW,
	LITTLEENDIAN_GZIP
	  = 256 + BIGENDIAN_GZIP,
	LITTLEENDIAN_DIFF
	  = 256 + BIGENDIAN_DIFF,
        LITTLEENDIAN_DIFF_GZIP
          = 256 + BIGENDIAN_DIFF_GZIP,
	LITTLEENDIAN_ZERO_SUPPRESS_SHORT
	  = 256 + BIGENDIAN_ZERO_SUPPRESS_SHORT,
	LITTLEENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER
	  = 256 + BIGENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER,
	LITTLEENDIAN_ZERO_SUPPRESS_INT_FLOAT
	  = 256 + BIGENDIAN_ZERO_SUPPRESS_INT_FLOAT,
	LITTLEENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP
	  = 256 + BIGENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP,
#if WORDS_BIGENDIAN
	RAW = BIGENDIAN_RAW,
	GZIP = BIGENDIAN_GZIP,
	DIFF = BIGENDIAN_DIFF,
        DIFF_GZIP = BIGENDIAN_DIFF_GZIP,
	ZERO_SUPPRESS_SHORT = BIGENDIAN_ZERO_SUPPRESS_SHORT,
	ZERO_SUPRESS_SHORT_GZIP_OTHER =
	  BIGENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER,
	ZERO_SUPPRESS_INT_FLOAT = BIGENDIAN_ZERO_SUPPRESS_INT_FLOAT,
	ZERO_SUPPRESS_OTHERWISE_GZIP
	  = BIGENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP,
#else // WORDS_BIGENDIAN
	RAW = LITTLEENDIAN_RAW,
	GZIP = LITTLEENDIAN_GZIP,
	DIFF = LITTLEENDIAN_DIFF,
        DIFF_GZIP = LITTLEENDIAN_DIFF_GZIP,
	ZERO_SUPPRESS_SHORT = LITTLEENDIAN_ZERO_SUPPRESS_SHORT,
	ZERO_SUPRESS_SHORT_GZIP_OTHER =
	  LITTLEENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER,
	ZERO_SUPPRESS_INT_FLOAT = LITTLEENDIAN_ZERO_SUPPRESS_INT_FLOAT,
	ZERO_SUPPRESS_OTHERWISE_GZIP
	  = LITTLEENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP,
#endif // WORDS_BIGENDIAN
#if ENABLE_DEPRICATED_FRAMECPP
	DIFF_ZERO_SUPPRESS_SHORT = ZERO_SUPPRESS_SHORT,
	GZIP_DIFF = DIFF_GZIP,
#endif /* ENABLE_DEPRICATED_FRAMECPP */
	NONE = RAW
      };

      enum data_types_type {
	FR_VECT_C = 0,
	FR_VECT_2S = 1,
	FR_VECT_8R = 2,
	FR_VECT_4R = 3,
	FR_VECT_4S = 4,
	FR_VECT_8S = 5,
	FR_VECT_8C = 6,
	FR_VECT_16C = 7,
	FR_VECT_STRING = 8,
	FR_VECT_2U = 9,
	FR_VECT_4U = 10,
	FR_VECT_8U = 11,
	FR_VECT_1U = 12
      };


      static const CHAR* getStaticName();

      static const int DEFAULT_GZIP_LEVEL;

      //: Default constructor
      FrVect( );

      //-----------------------------------------------------------------
      //
      //: Copy Constructor.  
      //
      //!param: const Vect& vect - Copied vector.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const FrVect& vect );


      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+       containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data. Default
      //+       is BIGENDIAN.   
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //    
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      const void* data = 0, 
	      const std::string& unitY = "" );

      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+     containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data.
      //+	Default is BIGENDIAN.      
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //!param: bool allocate - Whether to allocate data. Default: true.
      //!param: bool owns - Whether to destruct data in the destructor.
      //+	Default: true.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      void* data = 0,
	      const std::string& unitY = "", 
	      bool allocate = true, bool owns = true );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      const T* data = 0, const std::string& unitY = "" );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      T* data = 0, const std::string& unitY = "",
	      bool allocate = true, bool owns = true );

      virtual ~FrVect( );

      //: Comparison operator
      bool operator==( const FrVect& RHS ) const;
      bool operator!=( const FrVect& RHS ) const;
      //: Concatination operator
      FrVect& operator+=( const FrVect& RHS );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      // Accessors
      const std::string& GetName( ) const;
      INT_2U GetCompress( ) const;
      INT_2U GetType( ) const;
      INT_8U GetNData( ) const;
      INT_8U GetNBytes( ) const;
      const CHAR_U* GetData( bool Decompress = true ) const;
      CHAR_U* GetData( bool Decompress = true );
      template< class T > static INT_2U GetDataType();
      INT_4U GetNDim( ) const;
      const Dimension& GetDim( INT_4U Offset ) const;
      Dimension& GetDim( INT_4U Offset );
      const std::string GetUnitY( ) const;

      // Mutators
      void Compress( compression_scheme_type Scheme, int GZipLevel );
      void Uncompress( );
      void SetNData( INT_4U NData );

      size_t GetTypeSize( ) const;
      static size_t GetTypeSize( INT_2U type );

      //: Merge with another FrAdcData
      FrVect& Merge( const FrVect& RHS );

      //: Read the structure from stream
      static FrVect* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      std::auto_ptr< FrVect > SubFrVect( INT_4U Start, INT_4U Stop ) const;

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:

      struct data_definition_type
      {
      public:

	STRING				name;
	INT_2U				compress;
	INT_2U				type;
	INT_8U				nData;
	INT_8U				nBytes;
	CHAR_U*				data;
	std::vector< Dimension >	dimension;
	STRING				unitY;

        data_definition_type();
        data_definition_type( const std::string& n, INT_2U t, 
			      const byte_order_type byte_order, INT_4U ndim,
			      const Dimension* dims, const std::string& unit );

#if ENABLE_DEPRICATED_FRAMECPP
        data_definition_type( const std::string& n, INT_2U t, 
			      const byte_order_type byte_order, INT_4U ndim,
			      const INT_4U* dims, const std::string& unit );
#endif

	void copy_core( const data_definition_type& Source );

	bool operator==( const data_definition_type& RHS ) const;

      };
 
      data_definition_type	m_data;
      bool			m_owns;

      void copy_data( CHAR_U* Data, INT_8U size,
		      bool allocate, bool own );

      void read( IFrameStream& Stream );

#if ENABLE_DEPRICATED_FRAMECPP
    public:
      enum {
	BIG = ::FrameCPP::BYTE_ORDER_BIG_ENDIAN,
	LITTLE = ::FrameCPP::BYTE_ORDER_LITTLE_ENDIAN
      };

      typedef byte_order_type ByteOrder;

      typedef data_types_type Type;

      static const ByteOrder HOST_ORDER;

      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
      //+     containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data. Default
      //+	is BIGENDIAN.      
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const INT_4U* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      const void* data = 0, 
	      const std::string& unitY = "" );

      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
      //+     containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data. Default
      //+	is BIGENDIAN.      
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //!param: bool allocate - Whether to allocate data. Default: true.
      //!param: bool owns - Whether to destruct data in the destructor.
      //+	Default: true.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const INT_4U* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      void* data = 0, 
	      const std::string& unitY = "",
	      bool allocate = true, bool owns = true );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1, const INT_4U* dims = 0,
	      const T* data = 0, const std::string& unitY = "" );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1, const INT_4U* dims = 0,
	      T* data = 0, const std::string& unitY = "",
	      bool allocate = true, bool owns = true );
    
      const std::string& getName( ) const;
      INT_2U getCompress( ) const;
      INT_2U getType( ) const;
      INT_8U getNData( ) const;
      INT_8U getNBytes( ) const;
      const CHAR_U* getData( ) const;
      CHAR_U* getData( );
      INT_4U getNDim( ) const;
      const Dimension& getDimension( INT_4U Offset ) const;
      Dimension& getDimension( INT_4U Offset );
      const std::string getUnitY( ) const;
      static size_t getTypeSize( INT_2U type );
      void expand();
      void setNData( INT_4U NData );

      template< class T > static INT_2U getDataType( );

#endif // ENABLE_DEPRICATED_FRAMECPP
    };


    inline INT_2U FrVect::
    GetCompress( ) const
    {
      return m_data.compress;
    }

    inline INT_2U FrVect::
    GetType( ) const
    {
      return m_data.type;
    }

    inline size_t FrVect::
    GetTypeSize( ) const
    {
      return GetTypeSize( m_data.type );
    }

    inline INT_8U FrVect::
    GetNBytes( ) const
    {
      return m_data.nBytes;
    }

    inline INT_4U FrVect::GetNDim( ) const
    {
      return m_data.dimension.size( );
    }

    inline const Dimension& FrVect::
    GetDim( INT_4U Offset ) const
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline Dimension& FrVect::
    GetDim( INT_4U Offset )
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline const std::string FrVect::
    GetUnitY( ) const
    {
      return m_data.unitY;
    }

    inline void FrVect::
    SetNData( INT_4U NData )
    {
      m_data.nData = NData;
    }
   
#if ENABLE_DEPRICATED_FRAMECPP
    inline const Dimension& FrVect::getDimension( INT_4U Offset ) const
    {
       return GetDim( Offset );
    }
    inline Dimension& FrVect::getDimension( INT_4U Offset )
    {
       return GetDim( Offset );
    }   

    inline INT_2U FrVect::getCompress( ) const
    {
      return GetCompress( );
    }

    inline const std::string& FrVect::getName() const
    {
      return GetName( );
    }
   
    inline INT_2U FrVect::getType( ) const
    {
	return GetType( );
    }
    inline INT_8U FrVect::getNData( ) const
    {
	return GetNData( );
    }
    inline INT_8U FrVect::getNBytes( ) const
    {
       return GetNBytes( );
    }
    inline const CHAR_U* FrVect::getData( ) const
    {
	return GetData( );
    }
    inline CHAR_U* FrVect::getData( )
    {
	return GetData( );
    }
    inline INT_4U FrVect::getNDim( ) const
    {
	return GetNDim( );
    }
   
    inline const std::string FrVect::getUnitY( ) const
    {
	return GetUnitY( );
    }
    inline size_t FrVect::getTypeSize( INT_2U type )
    {
	return GetTypeSize( type );
    }
    inline void FrVect::setNData( INT_4U NData )
    {
	return SetNData( NData );
    }

#endif   
   
  } // namespace - Version_6
} // namespace FrameCPP

#if defined( __SUNPRO_CC ) && ( __SUNPRO_CC <= 0x550 )
#include "framecpp/Version6/FrVect.icc"
#endif /* */

#endif /* FrameCPP_VERSION_6_FrVect_HH */
