#ifndef FRAME_CPP__VERSION_6__REFERENCE_CC
#define FRAME_CPP__VERSION_6__REFERENCE_CC
/* -*- c-default-offset: 3 -*- */
#include "framecpp/config.h"

#include <cassert>

#include <algorithm>
#include <iostream>
#include <list>
#include <sstream>
#include <stdexcept>
#include <vector>

#include <iostream>
using std::endl;

#include "framecpp/Version6/Dictionary.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/Reference.hh"

#include "framecpp/ContainerBasic.hh"

using namespace FrameCPP::Version_6;
using std::runtime_error;   

#define LM_DEBUG 0
#define	LM_ASSERT 0

#if LM_ASSERT
#define ASSERT( a ) assert( a )
#else
#define ASSERT( a )
#endif

#if LM_DEBUG
#define ATM( LM_a ) std::cerr << "MESG: " << LM_a << std::endl
#define ATA( LM_a ) std::cerr << "INFO: " << LM_a << " " << __FILE__ << " " << __LINE__ << std::endl
#define AT(  ) ATA( "" )
#else
#define ATM( LM_a )
#define	ATA( LM_a )
#define AT( )
#endif

namespace FrameCPP
{
  namespace Version_6
  {
    namespace hidden
    {
      //---------------------------------------------------------------------
      // Classes only needed for this file
      //---------------------------------------------------------------------
      class RefHeadBase
      {
      public:
	RefHeadBase();
	virtual ~RefHeadBase();

	virtual void addItem( FrBase* b ) = 0;
      };

      template< class T >
      class RefHead : public RefHeadBase
      {
      public:

	RefHead( FrameCPP::ContainerBasic< T >& c );
      
	RefHead( const RefHead& ref );
  
	void addItem( FrBase* b );
   
	INLINE_KLUDGE const FrameCPP::ContainerBasic< T >* getContainer( ) const
	{
	  return &m_container;
	}

      private:

	FrameCPP::ContainerBasic< T >& m_container;
      };

      INLINE_KLUDGE RefHeadBase::
      RefHeadBase()
      {
      }

      INLINE_KLUDGE RefHeadBase::
      ~RefHeadBase()
      {
      }

      template< class T > INLINE_KLUDGE RefHead< T >::
      RefHead( FrameCPP::ContainerBasic< T >& c )
	:m_container( c )
      {
      }

      template< class T > INLINE_KLUDGE RefHead< T >::
      RefHead( const RefHead& ref )
	:m_container( ref.m_container )
      {
      }

      template< class T > INLINE_KLUDGE void RefHead< T >::
      addItem( FrBase* b )
      {
	T* p = dynamic_cast< T* >( b );
	if ( p )
	{
	  if ( b->GetClassId( ) == CLASS_FR_DETECTOR )
	  {
	    m_container.append( p, false /* allocate */, false /* own */ );
	  }
	  else
	  {
	    m_container.append( p, false /* allocate */, true /* own */ );
	  }
	}
	return;
      }   

      //---------------------------------------------------------------------
      // Typedefs to make life and typing easier
      //---------------------------------------------------------------------
      typedef General::hash_map< INT_4U, FrBase* > object_map_element_type;
      typedef std::vector< object_map_element_type* > object_map_type;
      typedef General::hash_map< INT_4U, PTR_STRUCT > next_element_type;
      typedef std::vector< next_element_type > next_map_type;
      typedef std::pair< RefHeadBase*, PTR_STRUCT > container_head_element_type;
      typedef std::list< container_head_element_type > container_head_list_type;
      typedef std::pair< FrBase**, PTR_STRUCT > pointer_element_type;
      typedef std::list< pointer_element_type > pointer_list_type;

#undef INLINE_KLUDGE

    }
  }
}
//-----------------------------------------------------------------------
// Declaration of Reference
//-----------------------------------------------------------------------

using namespace FrameCPP::Version_6::hidden;

class Reference::private_type
{
public:
  object_map_type			m_objects;
  container_head_list_type		m_container_heads;
  pointer_list_type			m_pointers;
  INT_4U				m_frame_version;
  INT_4U				m_frame_library_revision;
  bool					m_recursive_mode;
  std::list< PTR_STRUCT >		m_recursive_objects;
  std::list< PTR_STRUCT >::iterator	m_recursive_insert_pos;
  bool					m_tracking;

  //: Default Constructor
  private_type( const INT_4U num_objects );

  ~private_type();
  void DumpCache( ) const;
  void push_ref_for_recursion( const PTR_STRUCT& Ref );
  void remove_ref_for_recursion( const PTR_STRUCT& Ref );
};


Reference::
Reference( INT_4U ClassCount, const Dictionary& Dictionary )
  : m_dictionary( Dictionary ),
    m_class_count( ClassCount ),
    p( new private_type( ClassCount ) )
{
  ASSERT( &m_dictionary );
  ASSERT( p->m_objects.size( ) == m_class_count );
}

Reference::
~Reference( )
{
  AT( );
  ASSERT( &m_dictionary );
  ASSERT( p->m_objects.size( ) == m_class_count );
  Reset( true /* PurgeDetectors */ );
  ASSERT( p->m_objects.size( ) == m_class_count );
  delete p;
  p = (private_type*)NULL;
  ASSERT( &m_dictionary );
}

void Reference::
AddNextRef( const PTR_STRUCT& PtrStruct )
{
  ASSERT( &m_dictionary );
  ASSERT( p->m_objects.size( ) == m_class_count );
  p->push_ref_for_recursion( PtrStruct );
  ASSERT( p->m_objects.size( ) == m_class_count );
  ASSERT( &m_dictionary );
}

void Reference::
AddObject( FrBase* Base )
{
  ASSERT( &m_dictionary );
  ASSERT( p->m_objects.size( ) == m_class_count );
  if ( p->m_tracking )
  {
    ASSERT( &m_dictionary );
    ASSERT( p->m_objects.size( ) == m_class_count );
    switch( Base->GetClassId( ) )
    {
    case FrBase::CLASS_NULL:
    case FrBase::CLASS_FR_SE:
    case FrBase::CLASS_FR_SH:
      break;
    default:
      {
	const INT_4U class_id( Base->GetClassId() );
	ASSERT( class_id < p->m_objects.size( ) );
      
	if( class_id >= p->m_objects.size() )
	{
	  throw runtime_error( "Attempt to modify Reference::private_type::m_objects when "
			       "index is larger than vector size." );
	}
   

	if( p->m_objects[ class_id ] == 0 )   
	{
	  p->m_objects[ class_id ] = new object_map_element_type;
	}

	INT_4U size = p->m_objects[ class_id ]->size( );
	( *( p->m_objects[ class_id ] ) )[ Base->GetInstance( ) ] = Base;
	if ( size == p->m_objects[ class_id ]->size( ) )
	{
	  if ( ( m_dictionary.
		 MapStreamClassToLibraryClass( Base->GetClassId( ) )
		 == CLASS_FR_STAT_DATA ) &&
	       ( p->m_frame_version == 6 ) &&
	       ( p->m_frame_library_revision == 15 ) )
	  {
	    //----------------------------------------------------------
	    // FrameL violates the frame spec until v6r18 (PR#3216)
	    //----------------------------------------------------------
	  }
	  else
	  {
	    std::ostringstream	msg;

	    msg << "Duplicate object found: Class: "
		<< GetClassName( class_id )
		<< " Instance: " << Base->GetInstance( );
	    throw std::runtime_error( msg.str( ) );
	  }
	}
   
	ATA( "Adding Object: " << Base \
	     << " ClassId: " << Base->GetClassId( ) \
	     << " Instance: " << Base->GetInstance( ) );
	ATA( "Object in list: " << (*( p->m_objects[ class_id ] ) )[ Base->GetInstance( ) ] );

	ASSERT( &m_dictionary );
	ASSERT( p->m_objects.size( ) == m_class_count );
	break;
      }
    }
  }
  ASSERT( &m_dictionary );
   
  return;
}

template< class T >
void Reference::
AddContainer( FrameCPP::ContainerBasic< T >& Container, const PTR_STRUCT& Ref )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  Container.erase( Container.begin( ), Container.end( ) );
  if ( ( Ref != NULL_PTR_STRUCT ) && ( p->m_tracking ) )
  {
    p->m_container_heads.push_back
      ( std::pair< hidden::RefHeadBase*, PTR_STRUCT >( new RefHead< T >( Container ),
					       Ref ) );
    ASSERT( p->m_container_heads.back( ).first != (RefHead< T >*)NULL );
    p->push_ref_for_recursion( Ref );
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
AddPointer( FrBase** Pointer, const PTR_STRUCT& Ref )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  *Pointer = (FrBase*)NULL;
  if ( ( Ref != NULL_PTR_STRUCT ) && ( p->m_tracking ) )
  {
    p->m_pointers.push_back( std::pair< FrBase**, PTR_STRUCT >
			     ( Pointer, Ref ) );
    p->push_ref_for_recursion( Ref );
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
}

PTR_STRUCT Reference::
GetReferenceToPointer( const FrBase** Pointer ) const
{

  ASSERT( p->m_objects.size( ) == m_class_count );
  for ( pointer_list_type::iterator i( p->m_pointers.begin( ) );
	i != p->m_pointers.end( );
	i++ )
  {
    if ( (*i).first == Pointer )
    {
      ASSERT( p->m_objects.size( ) == m_class_count );
      return (*i).second;
    }
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
  return NULL_PTR_STRUCT;
}

bool Reference::
GetRecursiveReference( PTR_STRUCT& Reference )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  INT_2U	lib_class;
  while ( p->m_recursive_objects.size( ) > 0 )
  {
    Reference = p->m_recursive_objects.front( );
    p->m_recursive_objects.pop_front( );
    p->m_recursive_insert_pos = p->m_recursive_objects.begin( );

    lib_class =
      m_dictionary.MapStreamClassToLibraryClass( Reference.data_class );
    ASSERT( lib_class < p->m_objects.size( ) );

    // There's no map created yet
    if( p->m_objects[ lib_class ] == 0 )
    {
      return true;
    }

    General::hash_map< INT_4U, FrBase* >::iterator
      obj( (p->m_objects[ lib_class ])->find( Reference.data_instance ) );

    ASSERT( Reference.data_class < ( p->m_objects[ lib_class ] )->size( ) );
    if ( obj != ( p->m_objects[ lib_class ] )->end( ) )
    {
      // Already have read in the object. Look for the next one
      continue;
    }

    return true;
  }


  Reference = NULL_PTR_STRUCT;
  ASSERT( p->m_objects.size( ) == m_class_count );
  return false;
}

bool Reference::
IsRecursiveObject( const FrBase& Base ) const
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  if ( p->m_recursive_mode )
  {
    PTR_STRUCT	b( Base.GetClassId( ), Base.GetInstance( ) );
    for ( std::list< PTR_STRUCT >::const_iterator
	    ps( p->m_recursive_objects.begin( ) );
	  ps != p->m_recursive_objects.end( );
	  ps++
	  )
    {
      if ( *ps == b )
      {
	ASSERT( p->m_objects.size( ) == m_class_count );
	return true;
      }
    }
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
  return false;
}

template< class T >
void Reference::
RemoveContainer( FrameCPP::ContainerBasic< T >& Container )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  for ( container_head_list_type::iterator i( p->m_container_heads.begin( ) );
	i != p->m_container_heads.end( );
	i++ )
  {
    RefHead< T >*	refhead( dynamic_cast< RefHead< T >* >( (*i).first ) );

    if ( ( refhead ) && ( refhead->getContainer( ) == &Container ) )
    {
      delete refhead;   
      refhead = 0;
      (*i).first = (RefHead< T >*)NULL;
   
      p->remove_ref_for_recursion( (*i).second );
      p->m_container_heads.erase( i );

      ASSERT( p->m_objects.size( ) == m_class_count );
      return;
    }
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
RemovePointer( FrBase** Pointer )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  for ( pointer_list_type::iterator i( p->m_pointers.begin( ) );
	i != p->m_pointers.end( );
	i++ )
  {
    if ( (*i).first == Pointer )
    {
      p->remove_ref_for_recursion( (*i).second );
      p->m_pointers.erase( i );
      ASSERT( p->m_objects.size( ) == m_class_count );
      return;
    }
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
RemoveReference( const PTR_STRUCT& Object )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  ASSERT( &m_dictionary );
  ASSERT( Object.data_class < p->m_objects.size( ) );
  ATA( "Trying to remove: " << &Object );
   
  if( Object.data_class >= p->m_objects.size() )
  {
    throw runtime_error( "Reference::RemoveReference(): attempt to"
			 " access private_type->m_objects when index is"
			 " out of range." );
  }

  // Map wasn't created for the data class 
  if( p->m_objects[ Object.data_class ] == 0 )
  {
    ASSERT( &m_dictionary );
    return;
  }


  General::hash_map< INT_4U, FrBase* >::iterator
    obj( (p->m_objects[ Object.data_class ])->find( Object.data_instance ) );

  if ( obj != ( p->m_objects[ Object.data_class ] )->end( ) )
  {
    ATA( "Removing: 1: " << Object.data_class << "-" << Object.data_instance );
    ATA( "Removing: 2: " << obj->second->GetClassId( ) << "-" << obj->second->GetInstance( ) );
    ( p->m_objects[ Object.data_class ] )->erase( obj );
  }
  else
  {
    ATA( "Object not found: " << &Object );
  }
  ASSERT( &m_dictionary );
  ASSERT( p->m_objects.size( ) == m_class_count );

  return;
}

void Reference::
Reset( bool PurgeDetectors )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  AT( );
  const INT_4U	obj_size( p->m_objects.size( ) );

  for( INT_4U x = 0; x < obj_size; ++x )
  {
    if ( ( PurgeDetectors == false ) && ( x == CLASS_FR_DETECTOR ) )
    {
      continue;
    }

    // Map was created for class id
    if( p->m_objects[x] != 0 )
    {
      AT( );
      for( object_map_element_type::iterator
	     i = ( p->m_objects[x] )->begin(),
	     end_i = ( p->m_objects[x] )->end();
	   i != end_i;
	   ++i )
      {
	ATA( "x: " << x << " second: " << (*i).second );
	delete (*i).second;
	(*i).second = (FrBase*)NULL;
      }

      AT( );
      ( p->m_objects[x] )->clear();
    }
  }
  //---------------------------------------------------------------------
  AT( );
  ASSERT( p->m_objects.size( ) == m_class_count );
  for ( container_head_list_type::iterator i( p->m_container_heads.begin( ) );
	i != p->m_container_heads.end( );
	i++ )
  {
    delete (*i).first;
    (*i).first = (RefHeadBase*)NULL;
  }
  AT( );
  p->m_container_heads.resize( 0 );
  AT( );
  ASSERT( p->m_objects.size( ) == m_class_count );
  //---------------------------------------------------------------------
  p->m_pointers.resize( 0 );
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
ResolveLinks( const IFrameStream& Stream )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  PTR_STRUCT			ref;

  p->DumpCache( );
  //---------------------------------------------------------------------
  // Work on reconnecting pointers
  //---------------------------------------------------------------------
  AT( );
  for ( pointer_list_type::iterator i( p->m_pointers.begin( ) );
	i != p->m_pointers.end( );
	i++ )
  {
    AT( );
    ref = (*i).second;
    if ( ref.data_class == NULL_PTR_STRUCT.data_class )
    {
      AT( );
      *((*i).first) = (FrBase*)NULL;
    }
    else
    {
      AT( );
   
      ref.data_class =
	m_dictionary.MapStreamClassToLibraryClass( ref.data_class );
   
      if( ref.data_class >= p->m_objects.size() )
      {
	throw runtime_error( "Reference::ResolveLinks() FOR loop: attempt to"
			     " access private_type->m_objects when index is"
			     " out of range." );
      }   
   
      // Map was created for class id
      if( p->m_objects[ ref.data_class ] != 0 )
      {
	General::hash_map< INT_4U, FrBase* >::iterator
	  obj( (p->m_objects[ ref.data_class ])->find( ref.data_instance ) );

	AT( );
	if ( obj != ( p->m_objects[ ref.data_class ] )->end( ) )
	{
	  AT( );
	  *((*i).first) = (*obj).second;

	  if ( ref.data_class != CLASS_FR_DETECTOR )
	  {
	    //-------------------------------------------------------------
	    // Consume object since it cannot be referenced anywhere else
	    //-------------------------------------------------------------
	    ATA( "Link Resolved (Pointer): " << ref.data_class << "-" << ref.data_instance );
	    p->m_objects[ ref.data_class ]->erase( obj );
	  }
	}
	else
	{
	  AT( );
	  // Object was not found; assign to null pointer
	  *((*i).first) = (FrBase*)NULL;
	}
      }
    }
  }

  p->m_pointers.resize( 0 );


  //---------------------------------------------------------------------
  // Work on reconnecting the containers
  //---------------------------------------------------------------------
  PTR_STRUCT			tmp_ref;
  INT_2U			lib_class;
  RefHeadBase*			channel_head;

  AT( );
  for ( container_head_list_type::iterator ch( p->m_container_heads.begin( ) );
	ch != p->m_container_heads.end( );
	ch++ )
  {
    ref = (*ch).second;

    while( ref.data_class != NULL_PTR_STRUCT.data_class )
    {
      AT( );
      // Convert from external stream class id to internal library class id
      lib_class = ref.data_class;
      ref.data_class =
	m_dictionary.MapStreamClassToLibraryClass( ref.data_class );
      
      AT( );
   
      if( ref.data_class >= p->m_objects.size() )
      {
	throw runtime_error( "Reference::ResolveLinks(): WHILE loop: attempt to"
			     " access private_type->m_objects when index is"
			     " out of range." );
      }   
   
      if( p->m_objects[ ref.data_class ] != 0 )
      {
	General::hash_map< INT_4U, FrBase* >::iterator
	  obj( (p->m_objects[ ref.data_class ])->find( ref.data_instance ) );
   
	if ( obj != ( p->m_objects[ ref.data_class ] )->end( ) )
	{
	  AT( );
	  channel_head = (*ch).first;
	  channel_head->addItem( (*obj).second );
	  tmp_ref = (*obj).second->GetNext( );
   
	  if ( ref.data_class != CLASS_FR_DETECTOR )
	  {
	    //-------------------------------------------------------------
	    // Consume object since it cannot be referenced anywhere else
	    //-------------------------------------------------------------
	    ATA( "Link Resolved (Container): " << ref.data_class << "-" << ref.data_instance );
	    ( p->m_objects[ ref.data_class ] )->erase( obj );
	  }

	  AT( );
	  ref = tmp_ref;
	}
	else
	{
	  p->DumpCache( );
	  ATA( "Lookup failed for: ptr: " << p->m_objects[ ref.data_class ] << " ref: " << ref.data_class << "-"<<  ref.data_instance );
	  std::ostringstream	oss;

	  oss << "Unable to resolve link: ClassId: ";
	  if ( Stream.GetTOC( ) )
	  {
	    oss << Stream.GetTOC( )->GetClassName( lib_class );
	  }
	  else
	  {
	    oss << lib_class;
	  }
	  oss << " Instance: " << ref.data_instance;
	  throw std::runtime_error( oss.str( ).c_str( ) );
	}
      }
    }

    delete (*ch).first;
    (*ch).first = (RefHeadBase*)NULL;
  }
  AT( );
   
  p->m_container_heads.clear();
  p->m_container_heads.resize( 0 );
   
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
SetFrameVersion( INT_4U FrameVersion )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  p->m_frame_version = FrameVersion;
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
SetFrameLibraryRevision( INT_4U Revision )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  p->m_frame_library_revision = Revision;
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
SetRecursiveMode( bool Mode )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  p->m_recursive_mode = Mode;
  StartNewObject( );
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
SetTracking( bool Tracking )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  p->m_tracking = Tracking;
  ASSERT( p->m_objects.size( ) == m_class_count );
}

void Reference::
StartNewObject( )
{
  ASSERT( p->m_objects.size( ) == m_class_count );
  if ( p->m_recursive_mode )
  {
    p->m_recursive_insert_pos = p->m_recursive_objects.begin( );
  }
  ASSERT( p->m_objects.size( ) == m_class_count );
}

Reference::private_type::
private_type( const INT_4U num_objects )
  : m_objects( num_objects, 0 ),
    m_container_heads( ),
    m_pointers( ),
    m_frame_version( DATA_FORMAT_VERSION ),
    m_frame_library_revision( 0 ),
    m_recursive_mode( false ),
    m_recursive_objects( ),
    m_tracking( true )
{
  m_recursive_insert_pos = m_recursive_objects.begin( );
}

void Reference::private_type::
DumpCache( ) const
{
#if LM_DEBUG
  ATM( "OBJECTS: " );
  for ( object_map_type::const_iterator
	  cur = m_objects.begin( ),
	  end = m_objects.end( );
	cur != end;
	++cur )
  {
    ATM( "  object: " << *cur );
    if ( *cur == (object_map_element_type*)NULL )
    {
      continue;
    }
    for ( object_map_element_type::const_iterator
	    e_cur = (*cur)->begin( ),
	    e_end = (*cur)->end( );
	  e_cur != e_end;
	  ++e_cur )
    {
      ATM( "    instance: " << (*e_cur).first << " base: " << (*e_cur).second );
    }
  }
  ATM( "POINTERS: " );
  for( pointer_list_type::const_iterator
	 cur = m_pointers.begin( ),
	 end = m_pointers.end( );
       cur != end;
       ++cur )
  {
    // ATM( "  object: " << *cur );
  }
#endif	/* LM_DEBUG */
}

Reference::private_type::
~private_type()
{
  for( object_map_type::iterator iter = m_objects.begin(), end_iter = m_objects.end();
       iter != end_iter; ++iter )
  {
    delete ( *iter );
    *iter = 0;
  }

  m_objects.clear();
}

void Reference::private_type::
push_ref_for_recursion( const PTR_STRUCT& Ref )
{
  if ( ( m_recursive_mode ) && ( Ref != NULL_PTR_STRUCT ) )
  {
    m_recursive_insert_pos = 
      m_recursive_objects.insert( m_recursive_insert_pos, Ref );
  }
}

void Reference::private_type::
remove_ref_for_recursion( const PTR_STRUCT& Ref )
{
  for ( std::list< PTR_STRUCT >::iterator i( m_recursive_objects.begin( ) );
	i != m_recursive_objects.end( );
	i++ )
  {
    if ( Ref == (*i) )
    {
      m_recursive_objects.erase( i );
   
      // Reset m_recursive_insert_pos to the beginning of the list: to be safe
      m_recursive_insert_pos = m_recursive_objects.begin();
      return;
    }
  }
}

#define INSTANTIATE( LM_TYPE )						\
  template void Reference::AddContainer( FrameCPP::ContainerBasic< LM_TYPE >&, const PTR_STRUCT& ) ; \
  template void Reference::RemoveContainer( FrameCPP::ContainerBasic< LM_TYPE >& )

#include "ContainerInstantiation.icc"

#undef INSTANTIATE
#endif /* FRAME_CPP__VERSION_6__REFERENCE_CC */
