#include "framecpp/config.h"

#include <vector>

#include "general/types.hh"

#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrameH.hh"
#include "framecpp/Version6/FrDetector.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrEndOfFrame.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/Functions.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/Util.hh"
#include "framecpp/Version6/STRING.hh"

namespace
{
  namespace Version6
  {
    template< class T >
    std::vector< T >
    vect( void* Data, INT_4U Size )
    {
      T* data( reinterpret_cast< T* >( Data ) );
      std::vector< T > ret( Size );
      std::copy( &data[0], &data[ Size ], ret.begin( ) );
      return ret;
    }

    class dump_label
    {
    public:
      dump_label(	std::ostream& Stream,
			const std::string& Padding,
			int Width,
			bool Scientific )
	: m_stream( Stream ),
	  m_padding( Padding ),
	  m_scientific( Scientific ),
	  m_width( Width )
	
      {
      }

      template< class T >
      void operator()( const std::string& Label , const T& Data )
      {
	int	size( m_width - Label.length( ) );

	m_stream << m_padding
		 << Label << ":";
	while ( size-- > 0 )
	{
	  m_stream << " ";
	}
	m_stream << Data << std::endl;
      }

      template< class T >
      void operator()( const std::string& Label , const std::vector< T >& Data )
      {
	int	size( m_width - Label.length( ) );

	m_stream << m_padding
	       << Label << ":";
	while ( size-- > 0 )
	{
	  m_stream << " ";
	}
	print_vect( Data );
	m_stream << std::endl;
      }

    private:
      std::ostream&	m_stream;
      const std::string&	m_padding;
      bool			m_scientific;
      int			m_width;

      template< class T >
      void print_vect( const std::vector< T >& Data );
    };

    template<>
    void dump_label::operator()( const std::string& Label,
				 const FrameCPP::Version_6::PTR_STRUCT& Data )
    {
      int size( m_width - Label.length( ) );

      m_stream << m_padding
	       << Label << ":";
      while ( size-- > 0 )
      {
	m_stream << " ";
      }
      m_stream << Data.data_class << "-" << Data.data_instance << std::endl;
    }

    template<>
    void dump_label::operator()( const std::string& Label,
				 const FrameCPP::Version_6::STRING& Data )
    {
      int size( m_width - Label.length( ) );

      m_stream << m_padding
	       << Label << ":";
      while ( size-- > 0 )
      {
	m_stream << " ";
      }
      m_stream << Data.c_str( ) << std::endl;
    }


    template<>
    void dump_label::print_vect( const std::vector< FrameCPP::Version_6::STRING >& Data )
    {
      for( std::vector< FrameCPP::Version_6::STRING >::const_iterator
	     d( Data.begin( ) );
	   d != Data.end( );
	   ++d )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << ( const std::string )( *d );
      }
    }


    template< >
    void dump_label::
    print_vect( const std::vector< CHAR >& Data )
    {
      for ( std::vector< CHAR >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << ( INT_2S)( *d );
      }

      return;
    }

    template< >
    void dump_label::
    print_vect( const std::vector< CHAR_U >& Data )
    {
      for ( std::vector< CHAR_U >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << ( INT_2U)( *d );
      }
    }

    template< >
    void dump_label::
    print_vect( const std::vector< REAL_4 >& Data )
    {
      std::ios_base::fmtflags
	f( m_stream.flags( ) );	// Save flags
      static const std::streamsize
	prec = REAL_4_DIGITS + 1;

      if ( m_scientific )
      {
	m_stream.precision( prec );
	m_stream.setf(std::ios::scientific, std::ios::floatfield);
      }
      for ( std::vector< REAL_4 >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << *d;
      }
      m_stream.flags( f ); // Restore flags
    }

    template< >
    void dump_label::
    print_vect( const std::vector< REAL_8 >& Data )
    {
      std::ios_base::fmtflags
	f( m_stream.flags( ) );	// Save flags
      static const std::streamsize
	prec = REAL_8_DIGITS + 1;

      if ( m_scientific )
      {
	m_stream.precision( prec );
	m_stream.setf(std::ios::scientific, std::ios::floatfield);
      }
      for ( std::vector< REAL_8 >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << *d;
      }
      m_stream.flags( f ); // Restore flags
    }

    template< >
    void dump_label::
    print_vect( const std::vector< COMPLEX_8 >& Data )
    {
      std::ios_base::fmtflags
	f( m_stream.flags( ) );	// Save flags
      static const std::streamsize
	prec = REAL_4_DIGITS + 1;

      if ( m_scientific )
      {
	m_stream.precision( prec );
	m_stream.setf(std::ios::scientific, std::ios::floatfield);
      }
      for ( std::vector< COMPLEX_8 >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << ( *d ).real( ) << "+i" << (*d).imag( );
      }
      m_stream.flags( f ); // Restore flags
    }

    template< >
    void dump_label::
    print_vect( const std::vector< COMPLEX_16 >& Data )
    {
      std::ios_base::fmtflags
	f( m_stream.flags( ) );	// Save flags
      static const std::streamsize
	prec = REAL_8_DIGITS + 1;

      if ( m_scientific )
      {
	m_stream.precision( prec );
	m_stream.setf(std::ios::scientific, std::ios::floatfield);
      }
      for ( std::vector< COMPLEX_16 >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << ( *d ).real( ) << "+i" << (*d).imag( );
      }
      m_stream.flags( f ); // Restore flags
    }

    template< class T >
    void dump_label::
    print_vect( const std::vector< T >& Data )
    {
      for ( typename std::vector< T >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	if ( d != Data.begin( ) )
	{
	  m_stream << " ";
	}
	m_stream << *d;
      }
    }

    class DumpFrStruct
    {
    public:

      enum OutputMode_type {
	OUTPUT_MODE_DUMP_OBJECTS,
	OUTPUT_MODE_DUMP
      };
      
      DumpFrStruct( bool VerboseDataMode,
		    bool ScientificDataMode,
		    OutputMode_type OutputMode )
	: m_output_mode( OutputMode ),
	  m_scientific_data_mode( ScientificDataMode ),
	  m_verbose_data_mode( VerboseDataMode )
      {
      }

      void Dump( const FrBase& Base );

      void SetVerbosity( bool Verbosity )
      {
	m_verbose_data_mode = Verbosity;
      }

      void SetScientific( bool Scientific )
      {
	m_scientific_data_mode = Scientific;
      }

    private:
      OutputMode_type		m_output_mode;
      std::string		m_pad;
      bool			m_scientific_data_mode;
      static const std::string	m_spaces;
      bool			m_verbose_data_mode;

      void header( const std::string& tag );

      void trailer( );

      inline void pad_inc( )
      {
	m_pad += m_spaces;
      }

      inline void pad_dec( )
      {
	m_pad.resize( m_pad.length( ) - m_spaces.length( ) );
      }

      template< class T >
      void dump_container( const std::string& Name, const T& Container );
      
    };

    template< >
    void DumpFrStruct::
    dump_container( const std::string& Name, const std::vector< int >& Container )
    {
      std::cout << m_pad << Name << ":" << std::endl;
      std::cout << m_pad << "{" << std::endl;
      std::cout << m_pad << "}" << std::endl;
    }

    template< class T >
    void DumpFrStruct::
    dump_container( const std::string& Name, const T& Container )
    {
      if ( Container.size( ) == 0 )
      {
	return;
      }
      std::cout << m_pad << Name << ":" << std::endl;
      std::cout << m_pad << "{" << std::endl;
      pad_inc( );
      for ( INT_4U x = 0; x < Container.size( ); x++ )
      {
	Dump( *( Container[ x ] ) );
      }
      pad_dec( );
      std::cout << m_pad << "}" << std::endl;
    }

    void DumpFrStruct::
    Dump( const FrBase& Base )
    {
      switch( Base.GetClassId( ) )
      {
      case FrBase::CLASS_FR_SH:
	{
	  const FrSH& o( dynamic_cast< const FrSH& >( Base ) );
	
	  header( "FrSH" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "class", o.GetClass( ) );
	  l( "comment", o.GetComment( ) );

	  trailer( );
	}
	break;
      case FrBase::CLASS_FR_SE:
	{
	  const FrSE& o( dynamic_cast< const FrSE& >( Base ) );
	
	  header( "FrSE" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "class", o.GetClass( ) );
	  l( "comment", o.GetComment( ) );

	  trailer( );
	}
	break;
      case CLASS_FR_ADC_DATA:
	{
	  const FrAdcData& o( dynamic_cast< const FrAdcData& >( Base ) );
	  
	  header( "FrAdcData" );
	
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "channelGroup", o.GetChannelGroup( ) );
	  l( "channelNumber", o.GetChannelNumber( ) );
	  l( "nBits", o.GetNBits( ) );
	  l( "bias", o.GetBias( ) );
	  l( "slope", o.GetSlope( ) );
	  l( "units", o.GetUnits( ) );
	  l( "sampleRate", o.GetSampleRate( ) );
	  l( "timeOffset", o.GetTimeOffset( ) );
	  l( "fShift", o.GetFShift( ) );
	  l( "phase", o.GetPhase( ) );
	  l( "dataValid", o.GetDataValid( ) );
	
	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "data", o.RefData( ) );
	    dump_container( "aux", o.RefAux( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FRAME_H:
	{
	  const FrameH& o( dynamic_cast< const FrameH& >( Base ) );
	  dump_label l( std::cout, m_pad, 12, m_scientific_data_mode );
	
	  header( "FrameH" );
	
	  l( "name", o.GetName( ) );
	  l( "run", o.GetRun( ) );
	  l( "frame", o.GetFrame( ) );
	  l( "dataQuality", o.GetDataQuality( ) );
	  l( "GTimeS", o.GetGTime( ).GetSeconds( ) );
	  l( "GTimeN", o.GetGTime( ).GetNanoseconds( ) );
	  l( "ULeapS", o.GetULeapS( ) );
	  l( "dt", o.GetDt( ) );

	  switch( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "type", o.RefType( ) );
	    dump_container( "user", o.RefUser( ) );
	    dump_container( "detectSim", o.RefDetectSim( ) );
	    dump_container( "detectProc", o.RefDetectProc( ) );
	    dump_container( "history", o.RefHistory( ) );
	    if ( o.GetRawData( ) )
	    {
	      Dump( *(o.GetRawData( ) ) );
	    }
	    else
	    {
	      std::cout << m_pad << "FrRawData - NULL" << std::endl;
	    }
	    dump_container( "procData", o.RefProcData( ) );
	    dump_container( "simData", o.RefSimData( ) );
	    dump_container( "event", o.RefEvent( ) );
	    dump_container( "simEvent", o.RefSimEvent( ) );
	    dump_container( "summaryData", o.RefSummaryData( ) );
	    dump_container( "auxData", o.RefAuxData( ) );
	    dump_container( "auxTable", o.RefAuxTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_DETECTOR:
	{
	  const FrDetector& o( dynamic_cast< const FrDetector& >( Base ) );
	
	  header( "FrDetector" );
	  
	  std::string prefix;
	  prefix += o.GetPrefix( )[ 0 ];
	  prefix += o.GetPrefix( )[ 1 ];

	  dump_label l( std::cout, m_pad, 15, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "prefix", prefix );
	  l( "longitude", o.GetLongitude( ) );
	  l( "latitude", o.GetLatitude( ) );
	  l( "elevation", o.GetElevation( ) );
	  l( "armXazimuth", o.GetArmXazimuth( ) );
	  l( "armYazimuth", o.GetArmYazimuth( ) );
	  l( "armXaltitude", o.GetArmXaltitude( ) );
	  l( "armYaltitude", o.GetArmYaltitude( ) );
	  l( "armXmidpoint", o.GetArmXmidpoint( ) );
	  l( "armYmidpoint", o.GetArmYmidpoint( ) );
	  l( "localTime", o.GetLocalTime( ) );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "aux", o.RefAux( ) );
	    dump_container( "table", o.RefTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_END_OF_FILE:
	{
	  const FrEndOfFile& o( dynamic_cast< const FrEndOfFile& >( Base ) );
	
	  header( "FrEndOfFile" );
	
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "nFrames", o.GetNFrames( ) );
	  l( "nBytes", o.GetNBytes( ) );
	  l( "chkType", o.GetChkType( ) );
	  l( "chkSum", o.GetChkSum( ) );
	  l( "seekTOC", o.GetSeekTOC( ) );
	
	  trailer( );
	}
	break;
      case CLASS_FR_END_OF_FRAME:
	{
	  const FrEndOfFrame& o( dynamic_cast< const FrEndOfFrame& >( Base ) );
	  
	  header( "FrEndOfFrame" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "run", o.GetRun( ) );
	  l( "frame", o.GetFrame( ) );
	  l( "chkType", o.GetChkType( ) );
	  l( "chkSum", o.GetChkSum( ) );
      
	  trailer( );
	}
	break;
      case CLASS_FR_EVENT:
	{
	  const FrEvent& o( dynamic_cast< const FrEvent& >( Base ) );
	  std::vector< REAL_4 > parameters;
	  std::vector< std::string > parameterNames;

	  header( "FrEvent" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "inputs", o.GetInputs( ) );
	  l( "GTimeS", o.GetGTime( ).GetSeconds( ) );
	  l( "GTimeN", o.GetGTime( ).GetNanoseconds( ) );
	  l( "timeBefore", o.GetTimeBefore( ) );
	  l( "timeAfter", o.GetTimeAfter( ) );
	  l( "eventStatus", o.GetEventStatus( ) );
	  l( "amplitude", o.GetAmplitude( ) );
	  l( "probability", o.GetProbability( ) );
	  l( "statistics", o.GetStatistics( ) );
	  l( "nParam", o.GetParam( ).size( ) );
	  for( FrEvent::ParamList_type::const_iterator
		 cur = o.GetParam( ).begin( ),
		 end = o.GetParam( ).end( );
	       cur != end;
	       ++cur )
	  {
	    parameterNames.push_back( (*cur).first );
	    parameters.push_back( (*cur).second );
	  }
	  l( "parameters", parameters );
	  l( "parameterNames", parameterNames );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "data", o.RefData( ) );
	    dump_container( "table", o.RefTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_HISTORY:
	{
	  const FrHistory& o( dynamic_cast< const FrHistory& >( Base ) );
	
	  header( "FrHistory" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  
	  l( "name", o.GetName( ) );
	  l( "time", o.GetTime( ) );
	  l( "comment", o.GetComment( ) );

	  trailer( );
	}
	break;
      case CLASS_FR_PROC_DATA:
	{
	  const FrProcData& o( dynamic_cast< const FrProcData& >( Base ) );
	  std::vector< REAL_4 > parameters;
	  std::vector< std::string > parameterNames;
	  
	  header( "FrProcData" );
	  
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "type", o.GetType( ) );
	  l( "subType", o.GetSubType( ) );
	  l( "timeOffset", o.GetTimeOffset( ) );
	  l( "tRange", o.GetTRange( ) );
	  l( "fShift", o.GetFShift( ) );
	  l( "phase", o.GetPhase( ) );
	  l( "fRange", o.GetFRange( ) );
	  l( "BW", o.GetBW( ) );
	  l( "nAuxParam", o.GetAuxParam( ).size( ) );
	  for( FrProcData::AuxParamList_type::const_iterator
		 cur = o.GetAuxParam( ).begin( ),
		 end = o.GetAuxParam( ).end( );
	       cur != end;
	       ++cur )
	  {
	    parameters.push_back( (*cur).auxParam );
	    parameterNames.push_back( (*cur).auxParamName );
	  }
	  l( "auxParam", parameters );
	  l( "auxParamNames", parameterNames );
	
	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "data", o.RefData( ) );
	    dump_container( "aux", o.RefAux( ) );
	    dump_container( "table", o.RefTable( ) );
	    dump_container( "history", o.RefHistory( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_RAW_DATA:
	{
	  const FrRawData& o( dynamic_cast< const FrRawData& >( Base ) );
	
	  header( "FrRawData" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  
	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "firstSer", o.RefFirstSer( ) );
	    dump_container( "firstAdc", o.RefFirstAdc( ) );
	    dump_container( "firsTable", o.RefFirstTable( ) );
	    dump_container( "logMsg", o.RefLogMsg( ) );
	    dump_container( "more", o.RefMore( ) );
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_SER_DATA:
	{
	  const FrSerData& o( dynamic_cast< const FrSerData& >( Base ) );
	
	  header( "FrSerData" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "timeSec", o.GetTime( ).GetSeconds( ) );
	  l( "timeNsec", o.GetTime( ).GetNanoseconds( ) );
	  l( "sampleRate", o.GetSampleRate( ) );
	  l( "data", o.GetData( ) );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "serial", o.RefSerial( ) );
	    dump_container( "table", o.RefTable( ) );
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_SIM_DATA:
	{
	  const FrSimData& o( dynamic_cast< const FrSimData& >( Base ) );
	  
	  header( "FrSimData" );
	
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "sampleRate", o.GetSampleRate( ) );
	  l( "timeOffset", o.GetTimeOffset( ) );
	  l( "fShift", o.GetFShift( ) );
	  l( "phase", o.GetPhase( ) );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "data", o.RefData( ) );
	    dump_container( "input", o.RefInput( ) );
	    dump_container( "table", o.RefTable( ) );
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_SIM_EVENT:
	{
	  const FrSimEvent& o( dynamic_cast< const FrSimEvent& >( Base ) );
	  std::vector< REAL_4 > parameters;
	  std::vector< std::string > parameterNames;
	
	  header( "FrSimEvent" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "inputs", o.GetInputs( ) );
	  l( "GTimeS", o.GetGTime( ).GetSeconds( ) );
	  l( "GTimeN", o.GetGTime( ).GetNanoseconds( ) );
	  l( "timeBefore", o.GetTimeBefore( ) );
	  l( "timeAfter", o.GetTimeAfter( ) );
	  l( "amplitude", o.GetAmplitude( ) );
	  l( "nParam", o.GetParam( ).size( ) );
	  for( FrSimEvent::ParamList_type::const_iterator
		 cur = o.GetParam( ).begin( ),
		 end = o.GetParam( ).end( );
	       cur != end;
	       ++cur )
	  {
	    parameterNames.push_back( (*cur).first );
	    parameters.push_back( (*cur).second );
	  }
	  l( "parameters", parameters );
	  l( "parameterNames", parameterNames );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "data", o.RefData( ) );
	    dump_container( "table", o.RefTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_STAT_DATA:
	{
	  const FrStatData& o( dynamic_cast< const FrStatData& >( Base ) );

	  header( "FrStatData" );
	  
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "representation", o.GetRepresentation( ) );
	  l( "timeStart", o.GetTimeStart( ) );
	  l( "timeEnd", o.GetTimeEnd( ) );
	  l( "version", o.GetVersion( ) );

	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    //:TODO: Dump detector
	    dump_container( "data", o.RefData( ) );
	    dump_container( "table", o.RefTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_SUMMARY:
	{
	  const FrSummary& o( dynamic_cast< const FrSummary& >( Base ) );
	  
	  header( "FrSummary" );
	  
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "test", o.GetTest( ) );
	  l( "GTimeS", o.GetGTime( ).GetSeconds( ) );
	  l( "GTimeN", o.GetGTime( ).GetNanoseconds( ) );
	
	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "moments", o.RefMoments( ) );
	    dump_container( "table", o.RefTable( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_TABLE:
	{
	  const FrTable& o( dynamic_cast< const FrTable& >( Base ) );
	  
	  header( "FrTable" );
	
	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "comment", o.GetComment( ) );
	  l( "nColumn", o.GetNColumn( ) );
	  l( "nRow", o.GetNRow( ) );

	  //:TODO: dump columnName
	
	  switch ( m_output_mode )
	  {
	  case OUTPUT_MODE_DUMP:
	    dump_container( "column", o.RefColumn( ) );
	    break;
	  case OUTPUT_MODE_DUMP_OBJECTS:
	    break;
	  }

	  trailer( );
	}
	break;
      case CLASS_FR_TOC:
	{
	  const FrTOC& o( dynamic_cast< const FrTOC& >( Base ) );
	  
	  header( "FrTOC" );
	
	  dump_label l( std::cout, m_pad, 18, m_scientific_data_mode );
	  l( "ULeapS", o.GetULeapS( ) );
	  l( "dataQuality", o.GetDataQuality( ) );
	  l( "GTimeS", o.GetGTimeS( ) );
	  l( "GTimeN", o.GetGTimeN( ) );
	  l( "dt", o.GetDt( ) );
	  l( "runs", o.GetRuns( ) );
	  l( "frame", o.GetFrame( ) );
	  l( "positionH", o.GetPositionH( ) );
	  l( "nFirstADC", o.GetNFirstADC( ) );
	  l( "nFirstSer", o.GetNFirstSer( ) );
	  l( "nFirstTable", o.GetNFirstTable( ) );
	  l( "nFirstMsg", o.GetNFirstMsg( ) );
	  l( "nSH", o.GetNSH( ) );
	  l( "SHid", o.GetSHid( ) );
	  l( "SHname", o.GetSHname( ) );

	  //-------------------------------------------------------------
	  l( "nDetector", o.GetNDetector( ) );
	  {
	    std::vector< STRING >	name;
	    std::vector< INT_8U >	position;
	    
	    const FrTOC::MapDetector_type&	n( o.GetDetector( ) );

	    for ( FrTOC::MapDetector_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      name.push_back( cur->first );
	      position.push_back( cur->second );
	    }

	    l( "nameDetector", name );
	    l( "positionDetector", position );
	  }

	  //-------------------------------------------------------------
	  l( "nStatType", o.GetNStatType( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< STRING >		detector;
	    std::vector< INT_4U >		n_stat_inst;
	    std::vector< INT_4U >		tstart;
	    std::vector< INT_4U >		tend;
	    std::vector< INT_4U >		version;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapStatType_type&	n( o.GetStatType( ) );

	    for ( FrTOC::MapStatType_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      const FrTOC::StatType_type&	stat( cur->second );

	      name.push_back( cur->first.s_name );
	      detector.push_back( cur->first.s_detector );
	      n_stat_inst.push_back( stat.tStart.size( ) );

	      for ( INT_4U
		      x = 0,
		      xlast = stat.tStart.size( );
		    x != xlast;
		    ++x )
	      {
		tstart.push_back( stat.tStart[ x ] );
		tend.push_back( stat.tEnd[ x ] );
		version.push_back( stat.version[ x ] );
	      }
	      for ( FrTOC::position_list_type::const_iterator
		      y = stat.positionStat.begin( ),
		      ylast = stat.positionStat.end( );
		    y != ylast;
		    ++y )
	      {
		position.push_back( *y );
	      }
	    }

	    l( "nameStat", name );
	    l( "detector", detector);
	    l( "nStatInstance", n_stat_inst );
	    l( "tStart", tstart );
	    l( "tEnd", tend );
	    l( "version", version );
	    l( "positionStat", position );
	  }


	  //-------------------------------------------------------------
	  l( "nADC", o.GetNADC( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< INT_4U >		channel_id;
	    std::vector< INT_4U >		group_id;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapADC_type&	n( o.GetADC( ) );

	    for ( FrTOC::MapADC_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      const FrTOC::ADC_type&	adc( cur->second );

	      name.push_back( cur->first );
	      channel_id.push_back( adc.channelID );
	      group_id.push_back( adc.groupID );
	      
	      for ( INT_4U
		      x = 0,
		      xlast = adc.positionADC.size( );
		    x != xlast;
		    ++x )
	      {
		position.push_back( adc.positionADC[ x ] );
	      }
	    }

	    l( "nameADC", name );
	    l( "channelID", channel_id );
	    l( "groupID", group_id );
	    l( "positionADC", position );
	  }

	  //-------------------------------------------------------------
	  l( "nProc", o.GetNProc( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapProc_type&	n( o.GetProc( ) );

	    for ( FrTOC::MapProc_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      name.push_back( cur->first );
	      for ( INT_4U
		      x = 0,
		      xlast = cur->second.size( );
		    x != xlast;
		    ++x )
	      {
		position.push_back( cur->second[ x ] );
	      }
	    }

	    l( "nameProc", name );
	    l( "positionProc", position );
	  }

	  //-------------------------------------------------------------
	  l( "nSim", o.GetNSim( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapSim_type&	n( o.GetSim( ) );

	    for ( FrTOC::MapSim_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      name.push_back( cur->first );
	      for ( INT_4U
		      x = 0,
		      xlast = cur->second.size( );
		    x != xlast;
		    ++x )
	      {
		position.push_back( cur->second[ x ] );
	      }
	    }

	    l( "nameSim", name );
	    l( "positionSim", position );
	  }

	  //-------------------------------------------------------------
	  l( "nSer", o.GetNSer( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapSer_type&	n( o.GetSer( ) );

	    for ( FrTOC::MapSer_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      name.push_back( cur->first );
	      for ( INT_4U
		      x = 0,
		      xlast = cur->second.size( );
		    x != xlast;
		    ++x )
	      {
		position.push_back( cur->second[ x ] );
	      }
	    }

	    l( "nameSer", name );
	    l( "positionSer", position );
	  }

	  //-------------------------------------------------------------
	  l( "nSummary", o.GetNSummary( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapSummary_type&	n( o.GetSummary( ) );

	    for ( FrTOC::MapSummary_type::const_iterator
		    cur = n.begin( ),
		    last = n.end( );
		  cur != last;
		  ++cur )
	    {
	      name.push_back( cur->first );
	      for ( INT_4U
		      x = 0,
		      xlast = cur->second.size( );
		    x != xlast;
		    ++x )
	      {
		position.push_back( cur->second[ x ] );
	      }
	    }

	    l( "nameSum", name );
	    l( "positionSum", position );
	  }

	  //-------------------------------------------------------------
	  l( "nEventType", o.GetNEventType( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< INT_4U >		n;
	    std::vector< INT_4U >		gtimes;
	    std::vector< INT_4U>		gtimen;
	    std::vector< REAL_4 >		amplitude;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapEventType_type&	el( o.GetEventType( ) );

	    for ( FrTOC::MapEventType_type::const_iterator
		    cur = el.begin( ),
		    last = el.end( );
		  cur != last;
		  ++cur )
	    {
	      const FrTOC::EventList_type& e( cur->second );

	      n.push_back( e.size( ) );
	      name.push_back( cur->first );
	      for ( FrTOC::EventList_type::const_iterator
		      ecur = e.begin( ),
		      elast = e.end( );
		    ecur != elast;
		    ++ecur )
	      {
		gtimes.push_back( ecur->GTimeS );
		gtimen.push_back( ecur->GTimeN );
		amplitude.push_back( ecur->amplitude );
		position.push_back( ecur->position );
	      }
	    }
	    
	    l( "nameEventType", name );
	    l( "nEvent", n );
	    l( "GTimeS", gtimes );
	    l( "GTimeN", gtimen );
	    l( "amplitudeEvent", amplitude );
	    l( "positionEvent", position );
	  }

	  //-------------------------------------------------------------
	  l( "nSimEventType", o.GetNSimEventType( ) );
	  {
	    std::vector< STRING >		name;
	    std::vector< INT_4U >		n;
	    std::vector< INT_4U >		gtimes;
	    std::vector< INT_4U>		gtimen;
	    std::vector< REAL_4 >		amplitude;
	    std::vector< FrTOC::position_type >	position;
	    
	    const FrTOC::MapSimEventType_type&	el( o.GetSimEventType( ) );

	    for ( FrTOC::MapSimEventType_type::const_iterator
		    cur = el.begin( ),
		    last = el.end( );
		  cur != last;
		  ++cur )
	    {
	      const FrTOC::EventList_type& e( cur->second );

	      n.push_back( e.size( ) );
	      name.push_back( cur->first );
	      for ( FrTOC::EventList_type::const_iterator
		      ecur = e.begin( ),
		      elast = e.end( );
		    ecur != elast;
		    ++ecur )
	      {
		gtimes.push_back( ecur->GTimeS );
		gtimen.push_back( ecur->GTimeN );
		amplitude.push_back( ecur->amplitude );
		position.push_back( ecur->position );
	      }
	    }
	    
	    l( "nameSimEventType", name );
	    l( "nSimEvent", n );
	    l( "GTimeSSim", gtimes );
	    l( "GTimeNSim", gtimen );
	    l( "amplitudeSimEvent", amplitude );
	    l( "positionSimEvent", position );
	  }
	}
	break;
      case CLASS_FR_VECT:
	{
	  const FrVect& o( dynamic_cast< const FrVect& >( Base ) );
	  std::vector< INT_8U >	nx;
	  std::vector< REAL_8 >	dx;
	  std::vector< REAL_8 >	startX;
	  std::vector< STRING >	unitX;
	  for ( INT_4U x = 0; x < o.GetNDim( ); x++ )
	  {
	    const Dimension& d( o.GetDim( x ) );
	    nx.push_back( d.GetNx( ) );
	    dx.push_back( d.GetDx( ) );
	    startX.push_back( d.GetStartX( ) );
	    unitX.push_back( d.GetUnitX( ) );
	  }
	
	  header( "FrVect" );

	  dump_label l( std::cout, m_pad, 10, m_scientific_data_mode );
	  l( "name", o.GetName( ) );
	  l( "compress", o.GetCompress( ) );
	  l( "type", o.GetType( ) );
	  l( "nData", o.GetNData( ) );
	  l( "nBytes", o.GetNBytes( ) );
	  l( "nDim", o.GetNDim( ) );
	  l( "nx", nx );
	  l( "dx", dx );
	  l( "startX", startX );
	  l( "unitX", unitX );
	  l( "unitY", o.GetUnitY( ) );

	  if ( m_verbose_data_mode == false )
	  {
	    l( "data", "..." );
	  }
	  else
	  {
	    CHAR_U*	data( new CHAR_U[ o.GetNData( ) * o.GetTypeSize( ) ] );
	    try
	    {
	      // .Decompress( data );
	      memcpy( data, o.GetData( ), o.GetNData( ) * o.GetTypeSize( ) );
	      switch( o.GetType( ) )
	      {
	      case FrVect::FR_VECT_C:
		l( "data", vect< CHAR >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_2S:
		l( "data", vect< INT_2S >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_8R:
		l( "data", vect< REAL_8 >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_4R:
		l( "data", vect< REAL_4 >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_4S:
		l( "data", vect< INT_4S >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_8S:
		l( "data", vect< INT_8S >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_8C:
		l( "data", vect< COMPLEX_8 >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_16C:
		l( "data", vect< COMPLEX_16 >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_2U:
		l( "data", vect< INT_2U >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_4U:
		l( "data", vect< INT_4U >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_8U:
		l( "data", vect< INT_8U >( data, o.GetNData( ) ) );
		break;
	      case FrVect::FR_VECT_1U:
		l( "data", vect< CHAR_U >( data, o.GetNData( ) ) );
		break;
	      }
	      delete[] data;
	    }
	    catch( ... )
	    {
	      delete[] data;
	      throw;
	    }
	  }
	  
	  trailer( );
	}
	break;
      }
    }

    void DumpFrStruct::
    header( const std::string& tag )
    {
      switch( m_output_mode )
      {
      case OUTPUT_MODE_DUMP:
	std::cout << m_pad << tag << std::endl;
	std::cout << m_pad << "{" << std::endl;
	pad_inc( );
	break;
      case OUTPUT_MODE_DUMP_OBJECTS:
	pad_inc( );
	break;
      }
    }

    void DumpFrStruct::
    trailer( )
    {
      switch( m_output_mode )
      {
      case OUTPUT_MODE_DUMP:
	pad_dec( );
	std::cout << m_pad << "}" << std::endl;
	break;
      case OUTPUT_MODE_DUMP_OBJECTS:
	pad_dec( );
	break;
      }
    }
    const std::string DumpFrStruct::m_spaces( "  " );
  } // namespace - Version6
} // namespace - anonymous
