#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <string>

#include "general/unittest.h"

#include "framecpp/FrameFilename.hh"

General::UnitTest Test;

void
test( const std::string& Filename )
{
  try
  {
    FrameCPP::FrameFilename	ff( Filename );

    Test.Check( ff.S( ).find( '-' ) == std::string::npos )
      << Filename << ": S does not contain any hypens" << std::endl;
    Test.Check( ff.D( ).find( '-' ) == std::string::npos )
      << Filename << ": D does not contain any hypens" << std::endl;
    Test.Check( ff.G( ) > 0 )
      << Filename << ": GPS start time > 0" << std::endl;
    Test.Check( ff.T( ) > 0 )
      << Filename << ": Total Time > 0" << std::endl;
  }
  catch( const FrameCPP::FrameFilename::InvalidFrameFilename& e )
  {
    Test.Check( false )
      << Filename << " ";
    for ( size_t x = 0, end = e.getSize( ) ; x < end; ++x )
    {
      Test.Message( false )
	<< e[ x ].getMessage( )
	<< std::endl;
    }
  }
}

int
main( int ArgC, char* ArgV[] )
try
{
  Test.Init( ArgC, ArgV );

  //---------------------------------------------------------------------
  // Testing of class 1
  //---------------------------------------------------------------------
  // 1 second of raw GEO data
  test( "G-R-650123123-1.gwf" );
  // 8 seconds of raw LIGO data from Hanford
  test( "H-R-688775627-8.gwf" );
  // One minute of second-trend data from Livingston
  test( "L-T-688775600-60.gwf" );
  // One hour of minute-trend data fro mboth Hanford and Livingston
  test( "HL-M-68877500-3600.gwf" );
  //---------------------------------------------------------------------
  // Testing of class 2
  //---------------------------------------------------------------------
  // 256 seconds of "Level 2" data (e.g. selected channesl, possibly
  //   with some offline processing) from Hanford
  test( "H-level2-693020060-256.gwf" );
  // 1000 seocnds of PEM data from Livingston
  test( "L-PEM-720878444-1000.gwf" );
  // Some simultated data containing inspiral signals
  test( "HL-SimInspiral-730282200-16.gwf" );
  //---------------------------------------------------------------------
  // Testing of class 3
  //---------------------------------------------------------------------
  test( "pss-filter_test-787384722-8.gwf" );
  test( "iul-simset1-694215520-32.gwf" );
  test( "H-calibPEM-720878444-1000.gwf" );

  Test.Exit();
}
catch ( ... )
{
  Test.Check( false ) << "Caught an exception" << std::endl;
  Test.Exit( );
}
