/*
  frdump -- dump frame file contents
*/

#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fstream>
#include <iostream>

#include "framecpp/Version3_4_5/frame.hh"
#include "framecpp/Version3_4_5/framewritertoc.hh"
#include "framecpp/Version3_4_5/framereader.hh"
#include "framecpp/Version3_4_5/errors.hh"
#include "framecpp/Version3_4_5/dump.hh"

using namespace FrameCPP::Version_3_4_5;
using namespace std;

// The name of the program executable
char* programname = 0;

void
fatal( string msg ) {
  cerr << msg << endl;
  exit (1);
}

void
error_watch (const string& msg)
{
  cerr << "error: " << msg << endl;
}


void
usage( const char* f )
{
  if ( ! strncasecmp( f, "-h", 2 ) 
       || ! strncasecmp( f, "--h", 3 ) )
    {
      cout << "Dump frame file contents" << endl;
      cout << "Usage: " << (string)programname << " INFILE" << endl;
      exit(0);
    }
}

int
main( int argc, char* argv[] )
{  
  /* Determine program name (strip filesystem path) */
  for (programname = argv [0] + strlen (argv [0]) - 1;
       programname != argv [0] && *programname != '/';
       programname--)
    ;

  if (*programname == '/') 
    programname++;

  if ( argc > 1 )
    usage( argv[1] );
  
#if 0
  {
    Dictionary *dict = library.getCurrentVersionDictionary();
    // cerr << "Current version is " << dict->getVersion() << endl;
  }
#endif

  if (argc != 2) {
    usage("-h");
    return 0;
  }

  std::ifstream in( argv[ 1 ] );
  
  FrameReader fr( in );
  fr.setErrorWatch( error_watch );

  try {
    for(;!fr.isEndOfFile();) {
      Frame* frame = fr.readFrame();
      if (!frame)
	break;
      dump(cout, *frame);
      delete frame;
    }
  } catch ( read_failure& e ) {
     fatal( e.what() );
  }

  in.close();
  return 0;
}
