/*
  frtoc -- filter to process a frame file: add the TOC (Table Of Contents)
*/

#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fstream>
#include <iostream>

#include "framecpp/Version3_4_5/frame.hh"
#include "framecpp/Version3_4_5/framewritertoc.hh"
#include "framecpp/Version3_4_5/framereader.hh"
#include "framecpp/Version3_4_5/errors.hh"

using namespace FrameCPP::Version_3_4_5;
using namespace std;

// The name of the program executable
char* programname = 0;

void
fatal( string msg ) {
  cerr << msg << endl;
  exit (1);
}

void
error_watch (const string& msg)
{
  cerr << "error: " << msg << endl;
}


void
usage( char* f )
{
  if ( ! strncasecmp( f, "-h", 2 ) 
       || ! strncasecmp( f, "--h", 3 ) )
    {
      cout << "Filter to process a frame file: add the Table Of Contents." << endl;
      cout << "Usage: " << (string)programname << " INFILE OUTFILE" << endl;
      cout << "As filter: " << (string)programname << " < INFILE > OUTFILE" << endl;
      exit(0);
    }
}

int
main( int argc, char* argv[] )
{  
  std::istream* in=0;
  std::ostream* out=0;

  /* Determine program name (strip filesystem path) */
  for (programname = argv [0] + strlen (argv [0]) - 1;
       programname != argv [0] && *programname != '/';
       programname--)
    ;

  if (*programname == '/') 
    programname++;

  if ( argc > 1 )
    usage( argv[1] );
  
  if ( argc < 3 )
    {
      if ( argc < 2 )
	{
	  in = &( std::cin );
	  out = &( std::cout );
	}
      else
	{
	  in = new std::ifstream( argv[ 1 ] );
	  out = &( std::cout );
	}
  } else {
    in = new std::ifstream( argv[ 1 ] );
    out = new std::ofstream( argv[ 2 ] );
  }

  FrameReader fr( *in );
  fr.setErrorWatch( error_watch );
  FrameWriterTOC fw( *out );

  // :TODO: why don't we have the error watch in the output class???
  //  fw.setErrorWatch( error_watch );

  try {
    for(;;) {
      Frame* frame = fr.readFrame();
      if (!frame)
	break;
      fw.writeFrame( *frame );
      delete frame;
    }
  } catch ( read_failure& e ) {
    if ( e.code() != fr.ENDOFFILE )
      {
	fatal( e.what() );
      }
  }

  fw.close();   // build TOC
  if ( dynamic_cast< std::ofstream* >( out ) )
  {
    dynamic_cast< std::ofstream* >( out )->close();
  }
  if ( dynamic_cast< std::ifstream* >( in ) )
  {
    dynamic_cast< std::ifstream* >( in )->close();
  }
  return 0;
}
