#include "general/config.h"

#include "general/TaskThread.hh"
#include "general/ThreadPool.hh"

namespace General
{
  ThreadPoolClient& ThreadPool = ThreadPoolSingleton::Instance( );

  ThreadPoolClient::
  ThreadPoolClient( )
  {
    MutexLock::Initialize( m_queue_lock );
  }

  ThreadPoolClient::
  ~ThreadPoolClient( )
  {
    if ( m_in_use.size( ) > 0 )
    {
      MutexLock	lock( m_queue_lock );
      
      //-----------------------------------------------------------------
      // Get rid of threads currently doing something interesting
      //-----------------------------------------------------------------
      for ( thread_container_type::const_iterator
	      cur = m_in_use.begin( ),
	      last = m_in_use.end( );
	    cur != last;
	    ++cur )
      {
	delete( *cur );
      }
      m_in_use.erase( m_in_use.begin( ), m_in_use.end( ) );
    }
    //-------------------------------------------------------------------
    // Get rid of threads waiting to do something interesting
    //-------------------------------------------------------------------
    for ( thread_container_type::const_iterator
	    cur = m_available.begin( ),
	    last = m_available.end( );
	  cur != last;
	  ++cur )
    {
      delete( *cur );
    }
    m_available.erase( m_available.begin( ), m_available.end( ) );
  }

  TaskThread* ThreadPoolClient::
  Acquire( )
  {
    MutexLock	lock( m_queue_lock );

    if ( m_available.size( ) == 0 )
    {
      m_available.push_front( new TaskThread );
      m_available.front( )->Start( );
    }
    m_in_use.push_front( m_available.front( ) );
    m_available.pop_front( );
    return m_in_use.front( );
  }

  void ThreadPoolClient::
  Relinquish( TaskThread* Slot )
  {
    if ( Slot )
    {
      MutexLock lock( m_queue_lock );

      m_in_use.remove( Slot );
      m_available.push_back( Slot );
    }
  }
  
}
