#include "general/config.h"

#include <signal.h>

#include <list>
#include <sstream>

#include "general/ErrorLog.hh"
#include "general/TaskThread.hh"
#include "general/ThreadPool.hh"
#include "general/Timeout.hh"

namespace General
{
  void Timeout( Task& CallerTask, int Seconds )
  {
    TaskThread*	slave = ThreadPool.Acquire( );
    if ( slave )
    {
      MutexLock::Lock( slave->Mutex( ) );

      slave->AddTask( CallerTask );
      bool rc = slave->TimedWait( Seconds );
      bool joinable = true;
      slave->CriticalSection( false );
      if ( rc == false )
      {
	{
	  //-------------------------------------------------------------
	  // Mark the thread as timed out
	  //-------------------------------------------------------------
	  slave->TimedOut( );
	}
	// Notify the slave that it should stop whatever it is doing
	joinable = slave->Halt( );
	if ( joinable )
	{
	  // Need to wait for the slave to complete
	  slave->Wait( );
	}
      }
      if ( joinable )
      {
	ThreadPool.Relinquish( slave );
      }
    }
  }
}
