/* -*- mode: c++ ; c-basic-offset: 2;-*- */
#ifndef GENERAL__HASH_MAP_HH
#define	GENERAL__HASH_MAP_HH

#include "general/config.h"

#include <string>

#if HAVE_STLPORT_HASH_MAP  // stlport/hash_map was found
#include <hash_map>
#define HASH_MAP_NAMESPACE std
#elif HAVE_HASH_MAP        // hash_map was found
#include <hash_map>
#if defined( __GNUG__ ) && ( __GNUG__ < 3 )
#define HASH_NAMESPACE
#else
#define HASH_MAP_NAMESPACE std
#endif                  
#elif HAVE_EXT_HASH_MAP    // ext/hash_map was found
#include <ext/hash_map>
#if defined( __GNUG__ )
#define HASH_MAP_NAMESPACE __gnu_cxx
#endif
#else                       // no hash_map header file exists
#include <map>
#endif


#ifndef HASH_MAP_NAMESPACE
#define HASH_MAP_NAMESPACE std
#endif

#ifndef HASH_NAMESPACE
#define HASH_NAMESPACE HASH_MAP_NAMESPACE
#endif

namespace HASH_MAP_NAMESPACE
{
  template<>
  class hash< ::std::string >
  {
  public:
    //: Overloaded call operator.
    //  
    //!param: const std::string& s - A reference to the string representing
    //+		the search key.
    //
    //!return: size_t - Index into the hash.
    // 
    size_t operator()( const ::std::string& s ) const
    {
      HASH_NAMESPACE::hash< const char* > h;
      return h( s.c_str() );
    }
  }; // struct hash< string >

#if NEED_HASH_LONG_LONG
  template<>
  struct hash<long long>
  {
    size_t
    operator()(long long __x) const
    { return __x; }
  };
#endif /* NEED_HASH_LONG_LONG */

#if NEED_HASH_UNSIGNED_LONG_LONG
  template<>
  struct hash<unsigned long long>
  {
    size_t
    operator()(unsigned long long __x) const
    { return __x; }
  };
#endif /* NEED_HASH_UNSIGNED_LONG_LONG */
}

namespace General
{
  template < class Key >
  class hash
#if defined( HASH_NAMESPACE )
    : public HASH_NAMESPACE::hash< Key >
#endif /* HASH_NAMESPACE */
  {
  };

  template< class Key, class T,
	    class H = General::hash< Key >,
	    class EQ = std::equal_to< Key >,
	    class A = std::allocator< std::pair< const Key, T > > >
  class hash_map
    :
#if defined( HASH_MAP_NAMESPACE )
    public HASH_MAP_NAMESPACE::hash_map< Key, T, H, EQ, A >
#else /* HASH_MAP_NAMESPACE */
    public std::map< Key, T, EQ, A >
#endif /* HASH_MAP_NAMESPACE */
  {
  };

}

#endif /* GENERAL__HASH_MAP */
