/* -*- mode: c++; c-basic-offset: 4; -*- */

#include "general/config.h"

#include "general/DeadLockDetector.hh"
#include "general/mutexlock.hh"

#if DEAD_LOCK_DETECTOR_ENABLED
using General::DeadLockDetector;

namespace
{
    //: Log the thread state
    inline void log_lock( DeadLockDetector::state_type State,
			  MutexLock::lock_type* Lock,
			  char* const File,
			  unsigned int Line,
			  bool Logging )
    {
	if ( Logging )
	{
	    DeadLockDetector::SetState( DeadLockDetector::MUTEX,
					Lock,
					State,
					File, Line );
	}
    }
}

#else /* DEAD_LOCK_DETECTOR_ENABLED */
#define log(a)
#endif /* DEAD_LOCK_DETECTOR_ENABLED */

MutexLock::
MutexLock(lock_type& mutex, const char* const File, const unsigned int Line,
	  bool Logging )
    : m_mutex_ptr(&mutex),
      m_logging( Logging )
{
    int err;

    log_lock( DeadLockDetector::PENDING, m_mutex_ptr,
	      __FILE__, __LINE__,
	      m_logging );
    if ( ( err = pthread_mutex_lock(m_mutex_ptr) ) != 0 )
    {
	std::cerr << "MutexLock: "<< strerror( err )
		  << " File: " << __FILE__ << " Line: " << __LINE__
		  << std::endl;
	log_lock( DeadLockDetector::RELEASED, m_mutex_ptr,
		  __FILE__, __LINE__,
		  m_logging );
	assert( false );
    }
    else
    {
	log_lock( DeadLockDetector::ACQUIRED, m_mutex_ptr,
		  __FILE__, __LINE__,
		  m_logging );
    }
}

MutexLock::
MutexLock(lock_type* const mutex_ptr,
	  const char* const File, const unsigned int Line,
	  bool Logging )
    : m_mutex_ptr(mutex_ptr),
      m_logging( Logging )
      
{
    int err;

    log_lock( DeadLockDetector::PENDING, m_mutex_ptr,
	      __FILE__, __LINE__,
	      m_logging );
    if ( ( err = pthread_mutex_lock(m_mutex_ptr) ) != 0 )
    {
	std::cerr << "MutexLock: " << strerror( err )
		  << " File: " << __FILE__ << " Line: " << __LINE__
		  << std::endl;
	log_lock( DeadLockDetector::RELEASED, m_mutex_ptr,
		  __FILE__, __LINE__,
		  m_logging );
    }
    else
    {
	log_lock( DeadLockDetector::ACQUIRED, m_mutex_ptr,
		  __FILE__, __LINE__,
		  m_logging );
    }
}

MutexLock::
~MutexLock()
{
    Release( __FILE__, __LINE__ );
}

void MutexLock::
Release( const char* const File, const unsigned int Line )
{
    if ( m_mutex_ptr )
    {
	int err;

	if ( ( err = pthread_mutex_unlock(m_mutex_ptr) ) != 0 )
	{
	    std::cerr << "MutexLock::Release : "
		      << " File: " << __FILE__ << " Line: " << __LINE__
		      << strerror( err )
		      << std::endl;

	}
	else
	{
	    log_lock( DeadLockDetector::RELEASED, m_mutex_ptr,
		      __FILE__, __LINE__,
		      m_logging );
	}
	m_mutex_ptr = (lock_type*)NULL;
    }
}
