#include "general/config.h"

#include "regex.hh"
#include "errors.hh"

using General::Errors::getErrorString;

//-----------------------------------------------------------------------------
//
//: Constructor
//
// This constructor compiles the given regular expression.
//
//!param: const std::string& regex
//!param: int cflags
//
//!exc: LdasException
//
Regex::Regex( const std::string& regex, int cflags )
    throw( LdasException, std::bad_alloc )
        : mRegex( regex ), mPattern()
{
    int error = regcomp( &mPattern, mRegex.c_str(), cflags );
    if ( error != 0 )
    {
        size_t msgSize( regerror( error, &mPattern, 0, 0 ) );
        char* msg = new char[ msgSize ];
        regerror( error, &mPattern, msg, msgSize );
        
        LdasException e( Library::GENERAL, General::Errors::INVALID_REGEX,
                         getErrorString( General::Errors::INVALID_REGEX ),
			 msg,
                         __FILE__, __LINE__ );
        delete[] msg;
        msg = 0;

        throw e;
    }
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This just frees the compiled regex pattern.
//
Regex::~Regex() throw()
{
    regfree( &mPattern );
}


