/* -*- mode: C++ -*- */
#ifndef UNITTEST_H
#define	UNITTEST_H

#include <stdlib.h>
#include <strings.h>

#include <cstdlib>
#include <cstring>

#include <iostream>
#include <fstream>
#include <string>

namespace General {
  //: Provide standard routines for unit testing of code.
  //
  class UnitTest
  {
  public:
    //: Constructor
    UnitTest(void);
    //: Destructor
    ~UnitTest(void);
    //: Check for error conditions
    std::ostream& Check(bool TestResult);
    //: Check for error conditions
    void Check(bool TestResult, std::string Message);
    //: Exit with the appropriate exit code
    void Exit(void);
    //: Initialize the class with command line arguments
    void Init(int ArgC, char** ArgV);
    //: Check if in verbose mode
    bool IsVerbose( int Level = 0 ) const
    {
       return ( Level <= m_verbose );
    }   
    //: Output info message
    std::ostream& Message( bool Leader = true ) const;
    //: Output info message
    std::ostream& Message( int Level, bool Leader = true ) const;

  private:
    //: Level of verbosity
    int			m_verbose;
    //: Failure status flag
    bool		m_status;
    //: Pointer to stream when in non-verbose mode
    std::ostream*	m_null_stream;
  }; // class - UnitTest

  inline UnitTest::
  UnitTest(void)
    : m_verbose( -1 ),
      m_status(true)
  {
    m_null_stream = new std::ofstream("/dev/null");
    if ( getenv("TEST_VERBOSE_MODE") )
    {
      if ( 0 == strcmp( getenv( "TEST_VERBOSE_MODE" ), "true" ) )
      {
	m_verbose = 0;
      }
      else if ( 0 == strcmp( getenv( "TEST_VERBOSE_MODE" ), "false" ) )
      {
	m_verbose = -1;
      }
      else
      {
	m_verbose = atoi( getenv( "TEST_VERBOSE_MODE" ) );
      }
    }
  }
  
  inline UnitTest::
  ~UnitTest(void)
  {
    if (m_null_stream != 0)
    {
      dynamic_cast<std::ofstream*>(m_null_stream)->close();
      delete m_null_stream;
    }
  }

  inline std::ostream& UnitTest::
  Check(bool TestResult)
  {
    m_status = m_status && TestResult;
    if ( IsVerbose() )
    {
      std::cout << "-- " << ((TestResult) ? "PASS" : "FAIL") << ": ";
      return std::cout;
    }
    return *m_null_stream;
  }

  inline void UnitTest::
  Check(bool TestResult, std::string Message)
  {
    m_status = m_status && TestResult;
    if ( IsVerbose() )
    {
      std::cout << "-- " << ((TestResult) ? "PASS" : "FAIL") << ": " << Message
		<< std::endl << std::flush;
    }
  }

  inline void UnitTest::
  Exit(void)
  {
    exit ((m_status) ? 0 : 1);
  }

  inline void UnitTest::
  Init(int ArgC, char** ArgV)
  {
    for (int x = 1; x < ArgC; x++)
    {
      if (0 == (strcmp(ArgV[x], "--verbose")))
      {
	m_verbose = 0;
	continue;
      }
    }
  }

  inline std::ostream& UnitTest::
  Message( bool Leader ) const
  {
    if ( IsVerbose() )
    {
      if (Leader)
      {
	std::cout << "-- MESG: ";
      }
      return std::cout;
    }
    return *m_null_stream;
  }

  inline std::ostream& UnitTest::
  Message( int Level, bool Leader ) const
  {
    if ( IsVerbose( Level ) )
    {
      return Message( Leader );
    }
    return *m_null_stream;
  }
} // namespace - General

#endif	/* UNITTEST_H */
